CREATE DATABASE plone;
GRANT ALL ON plone.* TO 'plone'@'%' IDENTIFIED BY 'plonepass';
FLUSH PRIVILEGES;
CHANGE MASTER TO
    MASTER_HOST='192.168.1.61',
    MASTER_USER='repl',
    MASTER_PASSWORD='slavepass',
    MASTER_LOG_FILE='mysql-bin.000001',
    MASTER_LOG_POS=4,
    MASTER_CONNECT_RETRY=10;
START SLAVE;
