CREATE DATABASE plone;
GRANT ALL ON plone.* TO 'plone'@'%' IDENTIFIED BY 'plonepass';
GRANT REPLICATION SLAVE ON *.* TO 'repl'@'%' IDENTIFIED BY 'slavepass';
FLUSH PRIVILEGES;
RESET MASTER;
SHOW MASTER STATUS;
