/* Generated by Cython 3.0a6 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/jmadden/Projects/GithubSources/relstorage/src/relstorage/cache/c_cache.h"
        ],
        "include_dirs": [
            "/Users/jmadden/Projects/GithubSources/relstorage/src/relstorage/cache",
            "include"
        ],
        "language": "c++",
        "name": "relstorage.cache.cache",
        "sources": [
            "src/relstorage/cache/cache.pyx",
            "src/relstorage/cache/c_cache.cpp"
        ]
    },
    "module_name": "relstorage.cache.cache"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#define CYTHON_ABI "3_0a6"
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x030000A6
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP487_INIT_SUBCLASS
  #define CYTHON_PEP487_INIT_SUBCLASS 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(CYTHON_LIMITED_API)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030800B2
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#elif PY_VERSION_HEX >= 0x030800A4
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS(n)
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((n) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET)
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  (n)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static inline void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
#else
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__relstorage__cache__cache
#define __PYX_HAVE_API__relstorage__cache__cache
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <utility>

    #if __cplusplus > 199711L
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <string>
#include <vector>
#include "c_cache.h"

    #ifdef __clang__
    #pragma clang diagnostic push
    /* Cython generates lots of unreachable code diagnostics that flood the output */
    #pragma clang diagnostic ignored "-Wunreachable-code"
    /* As of Cython 3.0a6 and CPython 3.8 (at least) Cython generates
       deprecation warnings for tp_print */
    #pragma clang diagnostic ignored "-Wdeprecated-declarations"
    #endif

    template <typename T>
    T* array_new(int n) {
        return new T[n];
    }
    template <typename T>
    void array_delete(T* t) {
        delete[] t;
    }
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const wchar_t *u)
{
    const wchar_t *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/relstorage/cache/cache.pyx",
  "stringsource",
  "type.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_10relstorage_5cache_5cache_CachedValue;
struct __pyx_obj_10relstorage_5cache_5cache_SingleValue;
struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues;
struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration;
struct __pyx_obj_10relstorage_5cache_5cache_PyCache;
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__;
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__;
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems;
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys;
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values;

/* "relstorage/cache/cache.pyx":66
 *     void array_delete[T](T* t)
 * 
 * ctypedef const SVCacheEntry* SVCacheEntry_p             # <<<<<<<<<<<<<<
 * ctypedef const MVCacheEntry* MVCacheEntry_p
 * ctypedef const ICacheEntry*  ICacheEntry_p
 */
typedef relstorage::cache::SVCacheEntry const *__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p;

/* "relstorage/cache/cache.pyx":67
 * 
 * ctypedef const SVCacheEntry* SVCacheEntry_p
 * ctypedef const MVCacheEntry* MVCacheEntry_p             # <<<<<<<<<<<<<<
 * ctypedef const ICacheEntry*  ICacheEntry_p
 * ctypedef ICacheEntry** ICacheEntry_pp
 */
typedef relstorage::cache::MVCacheEntry const *__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p;

/* "relstorage/cache/cache.pyx":68
 * ctypedef const SVCacheEntry* SVCacheEntry_p
 * ctypedef const MVCacheEntry* MVCacheEntry_p
 * ctypedef const ICacheEntry*  ICacheEntry_p             # <<<<<<<<<<<<<<
 * ctypedef ICacheEntry** ICacheEntry_pp
 * 
 */
typedef relstorage::cache::ICacheEntry const *__pyx_t_10relstorage_5cache_5cache_ICacheEntry_p;

/* "relstorage/cache/cache.pyx":69
 * ctypedef const MVCacheEntry* MVCacheEntry_p
 * ctypedef const ICacheEntry*  ICacheEntry_p
 * ctypedef ICacheEntry** ICacheEntry_pp             # <<<<<<<<<<<<<<
 * 
 * cdef class CachedValue:
 */
typedef relstorage::cache::ICacheEntry **__pyx_t_10relstorage_5cache_5cache_ICacheEntry_pp;

/* "relstorage/cache/cache.pyx":71
 * ctypedef ICacheEntry** ICacheEntry_pp
 * 
 * cdef class CachedValue:             # <<<<<<<<<<<<<<
 *     """
 *     The base class for cached values.
 */
struct __pyx_obj_10relstorage_5cache_5cache_CachedValue {
  PyObject_HEAD
};


/* "relstorage/cache/cache.pyx":112
 * @cython.final
 * @cython.internal
 * cdef class SingleValue(CachedValue):             # <<<<<<<<<<<<<<
 *     cdef SVCacheEntry_p entry
 * 
 */
struct __pyx_obj_10relstorage_5cache_5cache_SingleValue {
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue __pyx_base;
  struct __pyx_vtabstruct_10relstorage_5cache_5cache_SingleValue *__pyx_vtab;
  __pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p entry;
};


/* "relstorage/cache/cache.pyx":204
 * 
 * @cython.final
 * cdef class MultipleValues(CachedValue):             # <<<<<<<<<<<<<<
 *     cdef MVCacheEntry_p entry
 * 
 */
struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues {
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue __pyx_base;
  struct __pyx_vtabstruct_10relstorage_5cache_5cache_MultipleValues *__pyx_vtab;
  __pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p entry;
};


/* "relstorage/cache/cache.pyx":245
 * 
 * 
 * cdef class PyGeneration:             # <<<<<<<<<<<<<<
 *     cdef Generation* generation
 *     cdef readonly object __name__
 */
struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration {
  PyObject_HEAD
  struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyGeneration *__pyx_vtab;
  relstorage::cache::Generation *generation;
  PyObject *__name__;
  struct __pyx_obj_10relstorage_5cache_5cache_PyCache *_cache;
};


/* "relstorage/cache/cache.pyx":289
 * 
 * @cython.final
 * cdef class PyCache:             # <<<<<<<<<<<<<<
 *     cdef Cache cache
 *     cdef readonly size_t sets
 */
struct __pyx_obj_10relstorage_5cache_5cache_PyCache {
  PyObject_HEAD
  struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyCache *__pyx_vtab;
  relstorage::cache::Cache cache;
  size_t sets;
  size_t hits;
  size_t misses;
};


/* "relstorage/cache/cache.pyx":279
 *         return not self.generation.empty()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         "Not thread safe."
 *         if self.generation.empty():
 */
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ {
  PyObject_HEAD
  relstorage::cache::Generation::iterator __pyx_v_it;
  struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self;
};


/* "relstorage/cache/cache.pyx":406
 *         self.cache.delitem(key)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across all the contained OID/TID pairs (the keys).
 */
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ {
  PyObject_HEAD
  relstorage::cache::Cache::iterator __pyx_v_end;
  relstorage::cache::Cache::iterator __pyx_v_it;
  relstorage::cache::OID_t __pyx_v_oid;
  struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self;
  PyObject *__pyx_v_t;
  PyObject *__pyx_v_tids;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "relstorage/cache/cache.pyx":428
 *             preincr(it)
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the oid/cache_value pairs.
 */
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems {
  PyObject_HEAD
  relstorage::cache::Cache::iterator __pyx_v_end;
  relstorage::cache::Cache::iterator __pyx_v_it;
  struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self;
};


/* "relstorage/cache/cache.pyx":441
 *             preincr(it)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the OIDs in the cache.
 */
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys {
  PyObject_HEAD
  relstorage::cache::Cache::iterator __pyx_v_end;
  relstorage::cache::Cache::iterator __pyx_v_it;
  struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self;
};


/* "relstorage/cache/cache.pyx":454
 *             preincr(it)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the values in the cache.
 */
struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values {
  PyObject_HEAD
  relstorage::cache::Cache::iterator __pyx_v_end;
  relstorage::cache::Cache::iterator __pyx_v_it;
  struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self;
};



/* "relstorage/cache/cache.pyx":112
 * @cython.final
 * @cython.internal
 * cdef class SingleValue(CachedValue):             # <<<<<<<<<<<<<<
 *     cdef SVCacheEntry_p entry
 * 
 */

struct __pyx_vtabstruct_10relstorage_5cache_5cache_SingleValue {
  struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *(*from_entry)(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p);
};
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_SingleValue *__pyx_vtabptr_10relstorage_5cache_5cache_SingleValue;
static struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_f_10relstorage_5cache_5cache_11SingleValue_from_entry(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p);


/* "relstorage/cache/cache.pyx":204
 * 
 * @cython.final
 * cdef class MultipleValues(CachedValue):             # <<<<<<<<<<<<<<
 *     cdef MVCacheEntry_p entry
 * 
 */

struct __pyx_vtabstruct_10relstorage_5cache_5cache_MultipleValues {
  struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *(*from_entry)(__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p);
};
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_MultipleValues *__pyx_vtabptr_10relstorage_5cache_5cache_MultipleValues;
static struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_f_10relstorage_5cache_5cache_14MultipleValues_from_entry(__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p);


/* "relstorage/cache/cache.pyx":245
 * 
 * 
 * cdef class PyGeneration:             # <<<<<<<<<<<<<<
 *     cdef Generation* generation
 *     cdef readonly object __name__
 */

struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyGeneration {
  PyObject *(*from_generation)(relstorage::cache::Generation &, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyGeneration *__pyx_vtabptr_10relstorage_5cache_5cache_PyGeneration;


/* "relstorage/cache/cache.pyx":289
 * 
 * @cython.final
 * cdef class PyCache:             # <<<<<<<<<<<<<<
 *     cdef Cache cache
 *     cdef readonly size_t sets
 */

struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyCache {
  PyObject *(*reset_stats)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, int __pyx_skip_dispatch);
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *(*get)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, int __pyx_skip_dispatch);
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *(*peek)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, int __pyx_skip_dispatch);
  PyObject *(*peek_item_with_tid)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, relstorage::cache::TID_t, int __pyx_skip_dispatch);
  int (*contains_oid_with_tid)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_item_with_tid)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_do_set)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, relstorage::cache::TID_t);
  PyObject *(*set_all_for_tid)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::TID_t, PyObject *, PyObject *, Py_ssize_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyCache *__pyx_vtabptr_10relstorage_5cache_5cache_PyCache;
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_reset_stats(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, int __pyx_skip_dispatch);
static struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_7PyCache_get(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, int __pyx_skip_dispatch);
static struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_7PyCache_peek(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, int __pyx_skip_dispatch);
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_peek_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, relstorage::cache::TID_t, int __pyx_skip_dispatch);
static int __pyx_f_10relstorage_5cache_5cache_7PyCache_contains_oid_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, int __pyx_skip_dispatch);
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_get_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, int __pyx_skip_dispatch);
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache__do_set(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, relstorage::cache::TID_t);
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_set_all_for_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::TID_t, PyObject *, PyObject *, Py_ssize_t, int __pyx_skip_dispatch);
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) (&args[nargs])
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, nargs, NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ReturnWithStopIteration.proto */
#define __Pyx_ReturnWithStopIteration(value)\
    if (value == Py_None) PyErr_SetNone(PyExc_StopIteration); else __Pyx__ReturnWithStopIteration(value)
static void __Pyx__ReturnWithStopIteration(PyObject* value);

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyObject_Str.proto */
#define __Pyx_PyObject_Str(obj)\
    (likely(PyString_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* MoveIfSupported.proto */
#if __cplusplus >= 201103L
  #include <utility>
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) std::move(x)
#else
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) x
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#if CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, CYTHON_UNUSED PyObject *typ) {
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, Py_ssize_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_CyFunctionType = 0;
#endif
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_IsCyOrPyCFunction(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func.vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_relstorage_3a__3a_cache_3a__3a_generation_num(relstorage::cache::generation_num value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_Occurred(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str);
#else
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);
#endif

/* #### Code section: module_declarations ### */
static struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_f_10relstorage_5cache_5cache_11SingleValue_from_entry(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p __pyx_v_entry); /* proto*/
static struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_f_10relstorage_5cache_5cache_14MultipleValues_from_entry(__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p __pyx_v_entry); /* proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_12PyGeneration_from_generation(relstorage::cache::Generation &__pyx_v_gen, PyObject *__pyx_v_name, PyObject *__pyx_v_cache); /* proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_reset_stats(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_7PyCache_get(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_7PyCache_peek(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_peek_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, relstorage::cache::TID_t __pyx_v_tid, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_10relstorage_5cache_5cache_7PyCache_contains_oid_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_get_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache__do_set(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_state, relstorage::cache::TID_t __pyx_v_tid); /* proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_set_all_for_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::TID_t __pyx_v_tid_int, PyObject *__pyx_v_state_oid_iter, PyObject *__pyx_v_compress, Py_ssize_t __pyx_v_value_limit, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/

/* Module declarations from "cython" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cpython.buffer" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libc.string" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libc.stdio" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "__builtin__" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cpython.type" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;
#endif

/* Module declarations from "cpython" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cpython.object" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "cpython.bytes" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libcpp.utility" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libcpp.pair" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libcpp.cast" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libcpp.string" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libcpp.vector" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "libcpp" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "relstorage.cache.c_cache" */
#if !CYTHON_COMPILING_IN_LIMITED_API
#endif

/* Module declarations from "relstorage.cache.cache" */
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_CachedValue = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_SingleValue = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_MultipleValues = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_PyGeneration = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_PyCache = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys = 0;
static PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values = 0;
#endif
static CYTHON_INLINE struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_python_from_entry_p(__pyx_t_10relstorage_5cache_5cache_ICacheEntry_p); /*proto*/
static CYTHON_INLINE struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_python_from_entry(relstorage::cache::ICacheEntry const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_10relstorage_5cache_5cache_bytes_from_pickle(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p const ); /*proto*/
static CYTHON_INLINE void __pyx_fuse_0__pyx_f_10relstorage_5cache_5cache_release_entry(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p *); /*proto*/
static CYTHON_INLINE void __pyx_fuse_1__pyx_f_10relstorage_5cache_5cache_release_entry(__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p *); /*proto*/
static PyObject *__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t(const std::vector<relstorage::cache::TID_t>  &); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "relstorage.cache.cache"
extern int __pyx_module_is_main_relstorage__cache__cache;
int __pyx_module_is_main_relstorage__cache__cache = 0;

/* Implementation of "relstorage.cache.cache" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_range;
/* #### Code section: string_decls ### */
static const char __pyx_k_i[] = "i";
static const char __pyx_k__2[] = "";
static const char __pyx_k__6[] = "*";
static const char __pyx_k_it[] = "it";
static const char __pyx_k__38[] = "?";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_oid[] = "oid";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_tid[] = "tid";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_eden[] = "eden";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_oids[] = "oids";
static const char __pyx_k_peek[] = "peek";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_filler[] = "filler";
static const char __pyx_k_freeze[] = "freeze";
static const char __pyx_k_frozen[] = "frozen";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_sizeof[] = "sizeof";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_PyCache[] = "PyCache";
static const char __pyx_k_tid_int[] = "tid_int";
static const char __pyx_k_add_MRUs[] = "add_MRUs";
static const char __pyx_k_compress[] = "compress";
static const char __pyx_k_del_oids[] = "del_oids";
static const char __pyx_k_delitems[] = "delitems";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_frequency[] = "frequency";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_oids_tids[] = "oids_tids";
static const char __pyx_k_probation[] = "probation";
static const char __pyx_k_protected[] = "protected";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_added_oids[] = "added_oids";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_CachedValue[] = "CachedValue";
static const char __pyx_k_PyCache_get[] = "PyCache.get";
static const char __pyx_k_reset_stats[] = "reset_stats";
static const char __pyx_k_value_limit[] = "value_limit";
static const char __pyx_k_PyCache_keys[] = "PyCache.keys";
static const char __pyx_k_PyCache_peek[] = "PyCache.peek";
static const char __pyx_k_PyGeneration[] = "PyGeneration";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_iteroiditems[] = "iteroiditems";
static const char __pyx_k_number_nodes[] = "number_nodes";
static const char __pyx_k_ordered_keys[] = "ordered_keys";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_MultipleValues[] = "MultipleValues";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_PyCache___iter[] = "PyCache.__iter__";
static const char __pyx_k_PyCache_freeze[] = "PyCache.freeze";
static const char __pyx_k_PyCache_values[] = "PyCache.values";
static const char __pyx_k_s_r_s_frozen_s[] = "%s(%r, %s, frozen=%s)";
static const char __pyx_k_state_oid_iter[] = "state_oid_iter";
static const char __pyx_k_age_frequencies[] = "age_frequencies";
static const char __pyx_k_set_all_for_tid[] = "set_all_for_tid";
static const char __pyx_k_NoSuchGeneration[] = "NoSuchGeneration";
static const char __pyx_k_PyCache_add_MRUs[] = "PyCache.add_MRUs";
static const char __pyx_k_PyCache_del_oids[] = "PyCache.del_oids";
static const char __pyx_k_PyCache_delitems[] = "PyCache.delitems";
static const char __pyx_k_PyCache_iteritems[] = "PyCache.iteritems";
static const char __pyx_k_get_item_with_tid[] = "get_item_with_tid";
static const char __pyx_k_return_count_only[] = "return_count_only";
static const char __pyx_k_SingleValue_sizeof[] = "SingleValue.sizeof";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_peek_item_with_tid[] = "peek_item_with_tid";
static const char __pyx_k_relstorage__compat[] = "relstorage._compat";
static const char __pyx_k_PyCache_reset_stats[] = "PyCache.reset_stats";
static const char __pyx_k_PyGeneration___iter[] = "PyGeneration.__iter__";
static const char __pyx_k_Invalid_pointer_cast[] = "Invalid pointer cast";
static const char __pyx_k_CacheConsistencyError[] = "CacheConsistencyError";
static const char __pyx_k_MultipleValues_sizeof[] = "MultipleValues.sizeof";
static const char __pyx_k_contains_oid_with_tid[] = "contains_oid_with_tid";
static const char __pyx_k_relstorage_cache_cache[] = "relstorage.cache.cache";
static const char __pyx_k_PyCache_age_frequencies[] = "PyCache.age_frequencies";
static const char __pyx_k_PyCache_set_all_for_tid[] = "PyCache.set_all_for_tid";
static const char __pyx_k_PyCache_get_item_with_tid[] = "PyCache.get_item_with_tid";
static const char __pyx_k_PyCache_peek_item_with_tid[] = "PyCache.peek_item_with_tid";
static const char __pyx_k_relstorage_cache_interfaces[] = "relstorage.cache.interfaces";
static const char __pyx_k_PyCache_contains_oid_with_tid[] = "PyCache.contains_oid_with_tid";
static const char __pyx_k_src_relstorage_cache_cache_pyx[] = "src/relstorage/cache/cache.pyx";
static const char __pyx_k_Python_wrappers_for_the_generat[] = "\nPython wrappers for the generational cache itself.\n\nThe cache controls object lifetime.\n\n";
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_n_s_CacheConsistencyError;
static PyObject *__pyx_n_s_CachedValue;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Invalid_pointer_cast;
static PyObject *__pyx_n_s_MultipleValues;
static PyObject *__pyx_n_s_MultipleValues_sizeof;
static PyObject *__pyx_n_s_NoSuchGeneration;
static PyObject *__pyx_n_s_NotImplemented;
static PyObject *__pyx_n_s_PyCache;
static PyObject *__pyx_n_s_PyCache___iter;
static PyObject *__pyx_n_s_PyCache_add_MRUs;
static PyObject *__pyx_n_s_PyCache_age_frequencies;
static PyObject *__pyx_n_s_PyCache_contains_oid_with_tid;
static PyObject *__pyx_n_s_PyCache_del_oids;
static PyObject *__pyx_n_s_PyCache_delitems;
static PyObject *__pyx_n_s_PyCache_freeze;
static PyObject *__pyx_n_s_PyCache_get;
static PyObject *__pyx_n_s_PyCache_get_item_with_tid;
static PyObject *__pyx_n_s_PyCache_iteritems;
static PyObject *__pyx_n_s_PyCache_keys;
static PyObject *__pyx_n_s_PyCache_peek;
static PyObject *__pyx_n_s_PyCache_peek_item_with_tid;
static PyObject *__pyx_n_s_PyCache_reset_stats;
static PyObject *__pyx_n_s_PyCache_set_all_for_tid;
static PyObject *__pyx_n_s_PyCache_values;
static PyObject *__pyx_n_s_PyGeneration;
static PyObject *__pyx_n_s_PyGeneration___iter;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_SingleValue_sizeof;
static PyObject *__pyx_kp_b__2;
static PyObject *__pyx_n_s__38;
static PyObject *__pyx_n_s__6;
static PyObject *__pyx_n_s_add_MRUs;
static PyObject *__pyx_n_s_added_oids;
static PyObject *__pyx_n_s_age_frequencies;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_compress;
static PyObject *__pyx_n_s_contains_oid_with_tid;
static PyObject *__pyx_n_s_del_oids;
static PyObject *__pyx_n_s_delitems;
static PyObject *__pyx_n_s_eden;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_filler;
static PyObject *__pyx_n_s_freeze;
static PyObject *__pyx_n_s_frequency;
static PyObject *__pyx_n_s_frozen;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_get_item_with_tid;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_initializing;
static PyObject *__pyx_n_s_it;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_n_s_iteritems;
static PyObject *__pyx_n_s_iteroiditems;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_number_nodes;
static PyObject *__pyx_n_s_oid;
static PyObject *__pyx_n_s_oids;
static PyObject *__pyx_n_s_oids_tids;
static PyObject *__pyx_n_s_ordered_keys;
static PyObject *__pyx_n_s_peek;
static PyObject *__pyx_n_s_peek_item_with_tid;
static PyObject *__pyx_n_s_probation;
static PyObject *__pyx_n_s_protected;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_relstorage__compat;
static PyObject *__pyx_n_s_relstorage_cache_cache;
static PyObject *__pyx_n_s_relstorage_cache_interfaces;
static PyObject *__pyx_n_s_reset_stats;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_return_count_only;
static PyObject *__pyx_kp_s_s_r_s_frozen_s;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_set_all_for_tid;
static PyObject *__pyx_n_s_sizeof;
static PyObject *__pyx_n_s_spec;
static PyObject *__pyx_kp_s_src_relstorage_cache_cache_pyx;
static PyObject *__pyx_n_s_state;
static PyObject *__pyx_n_s_state_oid_iter;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_tid;
static PyObject *__pyx_n_s_tid_int;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_value_limit;
static PyObject *__pyx_n_s_values;
#endif
/* #### Code section: decls ### */
static int __pyx_pf_10relstorage_5cache_5cache_11SingleValue___cinit__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static void __pyx_pf_10relstorage_5cache_5cache_11SingleValue_2__dealloc__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_4sizeof(CYTHON_UNUSED struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_6__iter__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_6frozen___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_5value___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_3key___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_9frequency___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_5state___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_3tid___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_7max_tid___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_12newest_value___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_8__eq__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_10__getitem__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_12__repr__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self); /* proto */
static void __pyx_pf_10relstorage_5cache_5cache_14MultipleValues___dealloc__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_2sizeof(CYTHON_UNUSED struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_5value___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_3key___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_9frequency___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_7max_tid___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_12newest_value___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_17generation_number___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_5limit___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_10relstorage_5cache_5cache_12PyGeneration___len__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self); /* proto */
static int __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_2__nonzero__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_4__iter__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_8__name_____get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self); /* proto */
static int __pyx_pf_10relstorage_5cache_5cache_7PyCache___cinit__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_eden, PyObject *__pyx_v_protected, PyObject *__pyx_v_probation); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_2reset_stats(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_5limit___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_4eden___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_9protected___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_9probation___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_11generations___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_4__nonzero__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_6__contains__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key); /* proto */
static Py_ssize_t __pyx_pf_10relstorage_5cache_5cache_7PyCache_8__len__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_10get(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_12peek(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_14peek_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, relstorage::cache::TID_t __pyx_v_tid); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_16contains_oid_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_18__getitem__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_20get_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid); /* proto */
static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_22__setitem__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_24__delitem__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_26__iter__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_29iteritems(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_32keys(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_35values(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_38set_all_for_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::TID_t __pyx_v_tid_int, PyObject *__pyx_v_state_oid_iter, PyObject *__pyx_v_compress, Py_ssize_t __pyx_v_value_limit); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_40add_MRUs(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_ordered_keys, PyObject *__pyx_v_return_count_only); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_42age_frequencies(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_44delitems(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_oids_tids); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_46del_oids(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_oids); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_48freeze(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_oids_tids); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_4sets___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_4hits___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_6misses___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self); /* proto */
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_CachedValue(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_SingleValue(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_MultipleValues(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_PyGeneration(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_PyCache(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_3_keys(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_4_values(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_int_0;
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_codeobj__3;
static PyObject *__pyx_codeobj__4;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__8;
static PyObject *__pyx_codeobj__10;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__37;
#endif
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_CachedValue;
  PyObject *__pyx_type_10relstorage_5cache_5cache_CachedValue;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_SingleValue;
  PyObject *__pyx_type_10relstorage_5cache_5cache_SingleValue;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_MultipleValues;
  PyObject *__pyx_type_10relstorage_5cache_5cache_MultipleValues;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_PyGeneration;
  PyObject *__pyx_type_10relstorage_5cache_5cache_PyGeneration;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache_PyCache;
  PyObject *__pyx_type_10relstorage_5cache_5cache_PyCache;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__;
  PyObject *__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__;
  PyObject *__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems;
  PyObject *__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys;
  PyObject *__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys;
  PyTypeObject *__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values;
  PyObject *__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_CacheConsistencyError;
  PyObject *__pyx_n_s_CachedValue;
  PyObject *__pyx_n_s_IndexError;
  PyObject *__pyx_kp_s_Invalid_pointer_cast;
  PyObject *__pyx_n_s_MultipleValues;
  PyObject *__pyx_n_s_MultipleValues_sizeof;
  PyObject *__pyx_n_s_NoSuchGeneration;
  PyObject *__pyx_n_s_NotImplemented;
  PyObject *__pyx_n_s_PyCache;
  PyObject *__pyx_n_s_PyCache___iter;
  PyObject *__pyx_n_s_PyCache_add_MRUs;
  PyObject *__pyx_n_s_PyCache_age_frequencies;
  PyObject *__pyx_n_s_PyCache_contains_oid_with_tid;
  PyObject *__pyx_n_s_PyCache_del_oids;
  PyObject *__pyx_n_s_PyCache_delitems;
  PyObject *__pyx_n_s_PyCache_freeze;
  PyObject *__pyx_n_s_PyCache_get;
  PyObject *__pyx_n_s_PyCache_get_item_with_tid;
  PyObject *__pyx_n_s_PyCache_iteritems;
  PyObject *__pyx_n_s_PyCache_keys;
  PyObject *__pyx_n_s_PyCache_peek;
  PyObject *__pyx_n_s_PyCache_peek_item_with_tid;
  PyObject *__pyx_n_s_PyCache_reset_stats;
  PyObject *__pyx_n_s_PyCache_set_all_for_tid;
  PyObject *__pyx_n_s_PyCache_values;
  PyObject *__pyx_n_s_PyGeneration;
  PyObject *__pyx_n_s_PyGeneration___iter;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_n_s_SingleValue_sizeof;
  PyObject *__pyx_kp_b__2;
  PyObject *__pyx_n_s__38;
  PyObject *__pyx_n_s__6;
  PyObject *__pyx_n_s_add_MRUs;
  PyObject *__pyx_n_s_added_oids;
  PyObject *__pyx_n_s_age_frequencies;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_class;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_compress;
  PyObject *__pyx_n_s_contains_oid_with_tid;
  PyObject *__pyx_n_s_del_oids;
  PyObject *__pyx_n_s_delitems;
  PyObject *__pyx_n_s_eden;
  PyObject *__pyx_n_s_end;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_filler;
  PyObject *__pyx_n_s_freeze;
  PyObject *__pyx_n_s_frequency;
  PyObject *__pyx_n_s_frozen;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_item_with_tid;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_it;
  PyObject *__pyx_n_s_iter;
  PyObject *__pyx_n_s_iteritems;
  PyObject *__pyx_n_s_iteroiditems;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_number_nodes;
  PyObject *__pyx_n_s_oid;
  PyObject *__pyx_n_s_oids;
  PyObject *__pyx_n_s_oids_tids;
  PyObject *__pyx_n_s_ordered_keys;
  PyObject *__pyx_n_s_peek;
  PyObject *__pyx_n_s_peek_item_with_tid;
  PyObject *__pyx_n_s_probation;
  PyObject *__pyx_n_s_protected;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_relstorage__compat;
  PyObject *__pyx_n_s_relstorage_cache_cache;
  PyObject *__pyx_n_s_relstorage_cache_interfaces;
  PyObject *__pyx_n_s_reset_stats;
  PyObject *__pyx_n_s_result;
  PyObject *__pyx_n_s_return_count_only;
  PyObject *__pyx_kp_s_s_r_s_frozen_s;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set_all_for_tid;
  PyObject *__pyx_n_s_sizeof;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_kp_s_src_relstorage_cache_cache_pyx;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_state_oid_iter;
  PyObject *__pyx_n_s_sys;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_tid;
  PyObject *__pyx_n_s_tid_int;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_value_limit;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_int_0;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__15;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__23;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__30;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_codeobj__3;
  PyObject *__pyx_codeobj__4;
  PyObject *__pyx_codeobj__5;
  PyObject *__pyx_codeobj__8;
  PyObject *__pyx_codeobj__10;
  PyObject *__pyx_codeobj__12;
  PyObject *__pyx_codeobj__14;
  PyObject *__pyx_codeobj__16;
  PyObject *__pyx_codeobj__18;
  PyObject *__pyx_codeobj__20;
  PyObject *__pyx_codeobj__22;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__37;
} __pyx_mstate;

#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache_CachedValue);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache_CachedValue);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache_SingleValue);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache_SingleValue);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache_MultipleValues);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache_MultipleValues);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache_PyGeneration);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache_PyGeneration);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache_PyCache);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache_PyCache);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys);
  Py_CLEAR(clear_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values);
  Py_CLEAR(clear_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CacheConsistencyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_CachedValue);
  Py_CLEAR(clear_module_state->__pyx_n_s_IndexError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Invalid_pointer_cast);
  Py_CLEAR(clear_module_state->__pyx_n_s_MultipleValues);
  Py_CLEAR(clear_module_state->__pyx_n_s_MultipleValues_sizeof);
  Py_CLEAR(clear_module_state->__pyx_n_s_NoSuchGeneration);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplemented);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache___iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_add_MRUs);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_age_frequencies);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_contains_oid_with_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_del_oids);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_delitems);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_freeze);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_get_item_with_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_iteritems);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_peek);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_peek_item_with_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_reset_stats);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_set_all_for_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyCache_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyGeneration);
  Py_CLEAR(clear_module_state->__pyx_n_s_PyGeneration___iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SingleValue_sizeof);
  Py_CLEAR(clear_module_state->__pyx_kp_b__2);
  Py_CLEAR(clear_module_state->__pyx_n_s__38);
  Py_CLEAR(clear_module_state->__pyx_n_s__6);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_MRUs);
  Py_CLEAR(clear_module_state->__pyx_n_s_added_oids);
  Py_CLEAR(clear_module_state->__pyx_n_s_age_frequencies);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_compress);
  Py_CLEAR(clear_module_state->__pyx_n_s_contains_oid_with_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_del_oids);
  Py_CLEAR(clear_module_state->__pyx_n_s_delitems);
  Py_CLEAR(clear_module_state->__pyx_n_s_eden);
  Py_CLEAR(clear_module_state->__pyx_n_s_end);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_filler);
  Py_CLEAR(clear_module_state->__pyx_n_s_freeze);
  Py_CLEAR(clear_module_state->__pyx_n_s_frequency);
  Py_CLEAR(clear_module_state->__pyx_n_s_frozen);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_item_with_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_it);
  Py_CLEAR(clear_module_state->__pyx_n_s_iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_iteritems);
  Py_CLEAR(clear_module_state->__pyx_n_s_iteroiditems);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_number_nodes);
  Py_CLEAR(clear_module_state->__pyx_n_s_oid);
  Py_CLEAR(clear_module_state->__pyx_n_s_oids);
  Py_CLEAR(clear_module_state->__pyx_n_s_oids_tids);
  Py_CLEAR(clear_module_state->__pyx_n_s_ordered_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_peek);
  Py_CLEAR(clear_module_state->__pyx_n_s_peek_item_with_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_probation);
  Py_CLEAR(clear_module_state->__pyx_n_s_protected);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_relstorage__compat);
  Py_CLEAR(clear_module_state->__pyx_n_s_relstorage_cache_cache);
  Py_CLEAR(clear_module_state->__pyx_n_s_relstorage_cache_interfaces);
  Py_CLEAR(clear_module_state->__pyx_n_s_reset_stats);
  Py_CLEAR(clear_module_state->__pyx_n_s_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_return_count_only);
  Py_CLEAR(clear_module_state->__pyx_kp_s_s_r_s_frozen_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_all_for_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_sizeof);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_relstorage_cache_cache_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_state_oid_iter);
  Py_CLEAR(clear_module_state->__pyx_n_s_sys);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_tid);
  Py_CLEAR(clear_module_state->__pyx_n_s_tid_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_value_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__15);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__23);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__30);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__3);
  Py_CLEAR(clear_module_state->__pyx_codeobj__4);
  Py_CLEAR(clear_module_state->__pyx_codeobj__5);
  Py_CLEAR(clear_module_state->__pyx_codeobj__8);
  Py_CLEAR(clear_module_state->__pyx_codeobj__10);
  Py_CLEAR(clear_module_state->__pyx_codeobj__12);
  Py_CLEAR(clear_module_state->__pyx_codeobj__14);
  Py_CLEAR(clear_module_state->__pyx_codeobj__16);
  Py_CLEAR(clear_module_state->__pyx_codeobj__18);
  Py_CLEAR(clear_module_state->__pyx_codeobj__20);
  Py_CLEAR(clear_module_state->__pyx_codeobj__22);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache_CachedValue);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache_CachedValue);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache_SingleValue);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache_SingleValue);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache_MultipleValues);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache_MultipleValues);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache_PyGeneration);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache_PyGeneration);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache_PyCache);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache_PyCache);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys);
  Py_VISIT(traverse_module_state->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values);
  Py_VISIT(traverse_module_state->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CacheConsistencyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_CachedValue);
  Py_VISIT(traverse_module_state->__pyx_n_s_IndexError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Invalid_pointer_cast);
  Py_VISIT(traverse_module_state->__pyx_n_s_MultipleValues);
  Py_VISIT(traverse_module_state->__pyx_n_s_MultipleValues_sizeof);
  Py_VISIT(traverse_module_state->__pyx_n_s_NoSuchGeneration);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplemented);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache___iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_add_MRUs);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_age_frequencies);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_contains_oid_with_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_del_oids);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_delitems);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_freeze);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_get_item_with_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_iteritems);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_peek);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_peek_item_with_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_reset_stats);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_set_all_for_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyCache_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyGeneration);
  Py_VISIT(traverse_module_state->__pyx_n_s_PyGeneration___iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SingleValue_sizeof);
  Py_VISIT(traverse_module_state->__pyx_kp_b__2);
  Py_VISIT(traverse_module_state->__pyx_n_s__38);
  Py_VISIT(traverse_module_state->__pyx_n_s__6);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_MRUs);
  Py_VISIT(traverse_module_state->__pyx_n_s_added_oids);
  Py_VISIT(traverse_module_state->__pyx_n_s_age_frequencies);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_compress);
  Py_VISIT(traverse_module_state->__pyx_n_s_contains_oid_with_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_del_oids);
  Py_VISIT(traverse_module_state->__pyx_n_s_delitems);
  Py_VISIT(traverse_module_state->__pyx_n_s_eden);
  Py_VISIT(traverse_module_state->__pyx_n_s_end);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_filler);
  Py_VISIT(traverse_module_state->__pyx_n_s_freeze);
  Py_VISIT(traverse_module_state->__pyx_n_s_frequency);
  Py_VISIT(traverse_module_state->__pyx_n_s_frozen);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_item_with_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_it);
  Py_VISIT(traverse_module_state->__pyx_n_s_iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_iteritems);
  Py_VISIT(traverse_module_state->__pyx_n_s_iteroiditems);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_number_nodes);
  Py_VISIT(traverse_module_state->__pyx_n_s_oid);
  Py_VISIT(traverse_module_state->__pyx_n_s_oids);
  Py_VISIT(traverse_module_state->__pyx_n_s_oids_tids);
  Py_VISIT(traverse_module_state->__pyx_n_s_ordered_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_peek);
  Py_VISIT(traverse_module_state->__pyx_n_s_peek_item_with_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_probation);
  Py_VISIT(traverse_module_state->__pyx_n_s_protected);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_relstorage__compat);
  Py_VISIT(traverse_module_state->__pyx_n_s_relstorage_cache_cache);
  Py_VISIT(traverse_module_state->__pyx_n_s_relstorage_cache_interfaces);
  Py_VISIT(traverse_module_state->__pyx_n_s_reset_stats);
  Py_VISIT(traverse_module_state->__pyx_n_s_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_return_count_only);
  Py_VISIT(traverse_module_state->__pyx_kp_s_s_r_s_frozen_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_all_for_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_sizeof);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_relstorage_cache_cache_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_state_oid_iter);
  Py_VISIT(traverse_module_state->__pyx_n_s_sys);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_tid);
  Py_VISIT(traverse_module_state->__pyx_n_s_tid_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_value_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__15);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__23);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__30);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__3);
  Py_VISIT(traverse_module_state->__pyx_codeobj__4);
  Py_VISIT(traverse_module_state->__pyx_codeobj__5);
  Py_VISIT(traverse_module_state->__pyx_codeobj__8);
  Py_VISIT(traverse_module_state->__pyx_codeobj__10);
  Py_VISIT(traverse_module_state->__pyx_codeobj__12);
  Py_VISIT(traverse_module_state->__pyx_codeobj__14);
  Py_VISIT(traverse_module_state->__pyx_codeobj__16);
  Py_VISIT(traverse_module_state->__pyx_codeobj__18);
  Py_VISIT(traverse_module_state->__pyx_codeobj__20);
  Py_VISIT(traverse_module_state->__pyx_codeobj__22);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#if CYTHON_COMPILING_IN_LIMITED_API
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#define __pyx_ptype_10relstorage_5cache_5cache_CachedValue __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache_CachedValue
#define __pyx_type_10relstorage_5cache_5cache_CachedValue __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache_CachedValue
#define __pyx_ptype_10relstorage_5cache_5cache_SingleValue __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache_SingleValue
#define __pyx_type_10relstorage_5cache_5cache_SingleValue __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache_SingleValue
#define __pyx_ptype_10relstorage_5cache_5cache_MultipleValues __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache_MultipleValues
#define __pyx_type_10relstorage_5cache_5cache_MultipleValues __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache_MultipleValues
#define __pyx_ptype_10relstorage_5cache_5cache_PyGeneration __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache_PyGeneration
#define __pyx_type_10relstorage_5cache_5cache_PyGeneration __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache_PyGeneration
#define __pyx_ptype_10relstorage_5cache_5cache_PyCache __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache_PyCache
#define __pyx_type_10relstorage_5cache_5cache_PyCache __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache_PyCache
#define __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__ __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__
#define __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__ __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__
#define __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__
#define __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__
#define __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems
#define __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems
#define __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys
#define __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys
#define __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values __pyx_mstate_global->__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values
#define __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values __pyx_mstate_global->__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_CacheConsistencyError __pyx_mstate_global->__pyx_n_s_CacheConsistencyError
#define __pyx_n_s_CachedValue __pyx_mstate_global->__pyx_n_s_CachedValue
#define __pyx_n_s_IndexError __pyx_mstate_global->__pyx_n_s_IndexError
#define __pyx_kp_s_Invalid_pointer_cast __pyx_mstate_global->__pyx_kp_s_Invalid_pointer_cast
#define __pyx_n_s_MultipleValues __pyx_mstate_global->__pyx_n_s_MultipleValues
#define __pyx_n_s_MultipleValues_sizeof __pyx_mstate_global->__pyx_n_s_MultipleValues_sizeof
#define __pyx_n_s_NoSuchGeneration __pyx_mstate_global->__pyx_n_s_NoSuchGeneration
#define __pyx_n_s_NotImplemented __pyx_mstate_global->__pyx_n_s_NotImplemented
#define __pyx_n_s_PyCache __pyx_mstate_global->__pyx_n_s_PyCache
#define __pyx_n_s_PyCache___iter __pyx_mstate_global->__pyx_n_s_PyCache___iter
#define __pyx_n_s_PyCache_add_MRUs __pyx_mstate_global->__pyx_n_s_PyCache_add_MRUs
#define __pyx_n_s_PyCache_age_frequencies __pyx_mstate_global->__pyx_n_s_PyCache_age_frequencies
#define __pyx_n_s_PyCache_contains_oid_with_tid __pyx_mstate_global->__pyx_n_s_PyCache_contains_oid_with_tid
#define __pyx_n_s_PyCache_del_oids __pyx_mstate_global->__pyx_n_s_PyCache_del_oids
#define __pyx_n_s_PyCache_delitems __pyx_mstate_global->__pyx_n_s_PyCache_delitems
#define __pyx_n_s_PyCache_freeze __pyx_mstate_global->__pyx_n_s_PyCache_freeze
#define __pyx_n_s_PyCache_get __pyx_mstate_global->__pyx_n_s_PyCache_get
#define __pyx_n_s_PyCache_get_item_with_tid __pyx_mstate_global->__pyx_n_s_PyCache_get_item_with_tid
#define __pyx_n_s_PyCache_iteritems __pyx_mstate_global->__pyx_n_s_PyCache_iteritems
#define __pyx_n_s_PyCache_keys __pyx_mstate_global->__pyx_n_s_PyCache_keys
#define __pyx_n_s_PyCache_peek __pyx_mstate_global->__pyx_n_s_PyCache_peek
#define __pyx_n_s_PyCache_peek_item_with_tid __pyx_mstate_global->__pyx_n_s_PyCache_peek_item_with_tid
#define __pyx_n_s_PyCache_reset_stats __pyx_mstate_global->__pyx_n_s_PyCache_reset_stats
#define __pyx_n_s_PyCache_set_all_for_tid __pyx_mstate_global->__pyx_n_s_PyCache_set_all_for_tid
#define __pyx_n_s_PyCache_values __pyx_mstate_global->__pyx_n_s_PyCache_values
#define __pyx_n_s_PyGeneration __pyx_mstate_global->__pyx_n_s_PyGeneration
#define __pyx_n_s_PyGeneration___iter __pyx_mstate_global->__pyx_n_s_PyGeneration___iter
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_n_s_SingleValue_sizeof __pyx_mstate_global->__pyx_n_s_SingleValue_sizeof
#define __pyx_kp_b__2 __pyx_mstate_global->__pyx_kp_b__2
#define __pyx_n_s__38 __pyx_mstate_global->__pyx_n_s__38
#define __pyx_n_s__6 __pyx_mstate_global->__pyx_n_s__6
#define __pyx_n_s_add_MRUs __pyx_mstate_global->__pyx_n_s_add_MRUs
#define __pyx_n_s_added_oids __pyx_mstate_global->__pyx_n_s_added_oids
#define __pyx_n_s_age_frequencies __pyx_mstate_global->__pyx_n_s_age_frequencies
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_class __pyx_mstate_global->__pyx_n_s_class
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_compress __pyx_mstate_global->__pyx_n_s_compress
#define __pyx_n_s_contains_oid_with_tid __pyx_mstate_global->__pyx_n_s_contains_oid_with_tid
#define __pyx_n_s_del_oids __pyx_mstate_global->__pyx_n_s_del_oids
#define __pyx_n_s_delitems __pyx_mstate_global->__pyx_n_s_delitems
#define __pyx_n_s_eden __pyx_mstate_global->__pyx_n_s_eden
#define __pyx_n_s_end __pyx_mstate_global->__pyx_n_s_end
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_filler __pyx_mstate_global->__pyx_n_s_filler
#define __pyx_n_s_freeze __pyx_mstate_global->__pyx_n_s_freeze
#define __pyx_n_s_frequency __pyx_mstate_global->__pyx_n_s_frequency
#define __pyx_n_s_frozen __pyx_mstate_global->__pyx_n_s_frozen
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_item_with_tid __pyx_mstate_global->__pyx_n_s_get_item_with_tid
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_it __pyx_mstate_global->__pyx_n_s_it
#define __pyx_n_s_iter __pyx_mstate_global->__pyx_n_s_iter
#define __pyx_n_s_iteritems __pyx_mstate_global->__pyx_n_s_iteritems
#define __pyx_n_s_iteroiditems __pyx_mstate_global->__pyx_n_s_iteroiditems
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_number_nodes __pyx_mstate_global->__pyx_n_s_number_nodes
#define __pyx_n_s_oid __pyx_mstate_global->__pyx_n_s_oid
#define __pyx_n_s_oids __pyx_mstate_global->__pyx_n_s_oids
#define __pyx_n_s_oids_tids __pyx_mstate_global->__pyx_n_s_oids_tids
#define __pyx_n_s_ordered_keys __pyx_mstate_global->__pyx_n_s_ordered_keys
#define __pyx_n_s_peek __pyx_mstate_global->__pyx_n_s_peek
#define __pyx_n_s_peek_item_with_tid __pyx_mstate_global->__pyx_n_s_peek_item_with_tid
#define __pyx_n_s_probation __pyx_mstate_global->__pyx_n_s_probation
#define __pyx_n_s_protected __pyx_mstate_global->__pyx_n_s_protected
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_relstorage__compat __pyx_mstate_global->__pyx_n_s_relstorage__compat
#define __pyx_n_s_relstorage_cache_cache __pyx_mstate_global->__pyx_n_s_relstorage_cache_cache
#define __pyx_n_s_relstorage_cache_interfaces __pyx_mstate_global->__pyx_n_s_relstorage_cache_interfaces
#define __pyx_n_s_reset_stats __pyx_mstate_global->__pyx_n_s_reset_stats
#define __pyx_n_s_result __pyx_mstate_global->__pyx_n_s_result
#define __pyx_n_s_return_count_only __pyx_mstate_global->__pyx_n_s_return_count_only
#define __pyx_kp_s_s_r_s_frozen_s __pyx_mstate_global->__pyx_kp_s_s_r_s_frozen_s
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set_all_for_tid __pyx_mstate_global->__pyx_n_s_set_all_for_tid
#define __pyx_n_s_sizeof __pyx_mstate_global->__pyx_n_s_sizeof
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_kp_s_src_relstorage_cache_cache_pyx __pyx_mstate_global->__pyx_kp_s_src_relstorage_cache_cache_pyx
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_state_oid_iter __pyx_mstate_global->__pyx_n_s_state_oid_iter
#define __pyx_n_s_sys __pyx_mstate_global->__pyx_n_s_sys
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_tid __pyx_mstate_global->__pyx_n_s_tid
#define __pyx_n_s_tid_int __pyx_mstate_global->__pyx_n_s_tid_int
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_value_limit __pyx_mstate_global->__pyx_n_s_value_limit
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__15 __pyx_mstate_global->__pyx_tuple__15
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__23 __pyx_mstate_global->__pyx_tuple__23
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__30 __pyx_mstate_global->__pyx_tuple__30
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_codeobj__3 __pyx_mstate_global->__pyx_codeobj__3
#define __pyx_codeobj__4 __pyx_mstate_global->__pyx_codeobj__4
#define __pyx_codeobj__5 __pyx_mstate_global->__pyx_codeobj__5
#define __pyx_codeobj__8 __pyx_mstate_global->__pyx_codeobj__8
#define __pyx_codeobj__10 __pyx_mstate_global->__pyx_codeobj__10
#define __pyx_codeobj__12 __pyx_mstate_global->__pyx_codeobj__12
#define __pyx_codeobj__14 __pyx_mstate_global->__pyx_codeobj__14
#define __pyx_codeobj__16 __pyx_mstate_global->__pyx_codeobj__16
#define __pyx_codeobj__18 __pyx_mstate_global->__pyx_codeobj__18
#define __pyx_codeobj__20 __pyx_mstate_global->__pyx_codeobj__20
#define __pyx_codeobj__22 __pyx_mstate_global->__pyx_codeobj__22
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#endif
/* #### Code section: module_code ### */

/* "relstorage/cache/cache.pyx":76
 *     """
 * 
 * cdef inline CachedValue python_from_entry_p(ICacheEntry_p entry):             # <<<<<<<<<<<<<<
 *     cdef const SVCacheEntry* sve_p = dynamic_cast[SVCacheEntry_p](entry)
 *     cdef const MVCacheEntry* mve_p = dynamic_cast[MVCacheEntry_p](entry)
 */

static CYTHON_INLINE struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_python_from_entry_p(__pyx_t_10relstorage_5cache_5cache_ICacheEntry_p __pyx_v_entry) {
  relstorage::cache::SVCacheEntry const *__pyx_v_sve_p;
  relstorage::cache::MVCacheEntry const *__pyx_v_mve_p;
  struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_sv = 0;
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p __pyx_t_1;
  __pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("python_from_entry_p", 0);

  /* "relstorage/cache/cache.pyx":77
 * 
 * cdef inline CachedValue python_from_entry_p(ICacheEntry_p entry):
 *     cdef const SVCacheEntry* sve_p = dynamic_cast[SVCacheEntry_p](entry)             # <<<<<<<<<<<<<<
 *     cdef const MVCacheEntry* mve_p = dynamic_cast[MVCacheEntry_p](entry)
 *     if not sve_p and not mve_p:
 */
  try {
    __pyx_t_1 = dynamic_cast<__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p>(__pyx_v_entry);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 77, __pyx_L1_error)
  }
  __pyx_v_sve_p = __pyx_t_1;

  /* "relstorage/cache/cache.pyx":78
 * cdef inline CachedValue python_from_entry_p(ICacheEntry_p entry):
 *     cdef const SVCacheEntry* sve_p = dynamic_cast[SVCacheEntry_p](entry)
 *     cdef const MVCacheEntry* mve_p = dynamic_cast[MVCacheEntry_p](entry)             # <<<<<<<<<<<<<<
 *     if not sve_p and not mve_p:
 *         # Most likely, entry is null
 */
  try {
    __pyx_t_2 = dynamic_cast<__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p>(__pyx_v_entry);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 78, __pyx_L1_error)
  }
  __pyx_v_mve_p = __pyx_t_2;

  /* "relstorage/cache/cache.pyx":79
 *     cdef const SVCacheEntry* sve_p = dynamic_cast[SVCacheEntry_p](entry)
 *     cdef const MVCacheEntry* mve_p = dynamic_cast[MVCacheEntry_p](entry)
 *     if not sve_p and not mve_p:             # <<<<<<<<<<<<<<
 *         # Most likely, entry is null
 *         raise AssertionError("Invalid pointer cast")
 */
  __pyx_t_4 = ((!(__pyx_v_sve_p != 0)) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_3 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = ((!(__pyx_v_mve_p != 0)) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_3)) {

    /* "relstorage/cache/cache.pyx":81
 *     if not sve_p and not mve_p:
 *         # Most likely, entry is null
 *         raise AssertionError("Invalid pointer cast")             # <<<<<<<<<<<<<<
 * 
 *     # Keeping in mind the semantics of our constructors
 */
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_AssertionError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 81, __pyx_L1_error)

    /* "relstorage/cache/cache.pyx":79
 *     cdef const SVCacheEntry* sve_p = dynamic_cast[SVCacheEntry_p](entry)
 *     cdef const MVCacheEntry* mve_p = dynamic_cast[MVCacheEntry_p](entry)
 *     if not sve_p and not mve_p:             # <<<<<<<<<<<<<<
 *         # Most likely, entry is null
 *         raise AssertionError("Invalid pointer cast")
 */
  }

  /* "relstorage/cache/cache.pyx":86
 *     # and Cython destructors, we'll let the class constructors handle that.
 *     cdef SingleValue sv
 *     if sve_p:             # <<<<<<<<<<<<<<
 *         sv = SingleValue.from_entry(sve_p)
 *         return sv
 */
  __pyx_t_3 = (__pyx_v_sve_p != 0);
  if (__pyx_t_3) {

    /* "relstorage/cache/cache.pyx":87
 *     cdef SingleValue sv
 *     if sve_p:
 *         sv = SingleValue.from_entry(sve_p)             # <<<<<<<<<<<<<<
 *         return sv
 * 
 */
    __pyx_t_5 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_11SingleValue_from_entry(__pyx_v_sve_p)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 87, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_v_sv = ((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "relstorage/cache/cache.pyx":88
 *     if sve_p:
 *         sv = SingleValue.from_entry(sve_p)
 *         return sv             # <<<<<<<<<<<<<<
 * 
 *     return MultipleValues.from_entry(mve_p)
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __Pyx_INCREF((PyObject *)__pyx_v_sv);
    __pyx_r = ((struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *)__pyx_v_sv);
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":86
 *     # and Cython destructors, we'll let the class constructors handle that.
 *     cdef SingleValue sv
 *     if sve_p:             # <<<<<<<<<<<<<<
 *         sv = SingleValue.from_entry(sve_p)
 *         return sv
 */
  }

  /* "relstorage/cache/cache.pyx":90
 *         return sv
 * 
 *     return MultipleValues.from_entry(mve_p)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_5 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_14MultipleValues_from_entry(__pyx_v_mve_p)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = ((struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":76
 *     """
 * 
 * cdef inline CachedValue python_from_entry_p(ICacheEntry_p entry):             # <<<<<<<<<<<<<<
 *     cdef const SVCacheEntry* sve_p = dynamic_cast[SVCacheEntry_p](entry)
 *     cdef const MVCacheEntry* mve_p = dynamic_cast[MVCacheEntry_p](entry)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("relstorage.cache.cache.python_from_entry_p", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sv);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":93
 * 
 * 
 * cdef inline CachedValue python_from_entry(const ICacheEntry& entry):             # <<<<<<<<<<<<<<
 *     return python_from_entry_p(&entry)
 * 
 */

static CYTHON_INLINE struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_python_from_entry(relstorage::cache::ICacheEntry const &__pyx_v_entry) {
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("python_from_entry", 0);

  /* "relstorage/cache/cache.pyx":94
 * 
 * cdef inline CachedValue python_from_entry(const ICacheEntry& entry):
 *     return python_from_entry_p(&entry)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object bytes_from_pickle(const SVCacheEntry_p entry):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_python_from_entry_p((&__pyx_v_entry))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":93
 * 
 * 
 * cdef inline CachedValue python_from_entry(const ICacheEntry& entry):             # <<<<<<<<<<<<<<
 *     return python_from_entry_p(&entry)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.python_from_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":96
 *     return python_from_entry_p(&entry)
 * 
 * cdef inline object bytes_from_pickle(const SVCacheEntry_p entry):             # <<<<<<<<<<<<<<
 *     return entry.as_object()
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_10relstorage_5cache_5cache_bytes_from_pickle(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p const __pyx_v_entry) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("bytes_from_pickle", 0);

  /* "relstorage/cache/cache.pyx":97
 * 
 * cdef inline object bytes_from_pickle(const SVCacheEntry_p entry):
 *     return entry.as_object()             # <<<<<<<<<<<<<<
 * 
 * ctypedef fused ConcreteCacheEntry:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_entry->as_object(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":96
 *     return python_from_entry_p(&entry)
 * 
 * cdef inline object bytes_from_pickle(const SVCacheEntry_p entry):             # <<<<<<<<<<<<<<
 *     return entry.as_object()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.bytes_from_pickle", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":103
 *     MVCacheEntry_p
 * 
 * cdef inline void release_entry(ConcreteCacheEntry* entry) except *:             # <<<<<<<<<<<<<<
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():
 */

static CYTHON_INLINE void __pyx_fuse_0__pyx_f_10relstorage_5cache_5cache_release_entry(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p *__pyx_v_entry) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  bool __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0release_entry", 0);

  /* "relstorage/cache/cache.pyx":105
 * cdef inline void release_entry(ConcreteCacheEntry* entry) except *:
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():             # <<<<<<<<<<<<<<
 *         del entry[0]
 *     entry[0] = NULL
 */
  __pyx_t_2 = ((__pyx_v_entry[0]) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  try {
    __pyx_t_3 = (__pyx_v_entry[0])->Py_release();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 105, __pyx_L1_error)
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":106
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():
 *         del entry[0]             # <<<<<<<<<<<<<<
 *     entry[0] = NULL
 * 
 */
    delete (__pyx_v_entry[0]);

    /* "relstorage/cache/cache.pyx":105
 * cdef inline void release_entry(ConcreteCacheEntry* entry) except *:
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():             # <<<<<<<<<<<<<<
 *         del entry[0]
 *     entry[0] = NULL
 */
  }

  /* "relstorage/cache/cache.pyx":107
 *     if entry[0] and entry[0].Py_release():
 *         del entry[0]
 *     entry[0] = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_entry[0]) = NULL;

  /* "relstorage/cache/cache.pyx":103
 *     MVCacheEntry_p
 * 
 * cdef inline void release_entry(ConcreteCacheEntry* entry) except *:             # <<<<<<<<<<<<<<
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.release_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

static CYTHON_INLINE void __pyx_fuse_1__pyx_f_10relstorage_5cache_5cache_release_entry(__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p *__pyx_v_entry) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  bool __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1release_entry", 0);

  /* "relstorage/cache/cache.pyx":105
 * cdef inline void release_entry(ConcreteCacheEntry* entry) except *:
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():             # <<<<<<<<<<<<<<
 *         del entry[0]
 *     entry[0] = NULL
 */
  __pyx_t_2 = ((__pyx_v_entry[0]) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  try {
    __pyx_t_3 = (__pyx_v_entry[0])->Py_release();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 105, __pyx_L1_error)
  }
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":106
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():
 *         del entry[0]             # <<<<<<<<<<<<<<
 *     entry[0] = NULL
 * 
 */
    delete (__pyx_v_entry[0]);

    /* "relstorage/cache/cache.pyx":105
 * cdef inline void release_entry(ConcreteCacheEntry* entry) except *:
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():             # <<<<<<<<<<<<<<
 *         del entry[0]
 *     entry[0] = NULL
 */
  }

  /* "relstorage/cache/cache.pyx":107
 *     if entry[0] and entry[0].Py_release():
 *         del entry[0]
 *     entry[0] = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  (__pyx_v_entry[0]) = NULL;

  /* "relstorage/cache/cache.pyx":103
 *     MVCacheEntry_p
 * 
 * cdef inline void release_entry(ConcreteCacheEntry* entry) except *:             # <<<<<<<<<<<<<<
 *     # Docs say to be very careful during __dealloc__
 *     if entry[0] and entry[0].Py_release():
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.release_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "relstorage/cache/cache.pyx":115
 *     cdef SVCacheEntry_p entry
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.entry = <SVCacheEntry_p>0
 * 
 */

/* Python wrapper */
static int __pyx_pw_10relstorage_5cache_5cache_11SingleValue_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10relstorage_5cache_5cache_11SingleValue_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue___cinit__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10relstorage_5cache_5cache_11SingleValue___cinit__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "relstorage/cache/cache.pyx":116
 * 
 *     def __cinit__(self):
 *         self.entry = <SVCacheEntry_p>0             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_v_self->entry = ((__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p)0);

  /* "relstorage/cache/cache.pyx":115
 *     cdef SVCacheEntry_p entry
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self.entry = <SVCacheEntry_p>0
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":119
 * 
 *     @staticmethod
 *     cdef SingleValue from_entry(SVCacheEntry_p entry):             # <<<<<<<<<<<<<<
 *         cdef SingleValue sv = SingleValue.__new__(SingleValue)
 *         sv.entry = entry.Py_use[SVCacheEntry]()
 */

static struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_f_10relstorage_5cache_5cache_11SingleValue_from_entry(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p __pyx_v_entry) {
  struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_sv = 0;
  struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_entry", 0);

  /* "relstorage/cache/cache.pyx":120
 *     @staticmethod
 *     cdef SingleValue from_entry(SVCacheEntry_p entry):
 *         cdef SingleValue sv = SingleValue.__new__(SingleValue)             # <<<<<<<<<<<<<<
 *         sv.entry = entry.Py_use[SVCacheEntry]()
 *         return sv
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10relstorage_5cache_5cache_SingleValue(((PyTypeObject *)__pyx_ptype_10relstorage_5cache_5cache_SingleValue), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_sv = ((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":121
 *     cdef SingleValue from_entry(SVCacheEntry_p entry):
 *         cdef SingleValue sv = SingleValue.__new__(SingleValue)
 *         sv.entry = entry.Py_use[SVCacheEntry]()             # <<<<<<<<<<<<<<
 *         return sv
 * 
 */
  __pyx_v_sv->entry = __pyx_v_entry->Py_use<relstorage::cache::SVCacheEntry>();

  /* "relstorage/cache/cache.pyx":122
 *         cdef SingleValue sv = SingleValue.__new__(SingleValue)
 *         sv.entry = entry.Py_use[SVCacheEntry]()
 *         return sv             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_sv);
  __pyx_r = __pyx_v_sv;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":119
 * 
 *     @staticmethod
 *     cdef SingleValue from_entry(SVCacheEntry_p entry):             # <<<<<<<<<<<<<<
 *         cdef SingleValue sv = SingleValue.__new__(SingleValue)
 *         sv.entry = entry.Py_use[SVCacheEntry]()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.from_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_sv);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":124
 *         return sv
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         release_entry(&self.entry)
 * 
 */

/* Python wrapper */
static void __pyx_pw_10relstorage_5cache_5cache_11SingleValue_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10relstorage_5cache_5cache_11SingleValue_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10relstorage_5cache_5cache_11SingleValue_2__dealloc__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10relstorage_5cache_5cache_11SingleValue_2__dealloc__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "relstorage/cache/cache.pyx":125
 * 
 *     def __dealloc__(self):
 *         release_entry(&self.entry)             # <<<<<<<<<<<<<<
 * 
 *     def sizeof(self):
 */
  __pyx_fuse_0__pyx_f_10relstorage_5cache_5cache_release_entry((&__pyx_v_self->entry)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 125, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":124
 *         return sv
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         release_entry(&self.entry)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("relstorage.cache.cache.SingleValue.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "relstorage/cache/cache.pyx":127
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 88
 *         return sizeof(SVCacheEntry)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5sizeof(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_11SingleValue_4sizeof, "SingleValue.sizeof(self)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_11SingleValue_5sizeof = {"sizeof", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5sizeof, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_11SingleValue_4sizeof};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5sizeof(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sizeof (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_4sizeof(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_4sizeof(CYTHON_UNUSED struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sizeof", 0);

  /* "relstorage/cache/cache.pyx":129
 *     def sizeof(self):
 *         # At this writing, reports 88
 *         return sizeof(SVCacheEntry)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(relstorage::cache::SVCacheEntry))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":127
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 88
 *         return sizeof(SVCacheEntry)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.sizeof", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":131
 *         return sizeof(SVCacheEntry)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter((
 *             bytes_from_pickle(self.entry),
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_7__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_7__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_6__iter__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_6__iter__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "relstorage/cache/cache.pyx":132
 * 
 *     def __iter__(self):
 *         return iter((             # <<<<<<<<<<<<<<
 *             bytes_from_pickle(self.entry),
 *             self.entry.tid()
 */
  __Pyx_XDECREF(__pyx_r);

  /* "relstorage/cache/cache.pyx":133
 *     def __iter__(self):
 *         return iter((
 *             bytes_from_pickle(self.entry),             # <<<<<<<<<<<<<<
 *             self.entry.tid()
 *         ))
 */
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_bytes_from_pickle(__pyx_v_self->entry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "relstorage/cache/cache.pyx":134
 *         return iter((
 *             bytes_from_pickle(self.entry),
 *             self.entry.tid()             # <<<<<<<<<<<<<<
 *         ))
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->entry->tid()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "relstorage/cache/cache.pyx":133
 *     def __iter__(self):
 *         return iter((
 *             bytes_from_pickle(self.entry),             # <<<<<<<<<<<<<<
 *             self.entry.tid()
 *         ))
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;

  /* "relstorage/cache/cache.pyx":132
 * 
 *     def __iter__(self):
 *         return iter((             # <<<<<<<<<<<<<<
 *             bytes_from_pickle(self.entry),
 *             self.entry.tid()
 */
  __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":131
 *         return sizeof(SVCacheEntry)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter((
 *             bytes_from_pickle(self.entry),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":137
 *         ))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def frozen(self):
 *         return self.entry.frozen()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_6frozen_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_6frozen_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_6frozen___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_6frozen___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":139
 *     @property
 *     def frozen(self):
 *         return self.entry.frozen()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->entry->frozen()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":137
 *         ))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def frozen(self):
 *         return self.entry.frozen()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.frozen.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":141
 *         return self.entry.frozen()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return self.state
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_5value___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_5value___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":143
 *     @property
 *     def value(self):
 *         return self.state             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":141
 *         return self.entry.frozen()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return self.state
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":145
 *         return self.state
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def key(self):
 *         return self.entry.key
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_3key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_3key___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_3key___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":147
 *     @property
 *     def key(self):
 *         return self.entry.key             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->entry->key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":145
 *         return self.state
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def key(self):
 *         return self.entry.key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":149
 *         return self.entry.key
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def frequency(self):
 *         return self.entry.frequency
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_9frequency_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_9frequency_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_9frequency___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_9frequency___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":151
 *     @property
 *     def frequency(self):
 *         return self.entry.frequency             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->entry->frequency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":149
 *         return self.entry.key
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def frequency(self):
 *         return self.entry.frequency
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.frequency.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":153
 *         return self.entry.frequency
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self):
 *         return bytes_from_pickle(self.entry)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5state_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5state_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_5state___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_5state___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":155
 *     @property
 *     def state(self):
 *         return bytes_from_pickle(self.entry)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_bytes_from_pickle(__pyx_v_self->entry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":153
 *         return self.entry.frequency
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def state(self):
 *         return bytes_from_pickle(self.entry)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.state.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":157
 *         return bytes_from_pickle(self.entry)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tid(self):
 *         return self.entry.tid()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_3tid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_3tid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_3tid___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_3tid___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":159
 *     @property
 *     def tid(self):
 *         return self.entry.tid()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->entry->tid()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":157
 *         return bytes_from_pickle(self.entry)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def tid(self):
 *         return self.entry.tid()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.tid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":161
 *         return self.entry.tid()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_tid(self):
 *         return self.entry.tid()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_7max_tid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_7max_tid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_7max_tid___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_7max_tid___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":163
 *     @property
 *     def max_tid(self):
 *         return self.entry.tid()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->entry->tid()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":161
 *         return self.entry.tid()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_tid(self):
 *         return self.entry.tid()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.max_tid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":165
 *         return self.entry.tid()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def newest_value(self):
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_12newest_value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_12newest_value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_12newest_value___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_12newest_value___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":167
 *     @property
 *     def newest_value(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":165
 *         return self.entry.tid()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def newest_value(self):
 *         return self
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":169
 *         return self
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.entry.weight()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_6weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_6weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_6weight___get__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":171
 *     @property
 *     def weight(self):
 *         return self.entry.weight()             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->entry->weight();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 171, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":169
 *         return self
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.entry.weight()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":173
 *         return self.entry.weight()
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef SingleValue p
 *         if other is self:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_9__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_9__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_8__eq__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_8__eq__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_p = 0;
  __pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p __pyx_v_my_entry;
  __pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p __pyx_v_other_entry;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "relstorage/cache/cache.pyx":175
 *     def __eq__(self, other):
 *         cdef SingleValue p
 *         if other is self:             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
  __pyx_t_1 = (__pyx_v_other == ((PyObject *)__pyx_v_self));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "relstorage/cache/cache.pyx":176
 *         cdef SingleValue p
 *         if other is self:
 *             return True             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, SingleValue):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":175
 *     def __eq__(self, other):
 *         cdef SingleValue p
 *         if other is self:             # <<<<<<<<<<<<<<
 *             return True
 * 
 */
  }

  /* "relstorage/cache/cache.pyx":178
 *             return True
 * 
 *         if isinstance(other, SingleValue):             # <<<<<<<<<<<<<<
 *             p = <SingleValue>other
 *             my_entry = self.entry
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_10relstorage_5cache_5cache_SingleValue); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":179
 * 
 *         if isinstance(other, SingleValue):
 *             p = <SingleValue>other             # <<<<<<<<<<<<<<
 *             my_entry = self.entry
 *             other_entry = p.entry
 */
    __pyx_t_3 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_p = ((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "relstorage/cache/cache.pyx":180
 *         if isinstance(other, SingleValue):
 *             p = <SingleValue>other
 *             my_entry = self.entry             # <<<<<<<<<<<<<<
 *             other_entry = p.entry
 *             return my_entry == other_entry
 */
    __pyx_t_4 = __pyx_v_self->entry;
    __pyx_v_my_entry = __pyx_t_4;

    /* "relstorage/cache/cache.pyx":181
 *             p = <SingleValue>other
 *             my_entry = self.entry
 *             other_entry = p.entry             # <<<<<<<<<<<<<<
 *             return my_entry == other_entry
 * 
 */
    __pyx_t_4 = __pyx_v_p->entry;
    __pyx_v_other_entry = __pyx_t_4;

    /* "relstorage/cache/cache.pyx":182
 *             my_entry = self.entry
 *             other_entry = p.entry
 *             return my_entry == other_entry             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(other, tuple):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_my_entry == __pyx_v_other_entry)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":178
 *             return True
 * 
 *         if isinstance(other, SingleValue):             # <<<<<<<<<<<<<<
 *             p = <SingleValue>other
 *             my_entry = self.entry
 */
  }

  /* "relstorage/cache/cache.pyx":184
 *             return my_entry == other_entry
 * 
 *         if isinstance(other, tuple):             # <<<<<<<<<<<<<<
 *             return len(other) == 2 and self.tid == other[1] and self.value == other[0]
 *         return NotImplemented
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_other); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "relstorage/cache/cache.pyx":185
 * 
 *         if isinstance(other, tuple):
 *             return len(other) == 2 and self.tid == other[1] and self.value == other[0]             # <<<<<<<<<<<<<<
 *         return NotImplemented
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 185, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_5 == 2);
    if (__pyx_t_2) {
    } else {
      __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_3 = __pyx_t_6;
      __pyx_t_6 = 0;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tid); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_other, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 185, __pyx_L1_error)
    if (__pyx_t_2) {
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __Pyx_INCREF(__pyx_t_8);
      __pyx_t_3 = __pyx_t_8;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_value); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_other, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_8, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_3 = __pyx_t_6;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_L6_bool_binop_done:;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":184
 *             return my_entry == other_entry
 * 
 *         if isinstance(other, tuple):             # <<<<<<<<<<<<<<
 *             return len(other) == 2 and self.tid == other[1] and self.value == other[0]
 *         return NotImplemented
 */
  }

  /* "relstorage/cache/cache.pyx":186
 *         if isinstance(other, tuple):
 *             return len(other) == 2 and self.tid == other[1] and self.value == other[0]
 *         return NotImplemented             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_builtin_NotImplemented);
  __pyx_r = __pyx_builtin_NotImplemented;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":173
 *         return self.entry.weight()
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef SingleValue p
 *         if other is self:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_p);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":188
 *         return NotImplemented
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             return self.state()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_11__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 188, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_10__getitem__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_10__getitem__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "relstorage/cache/cache.pyx":189
 * 
 *     def __getitem__(self, int i):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             return self.state()
 *         if i == 1:
 */
  __pyx_t_1 = ((__pyx_v_i == 0) != 0);
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":190
 *     def __getitem__(self, int i):
 *         if i == 0:
 *             return self.state()             # <<<<<<<<<<<<<<
 *         if i == 1:
 *             return self.entry.tid()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_state); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    {
      PyObject *__pyx_callargs[1] = {__pyx_t_4, };
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":189
 * 
 *     def __getitem__(self, int i):
 *         if i == 0:             # <<<<<<<<<<<<<<
 *             return self.state()
 *         if i == 1:
 */
  }

  /* "relstorage/cache/cache.pyx":191
 *         if i == 0:
 *             return self.state()
 *         if i == 1:             # <<<<<<<<<<<<<<
 *             return self.entry.tid()
 *         raise IndexError
 */
  __pyx_t_1 = ((__pyx_v_i == 1) != 0);
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":192
 *             return self.state()
 *         if i == 1:
 *             return self.entry.tid()             # <<<<<<<<<<<<<<
 *         raise IndexError
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int64_t(__pyx_v_self->entry->tid()); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":191
 *         if i == 0:
 *             return self.state()
 *         if i == 1:             # <<<<<<<<<<<<<<
 *             return self.entry.tid()
 *         raise IndexError
 */
  }

  /* "relstorage/cache/cache.pyx":193
 *         if i == 1:
 *             return self.entry.tid()
 *         raise IndexError             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_Raise(__pyx_builtin_IndexError, 0, 0, 0);
  __PYX_ERR(0, 193, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":188
 *         return NotImplemented
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i == 0:
 *             return self.state()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":195
 *         raise IndexError
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%r, %s, frozen=%s)" % (
 *             self.__class__.__name__,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_13__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_11SingleValue_13__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_11SingleValue_12__repr__(((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_11SingleValue_12__repr__(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "relstorage/cache/cache.pyx":196
 * 
 *     def __repr__(self):
 *         return "%s(%r, %s, frozen=%s)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.state,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "relstorage/cache/cache.pyx":197
 *     def __repr__(self):
 *         return "%s(%r, %s, frozen=%s)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.state,
 *             self.tid,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":198
 *         return "%s(%r, %s, frozen=%s)" % (
 *             self.__class__.__name__,
 *             self.state,             # <<<<<<<<<<<<<<
 *             self.tid,
 *             self.frozen,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_state); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "relstorage/cache/cache.pyx":199
 *             self.__class__.__name__,
 *             self.state,
 *             self.tid,             # <<<<<<<<<<<<<<
 *             self.frozen,
 *         )
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_tid); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "relstorage/cache/cache.pyx":200
 *             self.state,
 *             self.tid,
 *             self.frozen,             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_frozen); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "relstorage/cache/cache.pyx":197
 *     def __repr__(self):
 *         return "%s(%r, %s, frozen=%s)" % (
 *             self.__class__.__name__,             # <<<<<<<<<<<<<<
 *             self.state,
 *             self.tid,
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
  __pyx_t_2 = 0;
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "relstorage/cache/cache.pyx":196
 * 
 *     def __repr__(self):
 *         return "%s(%r, %s, frozen=%s)" % (             # <<<<<<<<<<<<<<
 *             self.__class__.__name__,
 *             self.state,
 */
  __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_s_r_s_frozen_s, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":195
 *         raise IndexError
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(%r, %s, frozen=%s)" % (
 *             self.__class__.__name__,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("relstorage.cache.cache.SingleValue.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":208
 * 
 *     @staticmethod
 *     cdef MultipleValues from_entry(MVCacheEntry_p entry):             # <<<<<<<<<<<<<<
 *         cdef MultipleValues mv = MultipleValues.__new__(MultipleValues)
 *         mv.entry = entry.Py_use[MVCacheEntry]()
 */

static struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_f_10relstorage_5cache_5cache_14MultipleValues_from_entry(__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p __pyx_v_entry) {
  struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_mv = 0;
  struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_entry", 0);

  /* "relstorage/cache/cache.pyx":209
 *     @staticmethod
 *     cdef MultipleValues from_entry(MVCacheEntry_p entry):
 *         cdef MultipleValues mv = MultipleValues.__new__(MultipleValues)             # <<<<<<<<<<<<<<
 *         mv.entry = entry.Py_use[MVCacheEntry]()
 *         return mv
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10relstorage_5cache_5cache_MultipleValues(((PyTypeObject *)__pyx_ptype_10relstorage_5cache_5cache_MultipleValues), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_mv = ((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":210
 *     cdef MultipleValues from_entry(MVCacheEntry_p entry):
 *         cdef MultipleValues mv = MultipleValues.__new__(MultipleValues)
 *         mv.entry = entry.Py_use[MVCacheEntry]()             # <<<<<<<<<<<<<<
 *         return mv
 * 
 */
  __pyx_v_mv->entry = __pyx_v_entry->Py_use<relstorage::cache::MVCacheEntry>();

  /* "relstorage/cache/cache.pyx":211
 *         cdef MultipleValues mv = MultipleValues.__new__(MultipleValues)
 *         mv.entry = entry.Py_use[MVCacheEntry]()
 *         return mv             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_mv);
  __pyx_r = __pyx_v_mv;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":208
 * 
 *     @staticmethod
 *     cdef MultipleValues from_entry(MVCacheEntry_p entry):             # <<<<<<<<<<<<<<
 *         cdef MultipleValues mv = MultipleValues.__new__(MultipleValues)
 *         mv.entry = entry.Py_use[MVCacheEntry]()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.from_entry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mv);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":213
 *         return mv
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         release_entry(&self.entry)
 * 
 */

/* Python wrapper */
static void __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_1__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_1__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_10relstorage_5cache_5cache_14MultipleValues___dealloc__(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_10relstorage_5cache_5cache_14MultipleValues___dealloc__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "relstorage/cache/cache.pyx":214
 * 
 *     def __dealloc__(self):
 *         release_entry(&self.entry)             # <<<<<<<<<<<<<<
 * 
 *     def sizeof(self):
 */
  __pyx_fuse_1__pyx_f_10relstorage_5cache_5cache_release_entry((&__pyx_v_self->entry)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":213
 *         return mv
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         release_entry(&self.entry)
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("relstorage.cache.cache.MultipleValues.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "relstorage/cache/cache.pyx":216
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 72.
 *         return sizeof(MVCacheEntry)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_3sizeof(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_14MultipleValues_2sizeof, "MultipleValues.sizeof(self)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_14MultipleValues_3sizeof = {"sizeof", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_3sizeof, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_14MultipleValues_2sizeof};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_3sizeof(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sizeof (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_14MultipleValues_2sizeof(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_2sizeof(CYTHON_UNUSED struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sizeof", 0);

  /* "relstorage/cache/cache.pyx":218
 *     def sizeof(self):
 *         # At this writing, reports 72.
 *         return sizeof(MVCacheEntry)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(relstorage::cache::MVCacheEntry))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":216
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 72.
 *         return sizeof(MVCacheEntry)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.sizeof", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":220
 *         return sizeof(MVCacheEntry)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return list(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_5value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_14MultipleValues_5value___get__(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_5value___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":222
 *     @property
 *     def value(self):
 *         return list(self)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":220
 *         return sizeof(MVCacheEntry)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def value(self):
 *         return list(self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":224
 *         return list(self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def key(self):
 *         return self.entry.key
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_3key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_3key_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_14MultipleValues_3key___get__(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_3key___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":226
 *     @property
 *     def key(self):
 *         return self.entry.key             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->entry->key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":224
 *         return list(self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def key(self):
 *         return self.entry.key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.key.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":228
 *         return self.entry.key
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def frequency(self):
 *         return self.entry.frequency
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_9frequency_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_9frequency_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_14MultipleValues_9frequency___get__(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_9frequency___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":230
 *     @property
 *     def frequency(self):
 *         return self.entry.frequency             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->entry->frequency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":228
 *         return self.entry.key
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def frequency(self):
 *         return self.entry.frequency
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.frequency.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":232
 *         return self.entry.frequency
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.entry.weight()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_6weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_6weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_14MultipleValues_6weight___get__(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":234
 *     @property
 *     def weight(self):
 *         return self.entry.weight()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->entry->weight();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 234, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":232
 *         return self.entry.frequency
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.entry.weight()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":236
 *         return self.entry.weight()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_tid(self):
 *         return self.entry.newest_tid()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_7max_tid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_7max_tid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_14MultipleValues_7max_tid___get__(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_7max_tid___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":238
 *     @property
 *     def max_tid(self):
 *         return self.entry.newest_tid()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->entry->newest_tid()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":236
 *         return self.entry.weight()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_tid(self):
 *         return self.entry.newest_tid()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.max_tid.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":240
 *         return self.entry.newest_tid()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def newest_value(self):
 *         return python_from_entry_p(self.entry.copy_newest_entry());
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_12newest_value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_12newest_value_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_14MultipleValues_12newest_value___get__(((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_14MultipleValues_12newest_value___get__(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  relstorage::cache::SVCacheEntry *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":242
 *     @property
 *     def newest_value(self):
 *         return python_from_entry_p(self.entry.copy_newest_entry());             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  try {
    __pyx_t_1 = __pyx_v_self->entry->copy_newest_entry();
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 242, __pyx_L1_error)
  }
  __pyx_t_2 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_python_from_entry_p(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":240
 *         return self.entry.newest_tid()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def newest_value(self):
 *         return python_from_entry_p(self.entry.copy_newest_entry());
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("relstorage.cache.cache.MultipleValues.newest_value.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":253
 * 
 *     @staticmethod
 *     cdef from_generation(Generation& gen, name, cache):             # <<<<<<<<<<<<<<
 *         cdef PyGeneration pygen = PyGeneration.__new__(PyGeneration)
 *         pygen.generation = &gen
 */

static PyObject *__pyx_f_10relstorage_5cache_5cache_12PyGeneration_from_generation(relstorage::cache::Generation &__pyx_v_gen, PyObject *__pyx_v_name, PyObject *__pyx_v_cache) {
  struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_pygen = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_generation", 0);

  /* "relstorage/cache/cache.pyx":254
 *     @staticmethod
 *     cdef from_generation(Generation& gen, name, cache):
 *         cdef PyGeneration pygen = PyGeneration.__new__(PyGeneration)             # <<<<<<<<<<<<<<
 *         pygen.generation = &gen
 *         pygen.__name__ = name
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_10relstorage_5cache_5cache_PyGeneration(((PyTypeObject *)__pyx_ptype_10relstorage_5cache_5cache_PyGeneration), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_pygen = ((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":255
 *     cdef from_generation(Generation& gen, name, cache):
 *         cdef PyGeneration pygen = PyGeneration.__new__(PyGeneration)
 *         pygen.generation = &gen             # <<<<<<<<<<<<<<
 *         pygen.__name__ = name
 *         pygen._cache = cache
 */
  __pyx_v_pygen->generation = (&__pyx_v_gen);

  /* "relstorage/cache/cache.pyx":256
 *         cdef PyGeneration pygen = PyGeneration.__new__(PyGeneration)
 *         pygen.generation = &gen
 *         pygen.__name__ = name             # <<<<<<<<<<<<<<
 *         pygen._cache = cache
 *         return pygen
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_pygen->__name__);
  __Pyx_DECREF(__pyx_v_pygen->__name__);
  __pyx_v_pygen->__name__ = __pyx_v_name;

  /* "relstorage/cache/cache.pyx":257
 *         pygen.generation = &gen
 *         pygen.__name__ = name
 *         pygen._cache = cache             # <<<<<<<<<<<<<<
 *         return pygen
 * 
 */
  if (!(likely(((__pyx_v_cache) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_cache, __pyx_ptype_10relstorage_5cache_5cache_PyCache))))) __PYX_ERR(0, 257, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_cache;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_pygen->_cache);
  __Pyx_DECREF((PyObject *)__pyx_v_pygen->_cache);
  __pyx_v_pygen->_cache = ((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":258
 *         pygen.__name__ = name
 *         pygen._cache = cache
 *         return pygen             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_pygen);
  __pyx_r = ((PyObject *)__pyx_v_pygen);
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":253
 * 
 *     @staticmethod
 *     cdef from_generation(Generation& gen, name, cache):             # <<<<<<<<<<<<<<
 *         cdef PyGeneration pygen = PyGeneration.__new__(PyGeneration)
 *         pygen.generation = &gen
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyGeneration.from_generation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_pygen);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":260
 *         return pygen
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def generation_number(self):
 *         return self.generation.generation
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_17generation_number_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_17generation_number_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_17generation_number___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_17generation_number___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":262
 *     @property
 *     def generation_number(self):
 *         return self.generation.generation             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_relstorage_3a__3a_cache_3a__3a_generation_num(__pyx_v_self->generation->generation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":260
 *         return pygen
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def generation_number(self):
 *         return self.generation.generation
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyGeneration.generation_number.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":264
 *         return self.generation.generation
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def limit(self):
 *         return self.generation.max_weight()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_5limit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_5limit_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_5limit___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_5limit___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":266
 *     @property
 *     def limit(self):
 *         return self.generation.max_weight()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->generation->max_weight()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":264
 *         return self.generation.generation
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def limit(self):
 *         return self.generation.max_weight()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyGeneration.limit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":268
 *         return self.generation.max_weight()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.generation.sum_weights()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_6weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_6weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_6weight___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":270
 *     @property
 *     def weight(self):
 *         return self.generation.sum_weights()             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->generation->sum_weights()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 270, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":268
 *         return self.generation.max_weight()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.generation.sum_weights()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyGeneration.weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":272
 *         return self.generation.sum_weights()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.generation.size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_1__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_1__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_12PyGeneration___len__(((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10relstorage_5cache_5cache_12PyGeneration___len__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "relstorage/cache/cache.pyx":273
 * 
 *     def __len__(self):
 *         return self.generation.size()             # <<<<<<<<<<<<<<
 * 
 *     # Cython still uses __nonzero__
 */
  __pyx_r = __pyx_v_self->generation->size();
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":272
 *         return self.generation.sum_weights()
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.generation.size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":276
 * 
 *     # Cython still uses __nonzero__
 *     def __nonzero__(self):             # <<<<<<<<<<<<<<
 *         return not self.generation.empty()
 * 
 */

/* Python wrapper */
static int __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_3__nonzero__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_3__nonzero__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__nonzero__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_2__nonzero__(((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_2__nonzero__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__nonzero__", 0);

  /* "relstorage/cache/cache.pyx":277
 *     # Cython still uses __nonzero__
 *     def __nonzero__(self):
 *         return not self.generation.empty()             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_r = (!(__pyx_v_self->generation->empty() != 0));
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":276
 * 
 *     # Cython still uses __nonzero__
 *     def __nonzero__(self):             # <<<<<<<<<<<<<<
 *         return not self.generation.empty()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10relstorage_5cache_5cache_12PyGeneration_6generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "relstorage/cache/cache.pyx":279
 *         return not self.generation.empty()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         "Not thread safe."
 *         if self.generation.empty():
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_5__iter__(PyObject *__pyx_v_self); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_12PyGeneration_4__iter__, "Not thread safe.");
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_10relstorage_5cache_5cache_12PyGeneration_4__iter__;
#endif
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_5__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_4__iter__(((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_4__iter__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct____iter__(__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 279, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10relstorage_5cache_5cache_12PyGeneration_6generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_PyGeneration___iter, __pyx_n_s_relstorage_cache_cache); if (unlikely(!gen)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyGeneration.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10relstorage_5cache_5cache_12PyGeneration_6generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 279, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":281
 *     def __iter__(self):
 *         "Not thread safe."
 *         if self.generation.empty():             # <<<<<<<<<<<<<<
 *             return ()
 *         it = self.generation.begin()
 */
  __pyx_t_1 = (__pyx_cur_scope->__pyx_v_self->generation->empty() != 0);
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":282
 *         "Not thread safe."
 *         if self.generation.empty():
 *             return ()             # <<<<<<<<<<<<<<
 *         it = self.generation.begin()
 *         while it != self.generation.end():
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_empty_tuple);
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":281
 *     def __iter__(self):
 *         "Not thread safe."
 *         if self.generation.empty():             # <<<<<<<<<<<<<<
 *             return ()
 *         it = self.generation.begin()
 */
  }

  /* "relstorage/cache/cache.pyx":283
 *         if self.generation.empty():
 *             return ()
 *         it = self.generation.begin()             # <<<<<<<<<<<<<<
 *         while it != self.generation.end():
 *             yield python_from_entry(deref(it))
 */
  __pyx_cur_scope->__pyx_v_it = __pyx_cur_scope->__pyx_v_self->generation->begin();

  /* "relstorage/cache/cache.pyx":284
 *             return ()
 *         it = self.generation.begin()
 *         while it != self.generation.end():             # <<<<<<<<<<<<<<
 *             yield python_from_entry(deref(it))
 *             preincr(it)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_it != __pyx_cur_scope->__pyx_v_self->generation->end()) != 0);
    if (!__pyx_t_1) break;

    /* "relstorage/cache/cache.pyx":285
 *         it = self.generation.begin()
 *         while it != self.generation.end():
 *             yield python_from_entry(deref(it))             # <<<<<<<<<<<<<<
 *             preincr(it)
 * 
 */
    __pyx_t_2 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_python_from_entry((*__pyx_cur_scope->__pyx_v_it))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 285, __pyx_L1_error)

    /* "relstorage/cache/cache.pyx":286
 *         while it != self.generation.end():
 *             yield python_from_entry(deref(it))
 *             preincr(it)             # <<<<<<<<<<<<<<
 * 
 * @cython.final
 */
    (void)((++__pyx_cur_scope->__pyx_v_it));
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "relstorage/cache/cache.pyx":279
 *         return not self.generation.empty()
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         "Not thread safe."
 *         if self.generation.empty():
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":247
 * cdef class PyGeneration:
 *     cdef Generation* generation
 *     cdef readonly object __name__             # <<<<<<<<<<<<<<
 *     # This reference is to keep the cache object alive while we're
 *     # alive since it owns the Generation*.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_8__name___1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_8__name___1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_12PyGeneration_8__name_____get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_12PyGeneration_8__name_____get__(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->__name__);
  __pyx_r = __pyx_v_self->__name__;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":295
 *     cdef readonly size_t misses
 * 
 *     def __cinit__(self, eden, protected, probation):             # <<<<<<<<<<<<<<
 *         self.cache.resize(eden, protected, probation)
 *         self.sets = self.hits = self.misses = 0
 */

/* Python wrapper */
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_eden = 0;
  PyObject *__pyx_v_protected = 0;
  PyObject *__pyx_v_probation = 0;
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_eden,&__pyx_n_s_protected,&__pyx_n_s_probation,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_eden,&__pyx_n_s_protected,&__pyx_n_s_probation,0};
    #endif
    PyObject* values[3] = {0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_eden)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_protected)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(0, 295, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_probation)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(0, 295, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 295, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
    }
    __pyx_v_eden = values[0];
    __pyx_v_protected = values[1];
    __pyx_v_probation = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 295, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache___cinit__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), __pyx_v_eden, __pyx_v_protected, __pyx_v_probation);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10relstorage_5cache_5cache_7PyCache___cinit__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_eden, PyObject *__pyx_v_protected, PyObject *__pyx_v_probation) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "relstorage/cache/cache.pyx":296
 * 
 *     def __cinit__(self, eden, protected, probation):
 *         self.cache.resize(eden, protected, probation)             # <<<<<<<<<<<<<<
 *         self.sets = self.hits = self.misses = 0
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_eden); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_As_size_t(__pyx_v_protected); if (unlikely((__pyx_t_2 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_v_probation); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_v_self->cache.resize(__pyx_t_1, __pyx_t_2, __pyx_t_3);

  /* "relstorage/cache/cache.pyx":297
 *     def __cinit__(self, eden, protected, probation):
 *         self.cache.resize(eden, protected, probation)
 *         self.sets = self.hits = self.misses = 0             # <<<<<<<<<<<<<<
 * 
 *     cpdef reset_stats(self):
 */
  __pyx_v_self->sets = 0;
  __pyx_v_self->hits = 0;
  __pyx_v_self->misses = 0;

  /* "relstorage/cache/cache.pyx":295
 *     cdef readonly size_t misses
 * 
 *     def __cinit__(self, eden, protected, probation):             # <<<<<<<<<<<<<<
 *         self.cache.resize(eden, protected, probation)
 *         self.sets = self.hits = self.misses = 0
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":299
 *         self.sets = self.hits = self.misses = 0
 * 
 *     cpdef reset_stats(self):             # <<<<<<<<<<<<<<
 *         self.hits = self.sets = self.misses = 0
 * 
 */

static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_3reset_stats(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_reset_stats(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset_stats", 0);

  /* "relstorage/cache/cache.pyx":300
 * 
 *     cpdef reset_stats(self):
 *         self.hits = self.sets = self.misses = 0             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_v_self->hits = 0;
  __pyx_v_self->sets = 0;
  __pyx_v_self->misses = 0;

  /* "relstorage/cache/cache.pyx":299
 *         self.sets = self.hits = self.misses = 0
 * 
 *     cpdef reset_stats(self):             # <<<<<<<<<<<<<<
 *         self.hits = self.sets = self.misses = 0
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_3reset_stats(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_2reset_stats, "PyCache.reset_stats(self)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_3reset_stats = {"reset_stats", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_3reset_stats, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_2reset_stats};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_3reset_stats(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset_stats (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_2reset_stats(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_2reset_stats(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("reset_stats", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_7PyCache_reset_stats(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.reset_stats", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":302
 *         self.hits = self.sets = self.misses = 0
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def limit(self):
 *         return self.cache.max_weight()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_5limit_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_5limit_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_5limit___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_5limit___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":304
 *     @property
 *     def limit(self):
 *         return self.cache.max_weight()             # <<<<<<<<<<<<<<
 * 
 *     # Access to generations
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->cache.max_weight()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":302
 *         self.hits = self.sets = self.misses = 0
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def limit(self):
 *         return self.cache.max_weight()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.limit.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":307
 * 
 *     # Access to generations
 *     @property             # <<<<<<<<<<<<<<
 *     def eden(self):
 *         return PyGeneration.from_generation(self.cache.ring_eden,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_4eden_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_4eden_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_4eden___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_4eden___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":309
 *     @property
 *     def eden(self):
 *         return PyGeneration.from_generation(self.cache.ring_eden,             # <<<<<<<<<<<<<<
 *                                             'eden',
 *                                             self)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "relstorage/cache/cache.pyx":311
 *         return PyGeneration.from_generation(self.cache.ring_eden,
 *                                             'eden',
 *                                             self)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_12PyGeneration_from_generation(__pyx_v_self->cache.ring_eden, __pyx_n_s_eden, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":307
 * 
 *     # Access to generations
 *     @property             # <<<<<<<<<<<<<<
 *     def eden(self):
 *         return PyGeneration.from_generation(self.cache.ring_eden,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.eden.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":313
 *                                             self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def protected(self):
 *         return PyGeneration.from_generation(self.cache.ring_protected,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_9protected_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_9protected_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_9protected___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_9protected___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":315
 *     @property
 *     def protected(self):
 *         return PyGeneration.from_generation(self.cache.ring_protected,             # <<<<<<<<<<<<<<
 *                                             'protected',
 *                                             self)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "relstorage/cache/cache.pyx":317
 *         return PyGeneration.from_generation(self.cache.ring_protected,
 *                                             'protected',
 *                                             self)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_12PyGeneration_from_generation(__pyx_v_self->cache.ring_protected, __pyx_n_s_protected, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":313
 *                                             self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def protected(self):
 *         return PyGeneration.from_generation(self.cache.ring_protected,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.protected.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":319
 *                                             self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def probation(self):
 *         return PyGeneration.from_generation(self.cache.ring_probation,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_9probation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_9probation_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_9probation___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_9probation___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":321
 *     @property
 *     def probation(self):
 *         return PyGeneration.from_generation(self.cache.ring_probation,             # <<<<<<<<<<<<<<
 *                                             'probation',
 *                                             self)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "relstorage/cache/cache.pyx":323
 *         return PyGeneration.from_generation(self.cache.ring_probation,
 *                                             'probation',
 *                                             self)             # <<<<<<<<<<<<<<
 *     @property
 *     def generations(self):
 */
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_12PyGeneration_from_generation(__pyx_v_self->cache.ring_probation, __pyx_n_s_probation, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":319
 *                                             self)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def probation(self):
 *         return PyGeneration.from_generation(self.cache.ring_probation,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.probation.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":324
 *                                             'probation',
 *                                             self)
 *     @property             # <<<<<<<<<<<<<<
 *     def generations(self):
 *         return [NoSuchGeneration(0),
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_11generations_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_11generations_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_11generations___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_11generations___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":326
 *     @property
 *     def generations(self):
 *         return [NoSuchGeneration(0),             # <<<<<<<<<<<<<<
 *                 self.eden,
 *                 self.protected,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_NoSuchGeneration); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_int_0};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 326, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "relstorage/cache/cache.pyx":327
 *     def generations(self):
 *         return [NoSuchGeneration(0),
 *                 self.eden,             # <<<<<<<<<<<<<<
 *                 self.protected,
 *                 self.probation,]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eden); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "relstorage/cache/cache.pyx":328
 *         return [NoSuchGeneration(0),
 *                 self.eden,
 *                 self.protected,             # <<<<<<<<<<<<<<
 *                 self.probation,]
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_protected); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "relstorage/cache/cache.pyx":329
 *                 self.eden,
 *                 self.protected,
 *                 self.probation,]             # <<<<<<<<<<<<<<
 * 
 *     # Mapping operations
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_probation); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "relstorage/cache/cache.pyx":326
 *     @property
 *     def generations(self):
 *         return [NoSuchGeneration(0),             # <<<<<<<<<<<<<<
 *                 self.eden,
 *                 self.protected,
 */
  __pyx_t_6 = PyList_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_6, 3, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":324
 *                                             'probation',
 *                                             self)
 *     @property             # <<<<<<<<<<<<<<
 *     def generations(self):
 *         return [NoSuchGeneration(0),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.generations.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":333
 *     # Mapping operations
 * 
 *     def __nonzero__(self):             # <<<<<<<<<<<<<<
 *         # Cython still uses __nonzero__
 *         return self.cache.size() > 0
 */

/* Python wrapper */
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_5__nonzero__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_5__nonzero__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__nonzero__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_4__nonzero__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_4__nonzero__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__nonzero__", 0);

  /* "relstorage/cache/cache.pyx":335
 *     def __nonzero__(self):
 *         # Cython still uses __nonzero__
 *         return self.cache.size() > 0             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, OID_t key):
 */
  __pyx_r = (__pyx_v_self->cache.size() > 0);
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":333
 *     # Mapping operations
 * 
 *     def __nonzero__(self):             # <<<<<<<<<<<<<<
 *         # Cython still uses __nonzero__
 *         return self.cache.size() > 0
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":337
 *         return self.cache.size() > 0
 * 
 *     def __contains__(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.cache.contains(key)
 * 
 */

/* Python wrapper */
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_7__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  relstorage::cache::OID_t __pyx_v_key;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_int64_t(__pyx_arg_key); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_6__contains__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((relstorage::cache::OID_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_6__contains__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "relstorage/cache/cache.pyx":338
 * 
 *     def __contains__(self, OID_t key):
 *         return self.cache.contains(key)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __pyx_r = __pyx_v_self->cache.contains(__pyx_v_key);
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":337
 *         return self.cache.size() > 0
 * 
 *     def __contains__(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.cache.contains(key)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":340
 *         return self.cache.contains(key)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.cache.size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_10relstorage_5cache_5cache_7PyCache_9__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_10relstorage_5cache_5cache_7PyCache_9__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_8__len__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_10relstorage_5cache_5cache_7PyCache_8__len__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "relstorage/cache/cache.pyx":341
 * 
 *     def __len__(self):
 *         return self.cache.size()             # <<<<<<<<<<<<<<
 * 
 *     cpdef CachedValue get(self, OID_t key):
 */
  __pyx_r = __pyx_v_self->cache.size();
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":340
 *         return self.cache.contains(key)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.cache.size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":343
 *         return self.cache.size()
 * 
 *     cpdef CachedValue get(self, OID_t key):             # <<<<<<<<<<<<<<
 *         entry = self.cache.get(key)
 *         if not entry:
 */

static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_11get(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_7PyCache_get(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch) {
  relstorage::cache::ICacheEntry *__pyx_v_entry;
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);

  /* "relstorage/cache/cache.pyx":344
 * 
 *     cpdef CachedValue get(self, OID_t key):
 *         entry = self.cache.get(key)             # <<<<<<<<<<<<<<
 *         if not entry:
 *             return None
 */
  __pyx_v_entry = __pyx_v_self->cache.get(__pyx_v_key);

  /* "relstorage/cache/cache.pyx":345
 *     cpdef CachedValue get(self, OID_t key):
 *         entry = self.cache.get(key)
 *         if not entry:             # <<<<<<<<<<<<<<
 *             return None
 *         return python_from_entry_p(entry)
 */
  __pyx_t_1 = ((!(__pyx_v_entry != 0)) != 0);
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":346
 *         entry = self.cache.get(key)
 *         if not entry:
 *             return None             # <<<<<<<<<<<<<<
 *         return python_from_entry_p(entry)
 * 
 */
    __Pyx_XDECREF((PyObject *)__pyx_r);
    __pyx_r = ((struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":345
 *     cpdef CachedValue get(self, OID_t key):
 *         entry = self.cache.get(key)
 *         if not entry:             # <<<<<<<<<<<<<<
 *             return None
 *         return python_from_entry_p(entry)
 */
  }

  /* "relstorage/cache/cache.pyx":347
 *         if not entry:
 *             return None
 *         return python_from_entry_p(entry)             # <<<<<<<<<<<<<<
 * 
 *     cpdef CachedValue peek(self, OID_t key):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_python_from_entry_p(__pyx_v_entry)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":343
 *         return self.cache.size()
 * 
 *     cpdef CachedValue get(self, OID_t key):             # <<<<<<<<<<<<<<
 *         entry = self.cache.get(key)
 *         if not entry:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_11get(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_10get, "PyCache.get(self, OID_t key) -> CachedValue");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_11get = {"get", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_11get, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_10get};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_11get(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  relstorage::cache::OID_t __pyx_v_key;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_int64_t(__pyx_arg_key); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_10get(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((relstorage::cache::OID_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_10get(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_7PyCache_get(__pyx_v_self, __pyx_v_key, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":349
 *         return python_from_entry_p(entry)
 * 
 *     cpdef CachedValue peek(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */

static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_13peek(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_f_10relstorage_5cache_5cache_7PyCache_peek(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek", 0);

  /* "relstorage/cache/cache.pyx":350
 * 
 *     cpdef CachedValue peek(self, OID_t key):
 *         return self.get(key)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_7PyCache_get(__pyx_v_self, __pyx_v_key, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":349
 *         return python_from_entry_p(entry)
 * 
 *     cpdef CachedValue peek(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.peek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_13peek(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_12peek, "PyCache.peek(self, OID_t key) -> CachedValue");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_13peek = {"peek", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_13peek, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_12peek};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_13peek(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  relstorage::cache::OID_t __pyx_v_key;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("peek (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_int64_t(__pyx_arg_key); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 349, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.peek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_12peek(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((relstorage::cache::OID_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_12peek(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_7PyCache_peek(__pyx_v_self, __pyx_v_key, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.peek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":352
 *         return self.get(key)
 * 
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):             # <<<<<<<<<<<<<<
 *         value = self.cache.peek(key, tid)
 *         if value:
 */

static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_15peek_item_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_peek_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, relstorage::cache::TID_t __pyx_v_tid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  relstorage::cache::SVCacheEntry *__pyx_v_value;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek_item_with_tid", 0);

  /* "relstorage/cache/cache.pyx":353
 * 
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):
 *         value = self.cache.peek(key, tid)             # <<<<<<<<<<<<<<
 *         if value:
 *             return python_from_entry_p(value)
 */
  __pyx_v_value = __pyx_v_self->cache.peek(__pyx_v_key, __pyx_v_tid);

  /* "relstorage/cache/cache.pyx":354
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):
 *         value = self.cache.peek(key, tid)
 *         if value:             # <<<<<<<<<<<<<<
 *             return python_from_entry_p(value)
 * 
 */
  __pyx_t_1 = (__pyx_v_value != 0);
  if (__pyx_t_1) {

    /* "relstorage/cache/cache.pyx":355
 *         value = self.cache.peek(key, tid)
 *         if value:
 *             return python_from_entry_p(value)             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint contains_oid_with_tid(self, OID_t key, tid):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_python_from_entry_p(__pyx_v_value)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":354
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):
 *         value = self.cache.peek(key, tid)
 *         if value:             # <<<<<<<<<<<<<<
 *             return python_from_entry_p(value)
 * 
 */
  }

  /* "relstorage/cache/cache.pyx":352
 *         return self.get(key)
 * 
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):             # <<<<<<<<<<<<<<
 *         value = self.cache.peek(key, tid)
 *         if value:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.peek_item_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_15peek_item_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_14peek_item_with_tid, "PyCache.peek_item_with_tid(self, OID_t key, TID_t tid)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_15peek_item_with_tid = {"peek_item_with_tid", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10relstorage_5cache_5cache_7PyCache_15peek_item_with_tid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_14peek_item_with_tid};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_15peek_item_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  relstorage::cache::OID_t __pyx_v_key;
  relstorage::cache::TID_t __pyx_v_tid;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("peek_item_with_tid (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_tid,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_tid,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tid)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("peek_item_with_tid", 1, 2, 2, 1); __PYX_ERR(0, 352, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "peek_item_with_tid") < 0)) __PYX_ERR(0, 352, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_key = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L3_error)
    __pyx_v_tid = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_tid == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("peek_item_with_tid", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 352, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.peek_item_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_14peek_item_with_tid(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), __pyx_v_key, __pyx_v_tid);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_14peek_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, relstorage::cache::TID_t __pyx_v_tid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("peek_item_with_tid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_7PyCache_peek_item_with_tid(__pyx_v_self, __pyx_v_key, __pyx_v_tid, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.peek_item_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":357
 *             return python_from_entry_p(value)
 * 
 *     cpdef bint contains_oid_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         """
 *         For use during cache validation. Can only be used when
 */

static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_17contains_oid_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int __pyx_f_10relstorage_5cache_5cache_7PyCache_contains_oid_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  relstorage::cache::TID_t __pyx_v_native_tid;
  relstorage::cache::SVCacheEntry *__pyx_v_entry;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  relstorage::cache::TID_t __pyx_t_1;
  int __pyx_t_2;
  relstorage::cache::TID_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("contains_oid_with_tid", 0);

  /* "relstorage/cache/cache.pyx":363
 *         to do this will lead to memory leaks.
 *         """
 *         cdef TID_t native_tid = -1 if tid is None else tid             # <<<<<<<<<<<<<<
 *         cdef SVCacheEntry* entry = self.cache.peek(key, native_tid)
 *         if not entry:
 */
  __pyx_t_2 = (__pyx_v_tid == Py_None);
  if ((__pyx_t_2 != 0)) {
    __pyx_t_1 = -1L;
  } else {
    __pyx_t_3 = __Pyx_PyInt_As_int64_t(__pyx_v_tid); if (unlikely((__pyx_t_3 == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 363, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_3;
  }
  __pyx_v_native_tid = __pyx_t_1;

  /* "relstorage/cache/cache.pyx":364
 *         """
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* entry = self.cache.peek(key, native_tid)             # <<<<<<<<<<<<<<
 *         if not entry:
 *             return False
 */
  __pyx_v_entry = __pyx_v_self->cache.peek(__pyx_v_key, __pyx_v_native_tid);

  /* "relstorage/cache/cache.pyx":365
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* entry = self.cache.peek(key, native_tid)
 *         if not entry:             # <<<<<<<<<<<<<<
 *             return False
 *         # We're guaranteed that entry is an existing object, not one we need
 */
  __pyx_t_2 = ((!(__pyx_v_entry != 0)) != 0);
  if (__pyx_t_2) {

    /* "relstorage/cache/cache.pyx":366
 *         cdef SVCacheEntry* entry = self.cache.peek(key, native_tid)
 *         if not entry:
 *             return False             # <<<<<<<<<<<<<<
 *         # We're guaranteed that entry is an existing object, not one we need
 *         # to manage the lifetime of, because of the requirement that only
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":365
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* entry = self.cache.peek(key, native_tid)
 *         if not entry:             # <<<<<<<<<<<<<<
 *             return False
 *         # We're guaranteed that entry is an existing object, not one we need
 */
  }

  /* "relstorage/cache/cache.pyx":370
 *         # to manage the lifetime of, because of the requirement that only
 *         # single values can be in the cache at this time.
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, OID_t key):
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":357
 *             return python_from_entry_p(value)
 * 
 *     cpdef bint contains_oid_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         """
 *         For use during cache validation. Can only be used when
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("relstorage.cache.cache.PyCache.contains_oid_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_17contains_oid_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_16contains_oid_with_tid, "PyCache.contains_oid_with_tid(self, OID_t key, tid) -> bool\n\n        For use during cache validation. Can only be used when\n        the cache only contains SingleValue entries; failure\n        to do this will lead to memory leaks.\n        ");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_17contains_oid_with_tid = {"contains_oid_with_tid", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10relstorage_5cache_5cache_7PyCache_17contains_oid_with_tid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_16contains_oid_with_tid};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_17contains_oid_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  relstorage::cache::OID_t __pyx_v_key;
  PyObject *__pyx_v_tid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("contains_oid_with_tid (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_tid,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_tid,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tid)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("contains_oid_with_tid", 1, 2, 2, 1); __PYX_ERR(0, 357, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "contains_oid_with_tid") < 0)) __PYX_ERR(0, 357, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_key = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L3_error)
    __pyx_v_tid = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("contains_oid_with_tid", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 357, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.contains_oid_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_16contains_oid_with_tid(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), __pyx_v_key, __pyx_v_tid);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_16contains_oid_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("contains_oid_with_tid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_10relstorage_5cache_5cache_7PyCache_contains_oid_with_tid(__pyx_v_self, __pyx_v_key, __pyx_v_tid, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.contains_oid_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":372
 *         return True
 * 
 *     def __getitem__(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_19__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_19__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  relstorage::cache::OID_t __pyx_v_key;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_int64_t(__pyx_arg_key); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 372, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_18__getitem__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((relstorage::cache::OID_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_18__getitem__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "relstorage/cache/cache.pyx":373
 * 
 *     def __getitem__(self, OID_t key):
 *         return self.get(key)             # <<<<<<<<<<<<<<
 * 
 *     cpdef get_item_with_tid(self, OID_t key, tid):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_7PyCache_get(__pyx_v_self, __pyx_v_key, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":372
 *         return True
 * 
 *     def __getitem__(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":375
 *         return self.get(key)
 * 
 *     cpdef get_item_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)
 */

static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_21get_item_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_get_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid, CYTHON_UNUSED int __pyx_skip_dispatch) {
  relstorage::cache::TID_t __pyx_v_native_tid;
  relstorage::cache::SVCacheEntry *__pyx_v_cvalue;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  relstorage::cache::TID_t __pyx_t_1;
  int __pyx_t_2;
  relstorage::cache::TID_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_with_tid", 0);

  /* "relstorage/cache/cache.pyx":376
 * 
 *     cpdef get_item_with_tid(self, OID_t key, tid):
 *         cdef TID_t native_tid = -1 if tid is None else tid             # <<<<<<<<<<<<<<
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)
 * 
 */
  __pyx_t_2 = (__pyx_v_tid == Py_None);
  if ((__pyx_t_2 != 0)) {
    __pyx_t_1 = -1L;
  } else {
    __pyx_t_3 = __Pyx_PyInt_As_int64_t(__pyx_v_tid); if (unlikely((__pyx_t_3 == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 376, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_3;
  }
  __pyx_v_native_tid = __pyx_t_1;

  /* "relstorage/cache/cache.pyx":377
 *     cpdef get_item_with_tid(self, OID_t key, tid):
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)             # <<<<<<<<<<<<<<
 * 
 *         if cvalue:
 */
  __pyx_v_cvalue = __pyx_v_self->cache.get(__pyx_v_key, __pyx_v_native_tid);

  /* "relstorage/cache/cache.pyx":379
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)
 * 
 *         if cvalue:             # <<<<<<<<<<<<<<
 *             self.hits += 1
 *             return SingleValue.from_entry(cvalue)
 */
  __pyx_t_2 = (__pyx_v_cvalue != 0);
  if (__pyx_t_2) {

    /* "relstorage/cache/cache.pyx":380
 * 
 *         if cvalue:
 *             self.hits += 1             # <<<<<<<<<<<<<<
 *             return SingleValue.from_entry(cvalue)
 * 
 */
    __pyx_v_self->hits = (__pyx_v_self->hits + 1);

    /* "relstorage/cache/cache.pyx":381
 *         if cvalue:
 *             self.hits += 1
 *             return SingleValue.from_entry(cvalue)             # <<<<<<<<<<<<<<
 * 
 *         self.misses += 1
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_11SingleValue_from_entry(__pyx_v_cvalue)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":379
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)
 * 
 *         if cvalue:             # <<<<<<<<<<<<<<
 *             self.hits += 1
 *             return SingleValue.from_entry(cvalue)
 */
  }

  /* "relstorage/cache/cache.pyx":383
 *             return SingleValue.from_entry(cvalue)
 * 
 *         self.misses += 1             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, OID_t key, tuple value):
 */
  __pyx_v_self->misses = (__pyx_v_self->misses + 1);

  /* "relstorage/cache/cache.pyx":375
 *         return self.get(key)
 * 
 *     cpdef get_item_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.get_item_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_21get_item_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_20get_item_with_tid, "PyCache.get_item_with_tid(self, OID_t key, tid)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_21get_item_with_tid = {"get_item_with_tid", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10relstorage_5cache_5cache_7PyCache_21get_item_with_tid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_20get_item_with_tid};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_21get_item_with_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  relstorage::cache::OID_t __pyx_v_key;
  PyObject *__pyx_v_tid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_item_with_tid (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_tid,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_tid,0};
    #endif
    PyObject* values[2] = {0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_key)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 375, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tid)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 375, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_item_with_tid", 1, 2, 2, 1); __PYX_ERR(0, 375, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_item_with_tid") < 0)) __PYX_ERR(0, 375, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_key = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 375, __pyx_L3_error)
    __pyx_v_tid = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_item_with_tid", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 375, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.get_item_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_20get_item_with_tid(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), __pyx_v_key, __pyx_v_tid);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_20get_item_with_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_tid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_with_tid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_7PyCache_get_item_with_tid(__pyx_v_self, __pyx_v_key, __pyx_v_tid, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.get_item_with_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":385
 *         self.misses += 1
 * 
 *     def __setitem__(self, OID_t key, tuple value):             # <<<<<<<<<<<<<<
 *         self._do_set(key, value[0], value[1])
 * 
 */

/* Python wrapper */
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_23__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_23__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key, PyObject *__pyx_v_value) {
  relstorage::cache::OID_t __pyx_v_key;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_int64_t(__pyx_arg_key); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 385, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyTuple_Type), 1, "value", 1))) __PYX_ERR(0, 385, __pyx_L1_error)
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_22__setitem__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((relstorage::cache::OID_t)__pyx_v_key), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_22__setitem__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  relstorage::cache::TID_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "relstorage/cache/cache.pyx":386
 * 
 *     def __setitem__(self, OID_t key, tuple value):
 *         self._do_set(key, value[0], value[1])             # <<<<<<<<<<<<<<
 * 
 *     cdef _do_set(self, OID_t key, object state, TID_t tid):
 */
  if (unlikely(__pyx_v_value == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 386, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_value == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 386, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_GetItemInt_Tuple(__pyx_v_value, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_int64_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_f_10relstorage_5cache_5cache_7PyCache__do_set(__pyx_v_self, __pyx_v_key, __pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "relstorage/cache/cache.pyx":385
 *         self.misses += 1
 * 
 *     def __setitem__(self, OID_t key, tuple value):             # <<<<<<<<<<<<<<
 *         self._do_set(key, value[0], value[1])
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":388
 *         self._do_set(key, value[0], value[1])
 * 
 *     cdef _do_set(self, OID_t key, object state, TID_t tid):             # <<<<<<<<<<<<<<
 *         # Do all this down here so we don't give up the GIL.
 *         cdef object b_state = state if state is not None else b''
 */

static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache__do_set(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key, PyObject *__pyx_v_state, relstorage::cache::TID_t __pyx_v_tid) {
  PyObject *__pyx_v_b_state = 0;
  relstorage::cache::ProposedCacheEntry __pyx_v_proposed;
  PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_do_set", 0);

  /* "relstorage/cache/cache.pyx":390
 *     cdef _do_set(self, OID_t key, object state, TID_t tid):
 *         # Do all this down here so we don't give up the GIL.
 *         cdef object b_state = state if state is not None else b''             # <<<<<<<<<<<<<<
 *         cdef ProposedCacheEntry proposed = ProposedCacheEntry(key, tid, b_state)
 *         if not self.cache.contains(key): # the long way to avoid type conversion
 */
  __pyx_t_2 = (__pyx_v_state != Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(__pyx_v_state);
    __pyx_t_1 = __pyx_v_state;
  } else {
    __Pyx_INCREF(__pyx_kp_b__2);
    __pyx_t_1 = __pyx_kp_b__2;
  }
  __pyx_v_b_state = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":391
 *         # Do all this down here so we don't give up the GIL.
 *         cdef object b_state = state if state is not None else b''
 *         cdef ProposedCacheEntry proposed = ProposedCacheEntry(key, tid, b_state)             # <<<<<<<<<<<<<<
 *         if not self.cache.contains(key): # the long way to avoid type conversion
 *             self.cache.add_to_eden(proposed)
 */
  __pyx_v_proposed = relstorage::cache::ProposedCacheEntry(__pyx_v_key, __pyx_v_tid, __pyx_v_b_state);

  /* "relstorage/cache/cache.pyx":392
 *         cdef object b_state = state if state is not None else b''
 *         cdef ProposedCacheEntry proposed = ProposedCacheEntry(key, tid, b_state)
 *         if not self.cache.contains(key): # the long way to avoid type conversion             # <<<<<<<<<<<<<<
 *             self.cache.add_to_eden(proposed)
 *         else:
 */
  __pyx_t_2 = ((!(__pyx_v_self->cache.contains(__pyx_v_key) != 0)) != 0);
  if (__pyx_t_2) {

    /* "relstorage/cache/cache.pyx":393
 *         cdef ProposedCacheEntry proposed = ProposedCacheEntry(key, tid, b_state)
 *         if not self.cache.contains(key): # the long way to avoid type conversion
 *             self.cache.add_to_eden(proposed)             # <<<<<<<<<<<<<<
 *         else:
 *             # We need to merge values.
 */
    try {
      __pyx_v_self->cache.add_to_eden(__pyx_v_proposed);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 393, __pyx_L1_error)
    }

    /* "relstorage/cache/cache.pyx":392
 *         cdef object b_state = state if state is not None else b''
 *         cdef ProposedCacheEntry proposed = ProposedCacheEntry(key, tid, b_state)
 *         if not self.cache.contains(key): # the long way to avoid type conversion             # <<<<<<<<<<<<<<
 *             self.cache.add_to_eden(proposed)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "relstorage/cache/cache.pyx":396
 *         else:
 *             # We need to merge values.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cache.store_and_make_MRU(proposed)
 *             except RuntimeError as e:
 */
  /*else*/ {
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "relstorage/cache/cache.pyx":397
 *             # We need to merge values.
 *             try:
 *                 self.cache.store_and_make_MRU(proposed)             # <<<<<<<<<<<<<<
 *             except RuntimeError as e:
 *                 raise CacheConsistencyError(str(e))
 */
        try {
          __pyx_v_self->cache.store_and_make_MRU(__pyx_v_proposed);
        } catch(...) {
          __Pyx_CppExn2PyErr();
          __PYX_ERR(0, 397, __pyx_L4_error)
        }

        /* "relstorage/cache/cache.pyx":396
 *         else:
 *             # We need to merge values.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cache.store_and_make_MRU(proposed)
 *             except RuntimeError as e:
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "relstorage/cache/cache.pyx":398
 *             try:
 *                 self.cache.store_and_make_MRU(proposed)
 *             except RuntimeError as e:             # <<<<<<<<<<<<<<
 *                 raise CacheConsistencyError(str(e))
 * 
 */
      __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_RuntimeError);
      if (__pyx_t_6) {
        __Pyx_AddTraceback("relstorage.cache.cache.PyCache._do_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 398, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __pyx_v_e = __pyx_t_7;
        /*try:*/ {

          /* "relstorage/cache/cache.pyx":399
 *                 self.cache.store_and_make_MRU(proposed)
 *             except RuntimeError as e:
 *                 raise CacheConsistencyError(str(e))             # <<<<<<<<<<<<<<
 * 
 *         self.sets += 1
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_CacheConsistencyError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 399, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_Str(__pyx_v_e); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 399, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          __pyx_t_6 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_6 = 1;
            }
          }
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_11};
            __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 399, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __Pyx_Raise(__pyx_t_9, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __PYX_ERR(0, 399, __pyx_L15_error)
        }

        /* "relstorage/cache/cache.pyx":398
 *             try:
 *                 self.cache.store_and_make_MRU(proposed)
 *             except RuntimeError as e:             # <<<<<<<<<<<<<<
 *                 raise CacheConsistencyError(str(e))
 * 
 */
        /*finally:*/ {
          __pyx_L15_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_15);
            __Pyx_XGOTREF(__pyx_t_16);
            __Pyx_XGOTREF(__pyx_t_17);
            __Pyx_XGOTREF(__pyx_t_18);
            __Pyx_XGOTREF(__pyx_t_19);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_6 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_15);
            __Pyx_XGIVEREF(__pyx_t_16);
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
            __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
            goto __pyx_L6_except_error;
          }
        }
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "relstorage/cache/cache.pyx":396
 *         else:
 *             # We need to merge values.
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cache.store_and_make_MRU(proposed)
 *             except RuntimeError as e:
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }
  }
  __pyx_L3:;

  /* "relstorage/cache/cache.pyx":401
 *                 raise CacheConsistencyError(str(e))
 * 
 *         self.sets += 1             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, OID_t key):
 */
  __pyx_v_self->sets = (__pyx_v_self->sets + 1);

  /* "relstorage/cache/cache.pyx":388
 *         self._do_set(key, value[0], value[1])
 * 
 *     cdef _do_set(self, OID_t key, object state, TID_t tid):             # <<<<<<<<<<<<<<
 *         # Do all this down here so we don't give up the GIL.
 *         cdef object b_state = state if state is not None else b''
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache._do_set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_b_state);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":403
 *         self.sets += 1
 * 
 *     def __delitem__(self, OID_t key):             # <<<<<<<<<<<<<<
 *         self.cache.delitem(key)
 * 
 */

/* Python wrapper */
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_25__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key); /*proto*/
static int __pyx_pw_10relstorage_5cache_5cache_7PyCache_25__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_key) {
  relstorage::cache::OID_t __pyx_v_key;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_key); {
    __pyx_v_key = __Pyx_PyInt_As_int64_t(__pyx_arg_key); if (unlikely((__pyx_v_key == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 403, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_24__delitem__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((relstorage::cache::OID_t)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10relstorage_5cache_5cache_7PyCache_24__delitem__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::OID_t __pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "relstorage/cache/cache.pyx":404
 * 
 *     def __delitem__(self, OID_t key):
 *         self.cache.delitem(key)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  try {
    __pyx_v_self->cache.delitem(__pyx_v_key);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 404, __pyx_L1_error)
  }

  /* "relstorage/cache/cache.pyx":403
 *         self.sets += 1
 * 
 *     def __delitem__(self, OID_t key):             # <<<<<<<<<<<<<<
 *         self.cache.delitem(key)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_28generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "relstorage/cache/cache.pyx":406
 *         self.cache.delitem(key)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across all the contained OID/TID pairs (the keys).
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_27__iter__(PyObject *__pyx_v_self); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_26__iter__, "\n        Iterate across all the contained OID/TID pairs (the keys).\n\n        Note that the number of items in the iterator may exceed\n        the number of items in the cache due to aliasing.\n\n        This is not thread safe.\n        ");
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_10relstorage_5cache_5cache_7PyCache_26__iter__;
#endif
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_27__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_26__iter__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_26__iter__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__(__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 406, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10relstorage_5cache_5cache_7PyCache_28generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_PyCache___iter, __pyx_n_s_relstorage_cache_cache); if (unlikely(!gen)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_28generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *__pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  relstorage::cache::OID_t __pyx_t_2;
  std::vector<relstorage::cache::TID_t>  __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 406, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":415
 *         This is not thread safe.
 *         """
 *         it = self.cache.begin()             # <<<<<<<<<<<<<<
 *         end = self.cache.end()
 * 
 */
  __pyx_cur_scope->__pyx_v_it = __pyx_cur_scope->__pyx_v_self->cache.begin();

  /* "relstorage/cache/cache.pyx":416
 *         """
 *         it = self.cache.begin()
 *         end = self.cache.end()             # <<<<<<<<<<<<<<
 * 
 *         while it != end:
 */
  __pyx_cur_scope->__pyx_v_end = __pyx_cur_scope->__pyx_v_self->cache.end();

  /* "relstorage/cache/cache.pyx":418
 *         end = self.cache.end()
 * 
 *         while it != end:             # <<<<<<<<<<<<<<
 *             oid = deref(it).key
 *             # This should be a small list. It mustn't persist
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_it != __pyx_cur_scope->__pyx_v_end) != 0);
    if (!__pyx_t_1) break;

    /* "relstorage/cache/cache.pyx":419
 * 
 *         while it != end:
 *             oid = deref(it).key             # <<<<<<<<<<<<<<
 *             # This should be a small list. It mustn't persist
 *             # across the yield boundary, we don't own it but Cython
 */
    __pyx_t_2 = (*__pyx_cur_scope->__pyx_v_it).key;
    __pyx_cur_scope->__pyx_v_oid = __pyx_t_2;

    /* "relstorage/cache/cache.pyx":423
 *             # across the yield boundary, we don't own it but Cython
 *             # tries to keep it as a reference
 *             tids = list(deref(it).all_tids())             # <<<<<<<<<<<<<<
 *             for t in tids:
 *                 yield (oid, t)
 */
    try {
      __pyx_t_3 = (*__pyx_cur_scope->__pyx_v_it).all_tids();
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 423, __pyx_L1_error)
    }
    __pyx_t_4 = __pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PySequence_ListKeepNew(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_tids);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_tids, ((PyObject*)__pyx_t_5));
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "relstorage/cache/cache.pyx":424
 *             # tries to keep it as a reference
 *             tids = list(deref(it).all_tids())
 *             for t in tids:             # <<<<<<<<<<<<<<
 *                 yield (oid, t)
 *             preincr(it)
 */
    __pyx_t_5 = __pyx_cur_scope->__pyx_v_tids; __Pyx_INCREF(__pyx_t_5); __pyx_t_6 = 0;
    for (;;) {
      if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_5)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_6); __Pyx_INCREF(__pyx_t_4); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 424, __pyx_L1_error)
      #else
      __pyx_t_4 = PySequence_ITEM(__pyx_t_5, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 424, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_t);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_t, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;

      /* "relstorage/cache/cache.pyx":425
 *             tids = list(deref(it).all_tids())
 *             for t in tids:
 *                 yield (oid, t)             # <<<<<<<<<<<<<<
 *             preincr(it)
 * 
 */
      __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_cur_scope->__pyx_v_oid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_t);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_t);
      PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_cur_scope->__pyx_v_t);
      __pyx_t_4 = 0;
      __pyx_r = __pyx_t_7;
      __pyx_t_7 = 0;
      __Pyx_XGIVEREF(__pyx_t_5);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_5;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_6;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L8_resume_from_yield:;
      __pyx_t_5 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_5);
      __pyx_t_6 = __pyx_cur_scope->__pyx_t_1;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 425, __pyx_L1_error)

      /* "relstorage/cache/cache.pyx":424
 *             # tries to keep it as a reference
 *             tids = list(deref(it).all_tids())
 *             for t in tids:             # <<<<<<<<<<<<<<
 *                 yield (oid, t)
 *             preincr(it)
 */
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "relstorage/cache/cache.pyx":426
 *             for t in tids:
 *                 yield (oid, t)
 *             preincr(it)             # <<<<<<<<<<<<<<
 * 
 *     def iteritems(self):
 */
    (void)((++__pyx_cur_scope->__pyx_v_it));
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "relstorage/cache/cache.pyx":406
 *         self.cache.delitem(key)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across all the contained OID/TID pairs (the keys).
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_31generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "relstorage/cache/cache.pyx":428
 *             preincr(it)
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the oid/cache_value pairs.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_30iteritems(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_29iteritems, "PyCache.iteritems(self)\n\n        Iterate across the oid/cache_value pairs.\n\n        Not thread safe.\n        ");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_30iteritems = {"iteritems", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_30iteritems, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_29iteritems};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_30iteritems(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iteritems (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_29iteritems(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_29iteritems(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("iteritems", 0);
  __pyx_cur_scope = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems(__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 428, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10relstorage_5cache_5cache_7PyCache_31generator2, __pyx_codeobj__3, (PyObject *) __pyx_cur_scope, __pyx_n_s_iteritems, __pyx_n_s_PyCache_iteritems, __pyx_n_s_relstorage_cache_cache); if (unlikely(!gen)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.iteritems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_31generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *__pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("iteritems", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 428, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":434
 *         Not thread safe.
 *         """
 *         it = self.cache.begin()             # <<<<<<<<<<<<<<
 *         end = self.cache.end()
 * 
 */
  __pyx_cur_scope->__pyx_v_it = __pyx_cur_scope->__pyx_v_self->cache.begin();

  /* "relstorage/cache/cache.pyx":435
 *         """
 *         it = self.cache.begin()
 *         end = self.cache.end()             # <<<<<<<<<<<<<<
 * 
 *         while it != end:
 */
  __pyx_cur_scope->__pyx_v_end = __pyx_cur_scope->__pyx_v_self->cache.end();

  /* "relstorage/cache/cache.pyx":437
 *         end = self.cache.end()
 * 
 *         while it != end:             # <<<<<<<<<<<<<<
 *             yield (deref(it).key, python_from_entry(deref(it)))
 *             preincr(it)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_it != __pyx_cur_scope->__pyx_v_end) != 0);
    if (!__pyx_t_1) break;

    /* "relstorage/cache/cache.pyx":438
 * 
 *         while it != end:
 *             yield (deref(it).key, python_from_entry(deref(it)))             # <<<<<<<<<<<<<<
 *             preincr(it)
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int64_t((*__pyx_cur_scope->__pyx_v_it).key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_python_from_entry((*__pyx_cur_scope->__pyx_v_it))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 438, __pyx_L1_error)

    /* "relstorage/cache/cache.pyx":439
 *         while it != end:
 *             yield (deref(it).key, python_from_entry(deref(it)))
 *             preincr(it)             # <<<<<<<<<<<<<<
 * 
 *     def keys(self):
 */
    (void)((++__pyx_cur_scope->__pyx_v_it));
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "relstorage/cache/cache.pyx":428
 *             preincr(it)
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the oid/cache_value pairs.
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("iteritems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_34generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "relstorage/cache/cache.pyx":441
 *             preincr(it)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the OIDs in the cache.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_33keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_32keys, "PyCache.keys(self)\n\n        Iterate across the OIDs in the cache.\n\n        Not thread safe.\n        ");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_33keys = {"keys", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_33keys, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_32keys};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_33keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_32keys(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_32keys(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);
  __pyx_cur_scope = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_3_keys(__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 441, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10relstorage_5cache_5cache_7PyCache_34generator3, __pyx_codeobj__4, (PyObject *) __pyx_cur_scope, __pyx_n_s_keys, __pyx_n_s_PyCache_keys, __pyx_n_s_relstorage_cache_cache); if (unlikely(!gen)) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_34generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *__pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 441, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":447
 *         Not thread safe.
 *         """
 *         it = self.cache.begin()             # <<<<<<<<<<<<<<
 *         end = self.cache.end()
 * 
 */
  __pyx_cur_scope->__pyx_v_it = __pyx_cur_scope->__pyx_v_self->cache.begin();

  /* "relstorage/cache/cache.pyx":448
 *         """
 *         it = self.cache.begin()
 *         end = self.cache.end()             # <<<<<<<<<<<<<<
 * 
 *         while it != end:
 */
  __pyx_cur_scope->__pyx_v_end = __pyx_cur_scope->__pyx_v_self->cache.end();

  /* "relstorage/cache/cache.pyx":450
 *         end = self.cache.end()
 * 
 *         while it != end:             # <<<<<<<<<<<<<<
 *             yield deref(it).key
 *             preincr(it)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_it != __pyx_cur_scope->__pyx_v_end) != 0);
    if (!__pyx_t_1) break;

    /* "relstorage/cache/cache.pyx":451
 * 
 *         while it != end:
 *             yield deref(it).key             # <<<<<<<<<<<<<<
 *             preincr(it)
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int64_t((*__pyx_cur_scope->__pyx_v_it).key); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 451, __pyx_L1_error)

    /* "relstorage/cache/cache.pyx":452
 *         while it != end:
 *             yield deref(it).key
 *             preincr(it)             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */
    (void)((++__pyx_cur_scope->__pyx_v_it));
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "relstorage/cache/cache.pyx":441
 *             preincr(it)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the OIDs in the cache.
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_37generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "relstorage/cache/cache.pyx":454
 *             preincr(it)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the values in the cache.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_36values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_35values, "PyCache.values(self)\n\n        Iterate across the values in the cache.\n\n        Not thread safe.\n        ");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_36values = {"values", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_36values, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_35values};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_36values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_35values(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_35values(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);
  __pyx_cur_scope = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_4_values(__pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 454, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10relstorage_5cache_5cache_7PyCache_37generator4, __pyx_codeobj__5, (PyObject *) __pyx_cur_scope, __pyx_n_s_values, __pyx_n_s_PyCache_values, __pyx_n_s_relstorage_cache_cache); if (unlikely(!gen)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10relstorage_5cache_5cache_7PyCache_37generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *__pyx_cur_scope = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 454, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":460
 *         Not thread safe.
 *         """
 *         it = self.cache.begin()             # <<<<<<<<<<<<<<
 *         end = self.cache.end()
 * 
 */
  __pyx_cur_scope->__pyx_v_it = __pyx_cur_scope->__pyx_v_self->cache.begin();

  /* "relstorage/cache/cache.pyx":461
 *         """
 *         it = self.cache.begin()
 *         end = self.cache.end()             # <<<<<<<<<<<<<<
 * 
 *         while it != end:
 */
  __pyx_cur_scope->__pyx_v_end = __pyx_cur_scope->__pyx_v_self->cache.end();

  /* "relstorage/cache/cache.pyx":463
 *         end = self.cache.end()
 * 
 *         while it != end:             # <<<<<<<<<<<<<<
 *             yield python_from_entry(deref(it))
 *             preincr(it)
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_it != __pyx_cur_scope->__pyx_v_end) != 0);
    if (!__pyx_t_1) break;

    /* "relstorage/cache/cache.pyx":464
 * 
 *         while it != end:
 *             yield python_from_entry(deref(it))             # <<<<<<<<<<<<<<
 *             preincr(it)
 * 
 */
    __pyx_t_2 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_python_from_entry((*__pyx_cur_scope->__pyx_v_it))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 464, __pyx_L1_error)

    /* "relstorage/cache/cache.pyx":465
 *         while it != end:
 *             yield python_from_entry(deref(it))
 *             preincr(it)             # <<<<<<<<<<<<<<
 * 
 *     # Cache specific operations
 */
    (void)((++__pyx_cur_scope->__pyx_v_it));
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "relstorage/cache/cache.pyx":454
 *             preincr(it)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the values in the cache.
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":469
 *     # Cache specific operations
 * 
 *     cpdef set_all_for_tid(self, TID_t tid_int, state_oid_iter, compress, Py_ssize_t value_limit):             # <<<<<<<<<<<<<<
 *         # Do all this down here so we don't give up the GIL, assuming the iter is actually a list
 *         # or otherwise not implemented in pure-python
 */

static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_39set_all_for_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_10relstorage_5cache_5cache_7PyCache_set_all_for_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::TID_t __pyx_v_tid_int, PyObject *__pyx_v_state_oid_iter, PyObject *__pyx_v_compress, Py_ssize_t __pyx_v_value_limit, CYTHON_UNUSED int __pyx_skip_dispatch) {
  relstorage::cache::OID_t __pyx_v_oid_int;
  PyObject *__pyx_v_state_bytes = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  relstorage::cache::OID_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_all_for_tid", 0);

  /* "relstorage/cache/cache.pyx":473
 *         # or otherwise not implemented in pure-python
 *         cdef OID_t oid_int
 *         for state_bytes, oid_int, _ in state_oid_iter:             # <<<<<<<<<<<<<<
 *             state_bytes = compress(state_bytes) if compress is not None else state_bytes
 *             state_bytes = state_bytes if state_bytes is not None else b''
 */
  if (likely(PyList_CheckExact(__pyx_v_state_oid_iter)) || PyTuple_CheckExact(__pyx_v_state_oid_iter)) {
    __pyx_t_1 = __pyx_v_state_oid_iter; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_state_oid_iter); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 473, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 473, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 473, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 473, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 473, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
      index = 0; __pyx_t_5 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      index = 2; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 3) < 0) __PYX_ERR(0, 473, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 473, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_6); if (unlikely((__pyx_t_10 == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_state_bytes, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_oid_int = __pyx_t_10;
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "relstorage/cache/cache.pyx":474
 *         cdef OID_t oid_int
 *         for state_bytes, oid_int, _ in state_oid_iter:
 *             state_bytes = compress(state_bytes) if compress is not None else state_bytes             # <<<<<<<<<<<<<<
 *             state_bytes = state_bytes if state_bytes is not None else b''
 *             if len(state_bytes) >= value_limit:
 */
    __pyx_t_11 = (__pyx_v_compress != Py_None);
    if ((__pyx_t_11 != 0)) {
      __Pyx_INCREF(__pyx_v_compress);
      __pyx_t_6 = __pyx_v_compress; __pyx_t_5 = NULL;
      __pyx_t_12 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_12 = 1;
        }
      }
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_state_bytes};
        __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_12, 1+__pyx_t_12);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 474, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_t_4 = __pyx_t_7;
      __pyx_t_7 = 0;
    } else {
      __Pyx_INCREF(__pyx_v_state_bytes);
      __pyx_t_4 = __pyx_v_state_bytes;
    }
    __Pyx_DECREF_SET(__pyx_v_state_bytes, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "relstorage/cache/cache.pyx":475
 *         for state_bytes, oid_int, _ in state_oid_iter:
 *             state_bytes = compress(state_bytes) if compress is not None else state_bytes
 *             state_bytes = state_bytes if state_bytes is not None else b''             # <<<<<<<<<<<<<<
 *             if len(state_bytes) >= value_limit:
 *                 # This value is too big, so don't cache it.
 */
    __pyx_t_11 = (__pyx_v_state_bytes != Py_None);
    if ((__pyx_t_11 != 0)) {
      __Pyx_INCREF(__pyx_v_state_bytes);
      __pyx_t_4 = __pyx_v_state_bytes;
    } else {
      __Pyx_INCREF(__pyx_kp_b__2);
      __pyx_t_4 = __pyx_kp_b__2;
    }
    __Pyx_DECREF_SET(__pyx_v_state_bytes, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "relstorage/cache/cache.pyx":476
 *             state_bytes = compress(state_bytes) if compress is not None else state_bytes
 *             state_bytes = state_bytes if state_bytes is not None else b''
 *             if len(state_bytes) >= value_limit:             # <<<<<<<<<<<<<<
 *                 # This value is too big, so don't cache it.
 *                 continue
 */
    __pyx_t_13 = PyObject_Length(__pyx_v_state_bytes); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 476, __pyx_L1_error)
    __pyx_t_11 = ((__pyx_t_13 >= __pyx_v_value_limit) != 0);
    if (__pyx_t_11) {

      /* "relstorage/cache/cache.pyx":478
 *             if len(state_bytes) >= value_limit:
 *                 # This value is too big, so don't cache it.
 *                 continue             # <<<<<<<<<<<<<<
 *             self._do_set(oid_int, state_bytes, tid_int)
 * 
 */
      goto __pyx_L3_continue;

      /* "relstorage/cache/cache.pyx":476
 *             state_bytes = compress(state_bytes) if compress is not None else state_bytes
 *             state_bytes = state_bytes if state_bytes is not None else b''
 *             if len(state_bytes) >= value_limit:             # <<<<<<<<<<<<<<
 *                 # This value is too big, so don't cache it.
 *                 continue
 */
    }

    /* "relstorage/cache/cache.pyx":479
 *                 # This value is too big, so don't cache it.
 *                 continue
 *             self._do_set(oid_int, state_bytes, tid_int)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = __pyx_f_10relstorage_5cache_5cache_7PyCache__do_set(__pyx_v_self, __pyx_v_oid_int, __pyx_v_state_bytes, __pyx_v_tid_int); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 479, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "relstorage/cache/cache.pyx":473
 *         # or otherwise not implemented in pure-python
 *         cdef OID_t oid_int
 *         for state_bytes, oid_int, _ in state_oid_iter:             # <<<<<<<<<<<<<<
 *             state_bytes = compress(state_bytes) if compress is not None else state_bytes
 *             state_bytes = state_bytes if state_bytes is not None else b''
 */
    __pyx_L3_continue:;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":469
 *     # Cache specific operations
 * 
 *     cpdef set_all_for_tid(self, TID_t tid_int, state_oid_iter, compress, Py_ssize_t value_limit):             # <<<<<<<<<<<<<<
 *         # Do all this down here so we don't give up the GIL, assuming the iter is actually a list
 *         # or otherwise not implemented in pure-python
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.set_all_for_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state_bytes);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_39set_all_for_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_38set_all_for_tid, "PyCache.set_all_for_tid(self, TID_t tid_int, state_oid_iter, compress, Py_ssize_t value_limit)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_39set_all_for_tid = {"set_all_for_tid", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10relstorage_5cache_5cache_7PyCache_39set_all_for_tid, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_38set_all_for_tid};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_39set_all_for_tid(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  relstorage::cache::TID_t __pyx_v_tid_int;
  PyObject *__pyx_v_state_oid_iter = 0;
  PyObject *__pyx_v_compress = 0;
  Py_ssize_t __pyx_v_value_limit;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_all_for_tid (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tid_int,&__pyx_n_s_state_oid_iter,&__pyx_n_s_compress,&__pyx_n_s_value_limit,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tid_int,&__pyx_n_s_state_oid_iter,&__pyx_n_s_compress,&__pyx_n_s_value_limit,0};
    #endif
    PyObject* values[4] = {0,0,0,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tid_int)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_state_oid_iter)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("set_all_for_tid", 1, 4, 4, 1); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_compress)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("set_all_for_tid", 1, 4, 4, 2); __PYX_ERR(0, 469, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value_limit)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("set_all_for_tid", 1, 4, 4, 3); __PYX_ERR(0, 469, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_all_for_tid") < 0)) __PYX_ERR(0, 469, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
      values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
    }
    __pyx_v_tid_int = __Pyx_PyInt_As_int64_t(values[0]); if (unlikely((__pyx_v_tid_int == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
    __pyx_v_state_oid_iter = values[1];
    __pyx_v_compress = values[2];
    __pyx_v_value_limit = __Pyx_PyIndex_AsSsize_t(values[3]); if (unlikely((__pyx_v_value_limit == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 469, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_all_for_tid", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 469, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.set_all_for_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_38set_all_for_tid(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), __pyx_v_tid_int, __pyx_v_state_oid_iter, __pyx_v_compress, __pyx_v_value_limit);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_38set_all_for_tid(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, relstorage::cache::TID_t __pyx_v_tid_int, PyObject *__pyx_v_state_oid_iter, PyObject *__pyx_v_compress, Py_ssize_t __pyx_v_value_limit) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_all_for_tid", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_10relstorage_5cache_5cache_7PyCache_set_all_for_tid(__pyx_v_self, __pyx_v_tid_int, __pyx_v_state_oid_iter, __pyx_v_compress, __pyx_v_value_limit, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.set_all_for_tid", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":482
 * 
 * 
 *     def add_MRUs(self, ordered_keys, return_count_only=False):             # <<<<<<<<<<<<<<
 *         cdef OID_t key
 *         cdef TID_t tid
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_41add_MRUs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_40add_MRUs, "PyCache.add_MRUs(self, ordered_keys, return_count_only=False)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_41add_MRUs = {"add_MRUs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10relstorage_5cache_5cache_7PyCache_41add_MRUs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_40add_MRUs};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_41add_MRUs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_ordered_keys = 0;
  PyObject *__pyx_v_return_count_only = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED const Py_ssize_t __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_MRUs (wrapper)", 0);
  {
    #if CYTHON_COMPILING_IN_LIMITED_API
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ordered_keys,&__pyx_n_s_return_count_only,0};
    #else
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ordered_keys,&__pyx_n_s_return_count_only,0};
    #endif
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_False);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_ordered_keys)) != 0)) kw_args--;
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_return_count_only);
          if (value) { values[1] = value; kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 482, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_MRUs") < 0)) __PYX_ERR(0, 482, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ordered_keys = values[0];
    __pyx_v_return_count_only = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_MRUs", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 482, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.add_MRUs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_40add_MRUs(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), __pyx_v_ordered_keys, __pyx_v_return_count_only);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_40add_MRUs(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_ordered_keys, PyObject *__pyx_v_return_count_only) {
  relstorage::cache::OID_t __pyx_v_key;
  relstorage::cache::TID_t __pyx_v_tid;
  int __pyx_v_i;
  int __pyx_v_number_nodes;
  relstorage::cache::TempCacheFiller __pyx_v_filler;
  PyObject *__pyx_v_state = NULL;
  PyObject *__pyx_v_frozen = NULL;
  PyObject *__pyx_v_frequency = NULL;
  std::vector<relstorage::cache::OID_t>  __pyx_v_added_oids;
  PyObject *__pyx_v_result = NULL;
  relstorage::cache::OID_t __pyx_7genexpr__pyx_v_key;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  relstorage::cache::TempCacheFiller __pyx_t_4;
  int __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  relstorage::cache::OID_t __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  relstorage::cache::TID_t __pyx_t_17;
  bool __pyx_t_18;
  int __pyx_t_19;
  std::vector<relstorage::cache::OID_t>  __pyx_t_20;
  std::vector<relstorage::cache::OID_t> ::iterator __pyx_t_21;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_MRUs", 0);
  __Pyx_INCREF(__pyx_v_ordered_keys);

  /* "relstorage/cache/cache.pyx":486
 *         cdef TID_t tid
 *         cdef int i
 *         cdef int number_nodes = len(ordered_keys)             # <<<<<<<<<<<<<<
 *         cdef TempCacheFiller filler
 *         if not number_nodes:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_ordered_keys); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 486, __pyx_L1_error)
  __pyx_v_number_nodes = __pyx_t_1;

  /* "relstorage/cache/cache.pyx":488
 *         cdef int number_nodes = len(ordered_keys)
 *         cdef TempCacheFiller filler
 *         if not number_nodes:             # <<<<<<<<<<<<<<
 *             return 0 if return_count_only else ()
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_number_nodes != 0)) != 0);
  if (__pyx_t_2) {

    /* "relstorage/cache/cache.pyx":489
 *         cdef TempCacheFiller filler
 *         if not number_nodes:
 *             return 0 if return_count_only else ()             # <<<<<<<<<<<<<<
 * 
 *         filler = TempCacheFiller(number_nodes)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_return_count_only); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 489, __pyx_L1_error)
    if (__pyx_t_2) {
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_3 = __pyx_int_0;
    } else {
      __Pyx_INCREF(__pyx_empty_tuple);
      __pyx_t_3 = __pyx_empty_tuple;
    }
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":488
 *         cdef int number_nodes = len(ordered_keys)
 *         cdef TempCacheFiller filler
 *         if not number_nodes:             # <<<<<<<<<<<<<<
 *             return 0 if return_count_only else ()
 * 
 */
  }

  /* "relstorage/cache/cache.pyx":491
 *             return 0 if return_count_only else ()
 * 
 *         filler = TempCacheFiller(number_nodes)             # <<<<<<<<<<<<<<
 * 
 *         for i, (key, (state, tid, frozen, frequency)) in enumerate(ordered_keys):
 */
  try {
    __pyx_t_4 = relstorage::cache::TempCacheFiller(__pyx_v_number_nodes);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 491, __pyx_L1_error)
  }
  __pyx_v_filler = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_4);

  /* "relstorage/cache/cache.pyx":493
 *         filler = TempCacheFiller(number_nodes)
 * 
 *         for i, (key, (state, tid, frozen, frequency)) in enumerate(ordered_keys):             # <<<<<<<<<<<<<<
 *             state = state or b''
 *             filler.set_once_at(i, key, tid, state, frozen, frequency)
 */
  __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_v_ordered_keys)) || PyTuple_CheckExact(__pyx_v_ordered_keys)) {
    __pyx_t_3 = __pyx_v_ordered_keys; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_ordered_keys); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 493, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_7); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 493, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_7); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 493, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 493, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 493, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_7))) || (PyList_CheckExact(__pyx_t_7))) {
      PyObject* sequence = __pyx_t_7;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 493, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_10 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 493, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __pyx_t_12 = __Pyx_PyInt_As_int64_t(__pyx_t_8); if (unlikely((__pyx_t_12 == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_v_key = __pyx_t_12;
    if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
      PyObject* sequence = __pyx_t_9;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 4)) {
        if (size > 4) __Pyx_RaiseTooManyValuesError(4);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 493, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_14 = PyTuple_GET_ITEM(sequence, 2); 
        __pyx_t_15 = PyTuple_GET_ITEM(sequence, 3); 
      } else {
        __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_14 = PyList_GET_ITEM(sequence, 2); 
        __pyx_t_15 = PyList_GET_ITEM(sequence, 3); 
      }
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_15);
      #else
      {
        Py_ssize_t i;
        PyObject** temps[4] = {&__pyx_t_10,&__pyx_t_13,&__pyx_t_14,&__pyx_t_15};
        for (i=0; i < 4; i++) {
          PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 493, __pyx_L1_error)
          __Pyx_GOTREF(item);
          *(temps[i]) = item;
        }
      }
      #endif
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else {
      Py_ssize_t index = -1;
      PyObject** temps[4] = {&__pyx_t_10,&__pyx_t_13,&__pyx_t_14,&__pyx_t_15};
      __pyx_t_16 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_16);
      for (index=0; index < 4; index++) {
        PyObject* item = __pyx_t_11(__pyx_t_16); if (unlikely(!item)) goto __pyx_L8_unpacking_failed;
        __Pyx_GOTREF(item);
        *(temps[index]) = item;
      }
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_16), 4) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 493, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    __pyx_t_17 = __Pyx_PyInt_As_int64_t(__pyx_t_13); if (unlikely((__pyx_t_17 == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF_SET(__pyx_v_state, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_v_tid = __pyx_t_17;
    __Pyx_XDECREF_SET(__pyx_v_frozen, __pyx_t_14);
    __pyx_t_14 = 0;
    __Pyx_XDECREF_SET(__pyx_v_frequency, __pyx_t_15);
    __pyx_t_15 = 0;
    __pyx_v_i = __pyx_t_5;
    __pyx_t_5 = (__pyx_t_5 + 1);

    /* "relstorage/cache/cache.pyx":494
 * 
 *         for i, (key, (state, tid, frozen, frequency)) in enumerate(ordered_keys):
 *             state = state or b''             # <<<<<<<<<<<<<<
 *             filler.set_once_at(i, key, tid, state, frozen, frequency)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_state); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 494, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __Pyx_INCREF(__pyx_v_state);
      __pyx_t_7 = __pyx_v_state;
      goto __pyx_L10_bool_binop_done;
    }
    __Pyx_INCREF(__pyx_kp_b__2);
    __pyx_t_7 = __pyx_kp_b__2;
    __pyx_L10_bool_binop_done:;
    __Pyx_DECREF_SET(__pyx_v_state, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "relstorage/cache/cache.pyx":495
 *         for i, (key, (state, tid, frozen, frequency)) in enumerate(ordered_keys):
 *             state = state or b''
 *             filler.set_once_at(i, key, tid, state, frozen, frequency)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_18 = __Pyx_PyObject_IsTrue(__pyx_v_frozen); if (unlikely((__pyx_t_18 == ((bool)-1)) && PyErr_Occurred())) __PYX_ERR(0, 495, __pyx_L1_error)
    __pyx_t_19 = __Pyx_PyInt_As_int(__pyx_v_frequency); if (unlikely((__pyx_t_19 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 495, __pyx_L1_error)
    __pyx_v_filler.set_once_at(__pyx_v_i, __pyx_v_key, __pyx_v_tid, __pyx_v_state, __pyx_t_18, __pyx_t_19);

    /* "relstorage/cache/cache.pyx":493
 *         filler = TempCacheFiller(number_nodes)
 * 
 *         for i, (key, (state, tid, frozen, frequency)) in enumerate(ordered_keys):             # <<<<<<<<<<<<<<
 *             state = state or b''
 *             filler.set_once_at(i, key, tid, state, frozen, frequency)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "relstorage/cache/cache.pyx":499
 * 
 *         # We're done with ordered_keys, free its memory
 *         ordered_keys = None             # <<<<<<<<<<<<<<
 * 
 *         added_oids = self.cache.add_many(filler)
 */
  __Pyx_INCREF(Py_None);
  __Pyx_DECREF_SET(__pyx_v_ordered_keys, Py_None);

  /* "relstorage/cache/cache.pyx":501
 *         ordered_keys = None
 * 
 *         added_oids = self.cache.add_many(filler)             # <<<<<<<<<<<<<<
 * 
 *         # Things that didn't get added have -1 for their generation.
 */
  try {
    __pyx_t_20 = __pyx_v_self->cache.add_many(__pyx_v_filler);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(0, 501, __pyx_L1_error)
  }
  __pyx_v_added_oids = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_20);

  /* "relstorage/cache/cache.pyx":504
 * 
 *         # Things that didn't get added have -1 for their generation.
 *         if return_count_only:             # <<<<<<<<<<<<<<
 *             return added_oids.size()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_return_count_only); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 504, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "relstorage/cache/cache.pyx":505
 *         # Things that didn't get added have -1 for their generation.
 *         if return_count_only:
 *             return added_oids.size()             # <<<<<<<<<<<<<<
 * 
 *         result = [self.get(key) for key in added_oids]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_added_oids.size()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "relstorage/cache/cache.pyx":504
 * 
 *         # Things that didn't get added have -1 for their generation.
 *         if return_count_only:             # <<<<<<<<<<<<<<
 *             return added_oids.size()
 * 
 */
  }

  /* "relstorage/cache/cache.pyx":507
 *             return added_oids.size()
 * 
 *         result = [self.get(key) for key in added_oids]             # <<<<<<<<<<<<<<
 * 
 *         return result
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_21 = __pyx_v_added_oids.begin();
    for (;;) {
      if (!(__pyx_t_21 != __pyx_v_added_oids.end())) break;
      __pyx_t_12 = *__pyx_t_21;
      ++__pyx_t_21;
      __pyx_7genexpr__pyx_v_key = __pyx_t_12;
      __pyx_t_7 = ((PyObject *)__pyx_f_10relstorage_5cache_5cache_7PyCache_get(__pyx_v_self, __pyx_7genexpr__pyx_v_key, 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 507, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_7))) __PYX_ERR(0, 507, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
  } /* exit inner scope */
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "relstorage/cache/cache.pyx":509
 *         result = [self.get(key) for key in added_oids]
 * 
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def age_frequencies(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":482
 * 
 * 
 *     def add_MRUs(self, ordered_keys, return_count_only=False):             # <<<<<<<<<<<<<<
 *         cdef OID_t key
 *         cdef TID_t tid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.add_MRUs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v_frozen);
  __Pyx_XDECREF(__pyx_v_frequency);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_ordered_keys);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":511
 *         return result
 * 
 *     def age_frequencies(self):             # <<<<<<<<<<<<<<
 *         self.cache.age_frequencies()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_43age_frequencies(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_42age_frequencies, "PyCache.age_frequencies(self)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_43age_frequencies = {"age_frequencies", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_43age_frequencies, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_42age_frequencies};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_43age_frequencies(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("age_frequencies (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_42age_frequencies(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_42age_frequencies(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("age_frequencies", 0);

  /* "relstorage/cache/cache.pyx":512
 * 
 *     def age_frequencies(self):
 *         self.cache.age_frequencies()             # <<<<<<<<<<<<<<
 * 
 *     def delitems(self, oids_tids):
 */
  __pyx_v_self->cache.age_frequencies();

  /* "relstorage/cache/cache.pyx":511
 *         return result
 * 
 *     def age_frequencies(self):             # <<<<<<<<<<<<<<
 *         self.cache.age_frequencies()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":514
 *         self.cache.age_frequencies()
 * 
 *     def delitems(self, oids_tids):             # <<<<<<<<<<<<<<
 *         """
 *         For each OID/TID pair in the items, remove all cached values
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_45delitems(PyObject *__pyx_v_self, PyObject *__pyx_v_oids_tids); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_44delitems, "PyCache.delitems(self, oids_tids)\n\n        For each OID/TID pair in the items, remove all cached values\n        for OID that are older than TID.\n        ");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_45delitems = {"delitems", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_45delitems, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_44delitems};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_45delitems(PyObject *__pyx_v_self, PyObject *__pyx_v_oids_tids) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("delitems (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_44delitems(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((PyObject *)__pyx_v_oids_tids));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_44delitems(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_oids_tids) {
  relstorage::cache::OID_t __pyx_v_oid;
  relstorage::cache::TID_t __pyx_v_tid;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  relstorage::cache::OID_t __pyx_t_10;
  relstorage::cache::TID_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("delitems", 0);

  /* "relstorage/cache/cache.pyx":522
 *         cdef TID_t tid
 * 
 *         for oid, tid in iteroiditems(oids_tids):             # <<<<<<<<<<<<<<
 *             self.cache.delitem(oid, tid)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iteroiditems); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_oids_tids};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 522, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 522, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 522, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 522, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 522, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 522, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 522, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 522, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
      index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 522, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 522, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_3); if (unlikely((__pyx_t_10 == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = __Pyx_PyInt_As_int64_t(__pyx_t_7); if (unlikely((__pyx_t_11 == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_oid = __pyx_t_10;
    __pyx_v_tid = __pyx_t_11;

    /* "relstorage/cache/cache.pyx":523
 * 
 *         for oid, tid in iteroiditems(oids_tids):
 *             self.cache.delitem(oid, tid)             # <<<<<<<<<<<<<<
 * 
 *     def del_oids(self, oids):
 */
    try {
      __pyx_v_self->cache.delitem(__pyx_v_oid, __pyx_v_tid);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 523, __pyx_L1_error)
    }

    /* "relstorage/cache/cache.pyx":522
 *         cdef TID_t tid
 * 
 *         for oid, tid in iteroiditems(oids_tids):             # <<<<<<<<<<<<<<
 *             self.cache.delitem(oid, tid)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "relstorage/cache/cache.pyx":514
 *         self.cache.age_frequencies()
 * 
 *     def delitems(self, oids_tids):             # <<<<<<<<<<<<<<
 *         """
 *         For each OID/TID pair in the items, remove all cached values
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.delitems", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":525
 *             self.cache.delitem(oid, tid)
 * 
 *     def del_oids(self, oids):             # <<<<<<<<<<<<<<
 *         """
 *         For each oid in OIDs, remove it.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_47del_oids(PyObject *__pyx_v_self, PyObject *__pyx_v_oids); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_46del_oids, "PyCache.del_oids(self, oids)\n\n        For each oid in OIDs, remove it.\n        ");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_47del_oids = {"del_oids", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_47del_oids, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_46del_oids};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_47del_oids(PyObject *__pyx_v_self, PyObject *__pyx_v_oids) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("del_oids (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_46del_oids(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((PyObject *)__pyx_v_oids));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_46del_oids(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_oids) {
  PyObject *__pyx_v_oid = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  relstorage::cache::OID_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("del_oids", 0);

  /* "relstorage/cache/cache.pyx":529
 *         For each oid in OIDs, remove it.
 *         """
 *         for oid in oids:             # <<<<<<<<<<<<<<
 *             self.cache.delitem(oid)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_oids)) || PyTuple_CheckExact(__pyx_v_oids)) {
    __pyx_t_1 = __pyx_v_oids; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_oids); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 529, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 529, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 529, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_oid, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "relstorage/cache/cache.pyx":530
 *         """
 *         for oid in oids:
 *             self.cache.delitem(oid)             # <<<<<<<<<<<<<<
 * 
 *     def freeze(self, oids_tids):
 */
    __pyx_t_5 = __Pyx_PyInt_As_int64_t(__pyx_v_oid); if (unlikely((__pyx_t_5 == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 530, __pyx_L1_error)
    try {
      __pyx_v_self->cache.delitem(__pyx_t_5);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 530, __pyx_L1_error)
    }

    /* "relstorage/cache/cache.pyx":529
 *         For each oid in OIDs, remove it.
 *         """
 *         for oid in oids:             # <<<<<<<<<<<<<<
 *             self.cache.delitem(oid)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":525
 *             self.cache.delitem(oid, tid)
 * 
 *     def del_oids(self, oids):             # <<<<<<<<<<<<<<
 *         """
 *         For each oid in OIDs, remove it.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.del_oids", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_oid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":532
 *             self.cache.delitem(oid)
 * 
 *     def freeze(self, oids_tids):             # <<<<<<<<<<<<<<
 *         # The idea is to *move* the data, or make it available,
 *         # *without* copying it.
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_49freeze(PyObject *__pyx_v_self, PyObject *__pyx_v_oids_tids); /*proto*/
PyDoc_STRVAR(__pyx_doc_10relstorage_5cache_5cache_7PyCache_48freeze, "PyCache.freeze(self, oids_tids)");
static PyMethodDef __pyx_mdef_10relstorage_5cache_5cache_7PyCache_49freeze = {"freeze", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_49freeze, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_48freeze};
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_49freeze(PyObject *__pyx_v_self, PyObject *__pyx_v_oids_tids) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("freeze (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_48freeze(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self), ((PyObject *)__pyx_v_oids_tids));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_48freeze(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self, PyObject *__pyx_v_oids_tids) {
  relstorage::cache::OID_t __pyx_v_oid;
  relstorage::cache::TID_t __pyx_v_tid;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  relstorage::cache::OID_t __pyx_t_10;
  relstorage::cache::TID_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("freeze", 0);

  /* "relstorage/cache/cache.pyx":538
 *         cdef TID_t tid
 * 
 *         for oid, tid in iteroiditems(oids_tids):             # <<<<<<<<<<<<<<
 *             self.cache.freeze(oid, tid)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_iteroiditems); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_oids_tids};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 538, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 538, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 538, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 538, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 538, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 538, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8);
      index = 0; __pyx_t_3 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_7);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
      __pyx_t_9 = NULL;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 538, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_3); if (unlikely((__pyx_t_10 == ((relstorage::cache::OID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_11 = __Pyx_PyInt_As_int64_t(__pyx_t_7); if (unlikely((__pyx_t_11 == ((relstorage::cache::TID_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_oid = __pyx_t_10;
    __pyx_v_tid = __pyx_t_11;

    /* "relstorage/cache/cache.pyx":539
 * 
 *         for oid, tid in iteroiditems(oids_tids):
 *             self.cache.freeze(oid, tid)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
    try {
      __pyx_v_self->cache.freeze(__pyx_v_oid, __pyx_v_tid);
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 539, __pyx_L1_error)
    }

    /* "relstorage/cache/cache.pyx":538
 *         cdef TID_t tid
 * 
 *         for oid, tid in iteroiditems(oids_tids):             # <<<<<<<<<<<<<<
 *             self.cache.freeze(oid, tid)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "relstorage/cache/cache.pyx":532
 *             self.cache.delitem(oid)
 * 
 *     def freeze(self, oids_tids):             # <<<<<<<<<<<<<<
 *         # The idea is to *move* the data, or make it available,
 *         # *without* copying it.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.freeze", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":541
 *             self.cache.freeze(oid, tid)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.cache.weight()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_6weight_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_6weight_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_6weight___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_6weight___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "relstorage/cache/cache.pyx":543
 *     @property
 *     def weight(self):
 *         return self.cache.weight()             # <<<<<<<<<<<<<<
 * 
 * # Local Variables:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->cache.weight()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "relstorage/cache/cache.pyx":541
 *             self.cache.freeze(oid, tid)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def weight(self):
 *         return self.cache.weight()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.weight.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":291
 * cdef class PyCache:
 *     cdef Cache cache
 *     cdef readonly size_t sets             # <<<<<<<<<<<<<<
 *     cdef readonly size_t hits
 *     cdef readonly size_t misses
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_4sets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_4sets_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_4sets___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_4sets___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->sets); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.sets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":292
 *     cdef Cache cache
 *     cdef readonly size_t sets
 *     cdef readonly size_t hits             # <<<<<<<<<<<<<<
 *     cdef readonly size_t misses
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_4hits_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_4hits_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_4hits___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_4hits___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->hits); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.hits.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "relstorage/cache/cache.pyx":293
 *     cdef readonly size_t sets
 *     cdef readonly size_t hits
 *     cdef readonly size_t misses             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, eden, protected, probation):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_6misses_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10relstorage_5cache_5cache_7PyCache_6misses_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_10relstorage_5cache_5cache_7PyCache_6misses___get__(((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10relstorage_5cache_5cache_7PyCache_6misses___get__(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_self->misses); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("relstorage.cache.cache.PyCache.misses.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t")
 * cdef object __pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

static PyObject *__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t(const std::vector<relstorage::cache::TID_t>  &__pyx_v_v) {
  size_t __pyx_7genexpr__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t", 0);

  /* "vector.to_py":61
 * @cname("__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t")
 * cdef object __pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t(vector[X]& v):
 *     return [v[i] for i in range(v.size())]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_v.size();
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_7genexpr__pyx_v_i = __pyx_t_4;
      __pyx_t_5 = __Pyx_PyInt_From_int64_t((__pyx_v_v[__pyx_7genexpr__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 61, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 61, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t")
 * cdef object __pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("vector.to_py.__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_CachedValue(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache_CachedValue(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache_CachedValue_slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_doc, (void *)PyDoc_STR("\n    The base class for cached values.\n    ")},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache_CachedValue},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache_CachedValue_spec = {
  "relstorage.cache.cache.CachedValue",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_CachedValue),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_10relstorage_5cache_5cache_CachedValue_slots,
};
#else

static PyTypeObject __pyx_type_10relstorage_5cache_5cache_CachedValue = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""CachedValue", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_CachedValue), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache_CachedValue, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  PyDoc_STR("\n    The base class for cached values.\n    "), /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache_CachedValue, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_SingleValue __pyx_vtable_10relstorage_5cache_5cache_SingleValue;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_SingleValue(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *p;
  #if CYTHON_COMPILING_IN_LIMITED_API
  newfunc new_func = (newfunc)PyType_GetSlot(__pyx_ptype_10relstorage_5cache_5cache_CachedValue, Py_tp_new);
  PyObject *o = new_func(t, a, k);
  #else
  PyObject *o = __pyx_tp_new_10relstorage_5cache_5cache_CachedValue(t, a, k);
  #endif
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *)o);
  p->__pyx_vtab = __pyx_vtabptr_10relstorage_5cache_5cache_SingleValue;
  if (unlikely(__pyx_pw_10relstorage_5cache_5cache_11SingleValue_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_10relstorage_5cache_5cache_SingleValue(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10relstorage_5cache_5cache_11SingleValue_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __pyx_tp_dealloc_10relstorage_5cache_5cache_CachedValue(o);
}
static PyObject *__pyx_sq_item_10relstorage_5cache_5cache_SingleValue(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_tp_richcompare_10relstorage_5cache_5cache_SingleValue(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_9__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_10relstorage_5cache_5cache_11SingleValue_9__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret); Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_frozen(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_6frozen_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_5value_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_3key_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_frequency(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_9frequency_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_state(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_5state_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_tid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_3tid_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_max_tid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_7max_tid_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_newest_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_12newest_value_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_11SingleValue_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_11SingleValue_6weight_1__get__(o);
}

static PyMethodDef __pyx_methods_10relstorage_5cache_5cache_SingleValue[] = {
  {"sizeof", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_11SingleValue_5sizeof, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_11SingleValue_4sizeof},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10relstorage_5cache_5cache_SingleValue[] = {
  {(char *)"frozen", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_frozen, 0, (char *)0, 0},
  {(char *)"value", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_value, 0, (char *)0, 0},
  {(char *)"key", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_key, 0, (char *)0, 0},
  {(char *)"frequency", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_frequency, 0, (char *)0, 0},
  {(char *)"state", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_state, 0, (char *)0, 0},
  {(char *)"tid", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_tid, 0, (char *)0, 0},
  {(char *)"max_tid", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_max_tid, 0, (char *)0, 0},
  {(char *)"newest_value", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_newest_value, 0, (char *)0, 0},
  {(char *)"weight", __pyx_getprop_10relstorage_5cache_5cache_11SingleValue_weight, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache_SingleValue_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10relstorage_5cache_5cache_SingleValue},
  {Py_tp_repr, (void *)__pyx_pw_10relstorage_5cache_5cache_11SingleValue_13__repr__},
  {Py_sq_item, (void *)__pyx_sq_item_10relstorage_5cache_5cache_SingleValue},
  {Py_mp_subscript, (void *)__pyx_pw_10relstorage_5cache_5cache_11SingleValue_11__getitem__},
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_10relstorage_5cache_5cache_SingleValue},
  {Py_tp_iter, (void *)__pyx_pw_10relstorage_5cache_5cache_11SingleValue_7__iter__},
  {Py_tp_methods, (void *)__pyx_methods_10relstorage_5cache_5cache_SingleValue},
  {Py_tp_getset, (void *)__pyx_getsets_10relstorage_5cache_5cache_SingleValue},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache_SingleValue},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache_SingleValue_spec = {
  "relstorage.cache.cache.SingleValue",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_10relstorage_5cache_5cache_SingleValue_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_SingleValue = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10relstorage_5cache_5cache_SingleValue, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_SingleValue = {
  0, /*mp_length*/
  __pyx_pw_10relstorage_5cache_5cache_11SingleValue_11__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10relstorage_5cache_5cache_SingleValue = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""SingleValue", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_SingleValue), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache_SingleValue, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10relstorage_5cache_5cache_11SingleValue_13__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_SingleValue, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_SingleValue, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_tp_richcompare_10relstorage_5cache_5cache_SingleValue, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_10relstorage_5cache_5cache_11SingleValue_7__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10relstorage_5cache_5cache_SingleValue, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10relstorage_5cache_5cache_SingleValue, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache_SingleValue, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_MultipleValues __pyx_vtable_10relstorage_5cache_5cache_MultipleValues;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_MultipleValues(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *p;
  #if CYTHON_COMPILING_IN_LIMITED_API
  newfunc new_func = (newfunc)PyType_GetSlot(__pyx_ptype_10relstorage_5cache_5cache_CachedValue, Py_tp_new);
  PyObject *o = new_func(t, a, k);
  #else
  PyObject *o = __pyx_tp_new_10relstorage_5cache_5cache_CachedValue(t, a, k);
  #endif
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *)o);
  p->__pyx_vtab = __pyx_vtabptr_10relstorage_5cache_5cache_MultipleValues;
  return o;
}

static void __pyx_tp_dealloc_10relstorage_5cache_5cache_MultipleValues(PyObject *o) {
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_1__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  __pyx_tp_dealloc_10relstorage_5cache_5cache_CachedValue(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_5value_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_3key_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_frequency(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_9frequency_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_6weight_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_max_tid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_7max_tid_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_newest_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_14MultipleValues_12newest_value_1__get__(o);
}

static PyMethodDef __pyx_methods_10relstorage_5cache_5cache_MultipleValues[] = {
  {"sizeof", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_14MultipleValues_3sizeof, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_14MultipleValues_2sizeof},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10relstorage_5cache_5cache_MultipleValues[] = {
  {(char *)"value", __pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_value, 0, (char *)0, 0},
  {(char *)"key", __pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_key, 0, (char *)0, 0},
  {(char *)"frequency", __pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_frequency, 0, (char *)0, 0},
  {(char *)"weight", __pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_weight, 0, (char *)0, 0},
  {(char *)"max_tid", __pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_max_tid, 0, (char *)0, 0},
  {(char *)"newest_value", __pyx_getprop_10relstorage_5cache_5cache_14MultipleValues_newest_value, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache_MultipleValues_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10relstorage_5cache_5cache_MultipleValues},
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_methods, (void *)__pyx_methods_10relstorage_5cache_5cache_MultipleValues},
  {Py_tp_getset, (void *)__pyx_getsets_10relstorage_5cache_5cache_MultipleValues},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache_MultipleValues},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache_MultipleValues_spec = {
  "relstorage.cache.cache.MultipleValues",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_10relstorage_5cache_5cache_MultipleValues_slots,
};
#else

static PyTypeObject __pyx_type_10relstorage_5cache_5cache_MultipleValues = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""MultipleValues", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache_MultipleValues, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10relstorage_5cache_5cache_MultipleValues, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10relstorage_5cache_5cache_MultipleValues, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache_MultipleValues, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyGeneration __pyx_vtable_10relstorage_5cache_5cache_PyGeneration;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_PyGeneration(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)o);
  p->__pyx_vtab = __pyx_vtabptr_10relstorage_5cache_5cache_PyGeneration;
  p->__name__ = Py_None; Py_INCREF(Py_None);
  p->_cache = ((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)Py_None); Py_INCREF(Py_None);
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache_PyGeneration(PyObject *o) {
  struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *p = (struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__name__);
  Py_CLEAR(p->_cache);
  (*Py_TYPE(o)->tp_free)(o);
}
#endif

static int __pyx_tp_traverse_10relstorage_5cache_5cache_PyGeneration(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *p = (struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)o;
  if (p->__name__) {
    e = (*v)(p->__name__, a); if (e) return e;
  }
  if (p->_cache) {
    e = (*v)(((PyObject *)p->_cache), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10relstorage_5cache_5cache_PyGeneration(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *p = (struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration *)o;
  tmp = ((PyObject*)p->__name__);
  p->__name__ = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_cache);
  p->_cache = ((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_12PyGeneration_generation_number(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_17generation_number_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_12PyGeneration_limit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_5limit_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_12PyGeneration_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_6weight_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_12PyGeneration___name__(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_8__name___1__get__(o);
}

static PyMethodDef __pyx_methods_10relstorage_5cache_5cache_PyGeneration[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10relstorage_5cache_5cache_PyGeneration[] = {
  {(char *)"generation_number", __pyx_getprop_10relstorage_5cache_5cache_12PyGeneration_generation_number, 0, (char *)0, 0},
  {(char *)"limit", __pyx_getprop_10relstorage_5cache_5cache_12PyGeneration_limit, 0, (char *)0, 0},
  {(char *)"weight", __pyx_getprop_10relstorage_5cache_5cache_12PyGeneration_weight, 0, (char *)0, 0},
  {(char *)"__name__", __pyx_getprop_10relstorage_5cache_5cache_12PyGeneration___name__, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache_PyGeneration_slots[] = {
  {Py_nb_nonzero, (void *)__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_3__nonzero__},
  {Py_sq_length, (void *)__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_1__len__},
  {Py_mp_length, (void *)__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_1__len__},
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10relstorage_5cache_5cache_PyGeneration},
  {Py_tp_clear, (void *)__pyx_tp_clear_10relstorage_5cache_5cache_PyGeneration},
  {Py_tp_iter, (void *)__pyx_pw_10relstorage_5cache_5cache_12PyGeneration_5__iter__},
  {Py_tp_methods, (void *)__pyx_methods_10relstorage_5cache_5cache_PyGeneration},
  {Py_tp_getset, (void *)__pyx_getsets_10relstorage_5cache_5cache_PyGeneration},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache_PyGeneration},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache_PyGeneration_spec = {
  "relstorage.cache.cache.PyGeneration",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10relstorage_5cache_5cache_PyGeneration_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_PyGeneration = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_3__nonzero__, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_PyGeneration = {
  __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_1__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_PyGeneration = {
  __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_1__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10relstorage_5cache_5cache_PyGeneration = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""PyGeneration", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_PyGeneration), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache_PyGeneration, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  &__pyx_tp_as_number_PyGeneration, /*tp_as_number*/
  &__pyx_tp_as_sequence_PyGeneration, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_PyGeneration, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10relstorage_5cache_5cache_PyGeneration, /*tp_traverse*/
  __pyx_tp_clear_10relstorage_5cache_5cache_PyGeneration, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_10relstorage_5cache_5cache_12PyGeneration_5__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10relstorage_5cache_5cache_PyGeneration, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10relstorage_5cache_5cache_PyGeneration, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache_PyGeneration, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10relstorage_5cache_5cache_PyCache __pyx_vtable_10relstorage_5cache_5cache_PyCache;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache_PyCache(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache_PyCache *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)o);
  p->__pyx_vtab = __pyx_vtabptr_10relstorage_5cache_5cache_PyCache;
  new((void*)&(p->cache)) relstorage::cache::Cache();
  if (unlikely(__pyx_pw_10relstorage_5cache_5cache_7PyCache_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache_PyCache(PyObject *o) {
  struct __pyx_obj_10relstorage_5cache_5cache_PyCache *p = (struct __pyx_obj_10relstorage_5cache_5cache_PyCache *)o;
  __Pyx_call_destructor(p->cache);
  (*Py_TYPE(o)->tp_free)(o);
}
#endif
static PyObject *__pyx_sq_item_10relstorage_5cache_5cache_PyCache(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_10relstorage_5cache_5cache_PyCache(PyObject *o, PyObject *i, PyObject *v) {
  __Pyx_TypeName o_type_name;
  if (v) {
    return __pyx_pw_10relstorage_5cache_5cache_7PyCache_23__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_10relstorage_5cache_5cache_7PyCache_25__delitem__(o, i);
  }
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_limit(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_5limit_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_eden(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_4eden_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_protected(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_9protected_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_probation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_9probation_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_generations(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_11generations_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_weight(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_6weight_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_sets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_4sets_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_hits(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_4hits_1__get__(o);
}

static PyObject *__pyx_getprop_10relstorage_5cache_5cache_7PyCache_misses(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10relstorage_5cache_5cache_7PyCache_6misses_1__get__(o);
}

static PyMethodDef __pyx_methods_10relstorage_5cache_5cache_PyCache[] = {
  {"iteritems", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_30iteritems, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_29iteritems},
  {"keys", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_33keys, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_32keys},
  {"values", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_36values, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_35values},
  {"add_MRUs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10relstorage_5cache_5cache_7PyCache_41add_MRUs, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_40add_MRUs},
  {"age_frequencies", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_43age_frequencies, METH_NOARGS, __pyx_doc_10relstorage_5cache_5cache_7PyCache_42age_frequencies},
  {"delitems", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_45delitems, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_44delitems},
  {"del_oids", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_47del_oids, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_46del_oids},
  {"freeze", (PyCFunction)__pyx_pw_10relstorage_5cache_5cache_7PyCache_49freeze, METH_O, __pyx_doc_10relstorage_5cache_5cache_7PyCache_48freeze},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10relstorage_5cache_5cache_PyCache[] = {
  {(char *)"limit", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_limit, 0, (char *)0, 0},
  {(char *)"eden", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_eden, 0, (char *)0, 0},
  {(char *)"protected", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_protected, 0, (char *)0, 0},
  {(char *)"probation", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_probation, 0, (char *)0, 0},
  {(char *)"generations", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_generations, 0, (char *)0, 0},
  {(char *)"weight", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_weight, 0, (char *)0, 0},
  {(char *)"sets", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_sets, 0, (char *)0, 0},
  {(char *)"hits", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_hits, 0, (char *)0, 0},
  {(char *)"misses", __pyx_getprop_10relstorage_5cache_5cache_7PyCache_misses, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache_PyCache_slots[] = {
  {Py_nb_nonzero, (void *)__pyx_pw_10relstorage_5cache_5cache_7PyCache_5__nonzero__},
  {Py_sq_length, (void *)__pyx_pw_10relstorage_5cache_5cache_7PyCache_9__len__},
  {Py_sq_item, (void *)__pyx_sq_item_10relstorage_5cache_5cache_PyCache},
  {Py_sq_contains, (void *)__pyx_pw_10relstorage_5cache_5cache_7PyCache_7__contains__},
  {Py_mp_length, (void *)__pyx_pw_10relstorage_5cache_5cache_7PyCache_9__len__},
  {Py_mp_subscript, (void *)__pyx_pw_10relstorage_5cache_5cache_7PyCache_19__getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_10relstorage_5cache_5cache_PyCache},
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_iter, (void *)__pyx_pw_10relstorage_5cache_5cache_7PyCache_27__iter__},
  {Py_tp_methods, (void *)__pyx_methods_10relstorage_5cache_5cache_PyCache},
  {Py_tp_getset, (void *)__pyx_getsets_10relstorage_5cache_5cache_PyCache},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache_PyCache},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache_PyCache_spec = {
  "relstorage.cache.cache.PyCache",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_PyCache),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER,
  __pyx_type_10relstorage_5cache_5cache_PyCache_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_PyCache = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  __pyx_pw_10relstorage_5cache_5cache_7PyCache_5__nonzero__, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_PyCache = {
  __pyx_pw_10relstorage_5cache_5cache_7PyCache_9__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_10relstorage_5cache_5cache_PyCache, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_10relstorage_5cache_5cache_7PyCache_7__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_PyCache = {
  __pyx_pw_10relstorage_5cache_5cache_7PyCache_9__len__, /*mp_length*/
  __pyx_pw_10relstorage_5cache_5cache_7PyCache_19__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_10relstorage_5cache_5cache_PyCache, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_10relstorage_5cache_5cache_PyCache = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""PyCache", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache_PyCache), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache_PyCache, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  &__pyx_tp_as_number_PyCache, /*tp_as_number*/
  &__pyx_tp_as_sequence_PyCache, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_PyCache, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_10relstorage_5cache_5cache_7PyCache_27__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10relstorage_5cache_5cache_PyCache, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10relstorage_5cache_5cache_PyCache, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache_PyCache, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct____iter__[8];
static int __pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct____iter__[--__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *)o);
  new((void*)&(p->__pyx_v_it)) relstorage::cache::Generation::iterator();
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->__pyx_v_it);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__)))) {
    __pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct____iter__[__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct____iter__++] = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#endif

static int __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter___slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct____iter__},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct____iter__},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter___spec = {
  "relstorage.cache.cache.__pyx_scope_struct____iter__",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter___slots,
};
#else

static PyTypeObject __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct____iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__[8];
static int __pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ = 0;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__)))) {
    o = (PyObject*)__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__[--__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *)o);
  new((void*)&(p->__pyx_v_end)) relstorage::cache::Cache::iterator();
  new((void*)&(p->__pyx_v_it)) relstorage::cache::Cache::iterator();
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__(PyObject *o) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *)o;
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->__pyx_v_end);
  __Pyx_call_destructor(p->__pyx_v_it);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_t);
  Py_CLEAR(p->__pyx_v_tids);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__)))) {
    __pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__[__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__++] = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#endif

static int __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_t) {
    e = (*v)(p->__pyx_v_t, a); if (e) return e;
  }
  if (p->__pyx_v_tids) {
    e = (*v)(p->__pyx_v_tids, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter___slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter___spec = {
  "relstorage.cache.cache.__pyx_scope_struct_1___iter__",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter___slots,
};
#else

static PyTypeObject __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""__pyx_scope_struct_1___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems[8];
static int __pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems = 0;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems)))) {
    o = (PyObject*)__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems[--__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems];
    memset(o, 0, sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *)o);
  new((void*)&(p->__pyx_v_end)) relstorage::cache::Cache::iterator();
  new((void*)&(p->__pyx_v_it)) relstorage::cache::Cache::iterator();
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems(PyObject *o) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *)o;
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->__pyx_v_end);
  __Pyx_call_destructor(p->__pyx_v_it);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems)))) {
    __pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems[__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems++] = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#endif

static int __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems_slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems_spec = {
  "relstorage.cache.cache.__pyx_scope_struct_2_iteritems",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems_slots,
};
#else

static PyTypeObject __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""__pyx_scope_struct_2_iteritems", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_3_keys[8];
static int __pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_3_keys = 0;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_3_keys(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_3_keys > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys)))) {
    o = (PyObject*)__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_3_keys[--__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_3_keys];
    memset(o, 0, sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *)o);
  new((void*)&(p->__pyx_v_end)) relstorage::cache::Cache::iterator();
  new((void*)&(p->__pyx_v_it)) relstorage::cache::Cache::iterator();
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_3_keys(PyObject *o) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *)o;
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->__pyx_v_end);
  __Pyx_call_destructor(p->__pyx_v_it);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_3_keys < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys)))) {
    __pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_3_keys[__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_3_keys++] = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#endif

static int __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_3_keys(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys_slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_3_keys},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_3_keys},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys_spec = {
  "relstorage.cache.cache.__pyx_scope_struct_3_keys",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys_slots,
};
#else

static PyTypeObject __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""__pyx_scope_struct_3_keys", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_3_keys), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_3_keys, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_3_keys, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_3_keys, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_4_values[8];
static int __pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_4_values = 0;

static PyObject *__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_4_values(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_4_values > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values)))) {
    o = (PyObject*)__pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_4_values[--__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_4_values];
    memset(o, 0, sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  p = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *)o);
  new((void*)&(p->__pyx_v_end)) relstorage::cache::Cache::iterator();
  new((void*)&(p->__pyx_v_it)) relstorage::cache::Cache::iterator();
  return o;
}

#if !CYTHON_COMPILING_IN_LIMITED_API
static void __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_4_values(PyObject *o) {
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *)o;
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->__pyx_v_end);
  __Pyx_call_destructor(p->__pyx_v_it);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_4_values < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values)))) {
    __pyx_freelist_10relstorage_5cache_5cache___pyx_scope_struct_4_values[__pyx_freecount_10relstorage_5cache_5cache___pyx_scope_struct_4_values++] = ((struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}
#endif

static int __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_4_values(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *p = (struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values_slots[] = {
  {Py_tp_getattro, (void *)PyObject_GenericGetAttr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_4_values},
  {Py_tp_new, (void *)__pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_4_values},
  {0, 0},
};
static PyType_Spec __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values_spec = {
  "relstorage.cache.cache.__pyx_scope_struct_4_values",
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values_slots,
};
#else

static PyTypeObject __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values = {
  PyVarObject_HEAD_INIT(0, 0)
  "relstorage.cache.cache.""__pyx_scope_struct_4_values", /*tp_name*/
  sizeof(struct __pyx_obj_10relstorage_5cache_5cache___pyx_scope_struct_4_values), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10relstorage_5cache_5cache___pyx_scope_struct_4_values, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10relstorage_5cache_5cache___pyx_scope_struct_4_values, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10relstorage_5cache_5cache___pyx_scope_struct_4_values, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  #if CYTHON_COMPILING_IN_LIMITED_API
  {0, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {0, __pyx_k_CacheConsistencyError, sizeof(__pyx_k_CacheConsistencyError), 0, 0, 1, 1},
  {0, __pyx_k_CachedValue, sizeof(__pyx_k_CachedValue), 0, 0, 1, 1},
  {0, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {0, __pyx_k_Invalid_pointer_cast, sizeof(__pyx_k_Invalid_pointer_cast), 0, 0, 1, 0},
  {0, __pyx_k_MultipleValues, sizeof(__pyx_k_MultipleValues), 0, 0, 1, 1},
  {0, __pyx_k_MultipleValues_sizeof, sizeof(__pyx_k_MultipleValues_sizeof), 0, 0, 1, 1},
  {0, __pyx_k_NoSuchGeneration, sizeof(__pyx_k_NoSuchGeneration), 0, 0, 1, 1},
  {0, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {0, __pyx_k_PyCache, sizeof(__pyx_k_PyCache), 0, 0, 1, 1},
  {0, __pyx_k_PyCache___iter, sizeof(__pyx_k_PyCache___iter), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_add_MRUs, sizeof(__pyx_k_PyCache_add_MRUs), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_age_frequencies, sizeof(__pyx_k_PyCache_age_frequencies), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_contains_oid_with_tid, sizeof(__pyx_k_PyCache_contains_oid_with_tid), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_del_oids, sizeof(__pyx_k_PyCache_del_oids), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_delitems, sizeof(__pyx_k_PyCache_delitems), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_freeze, sizeof(__pyx_k_PyCache_freeze), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_get, sizeof(__pyx_k_PyCache_get), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_get_item_with_tid, sizeof(__pyx_k_PyCache_get_item_with_tid), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_iteritems, sizeof(__pyx_k_PyCache_iteritems), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_keys, sizeof(__pyx_k_PyCache_keys), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_peek, sizeof(__pyx_k_PyCache_peek), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_peek_item_with_tid, sizeof(__pyx_k_PyCache_peek_item_with_tid), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_reset_stats, sizeof(__pyx_k_PyCache_reset_stats), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_set_all_for_tid, sizeof(__pyx_k_PyCache_set_all_for_tid), 0, 0, 1, 1},
  {0, __pyx_k_PyCache_values, sizeof(__pyx_k_PyCache_values), 0, 0, 1, 1},
  {0, __pyx_k_PyGeneration, sizeof(__pyx_k_PyGeneration), 0, 0, 1, 1},
  {0, __pyx_k_PyGeneration___iter, sizeof(__pyx_k_PyGeneration___iter), 0, 0, 1, 1},
  {0, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {0, __pyx_k_SingleValue_sizeof, sizeof(__pyx_k_SingleValue_sizeof), 0, 0, 1, 1},
  {0, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 0, 0},
  {0, __pyx_k__38, sizeof(__pyx_k__38), 0, 0, 1, 1},
  {0, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 1},
  {0, __pyx_k_add_MRUs, sizeof(__pyx_k_add_MRUs), 0, 0, 1, 1},
  {0, __pyx_k_added_oids, sizeof(__pyx_k_added_oids), 0, 0, 1, 1},
  {0, __pyx_k_age_frequencies, sizeof(__pyx_k_age_frequencies), 0, 0, 1, 1},
  {0, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {0, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {0, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {0, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {0, __pyx_k_compress, sizeof(__pyx_k_compress), 0, 0, 1, 1},
  {0, __pyx_k_contains_oid_with_tid, sizeof(__pyx_k_contains_oid_with_tid), 0, 0, 1, 1},
  {0, __pyx_k_del_oids, sizeof(__pyx_k_del_oids), 0, 0, 1, 1},
  {0, __pyx_k_delitems, sizeof(__pyx_k_delitems), 0, 0, 1, 1},
  {0, __pyx_k_eden, sizeof(__pyx_k_eden), 0, 0, 1, 1},
  {0, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {0, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {0, __pyx_k_filler, sizeof(__pyx_k_filler), 0, 0, 1, 1},
  {0, __pyx_k_freeze, sizeof(__pyx_k_freeze), 0, 0, 1, 1},
  {0, __pyx_k_frequency, sizeof(__pyx_k_frequency), 0, 0, 1, 1},
  {0, __pyx_k_frozen, sizeof(__pyx_k_frozen), 0, 0, 1, 1},
  {0, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {0, __pyx_k_get_item_with_tid, sizeof(__pyx_k_get_item_with_tid), 0, 0, 1, 1},
  {0, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {0, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {0, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {0, __pyx_k_it, sizeof(__pyx_k_it), 0, 0, 1, 1},
  {0, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {0, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {0, __pyx_k_iteroiditems, sizeof(__pyx_k_iteroiditems), 0, 0, 1, 1},
  {0, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {0, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {0, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {0, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {0, __pyx_k_number_nodes, sizeof(__pyx_k_number_nodes), 0, 0, 1, 1},
  {0, __pyx_k_oid, sizeof(__pyx_k_oid), 0, 0, 1, 1},
  {0, __pyx_k_oids, sizeof(__pyx_k_oids), 0, 0, 1, 1},
  {0, __pyx_k_oids_tids, sizeof(__pyx_k_oids_tids), 0, 0, 1, 1},
  {0, __pyx_k_ordered_keys, sizeof(__pyx_k_ordered_keys), 0, 0, 1, 1},
  {0, __pyx_k_peek, sizeof(__pyx_k_peek), 0, 0, 1, 1},
  {0, __pyx_k_peek_item_with_tid, sizeof(__pyx_k_peek_item_with_tid), 0, 0, 1, 1},
  {0, __pyx_k_probation, sizeof(__pyx_k_probation), 0, 0, 1, 1},
  {0, __pyx_k_protected, sizeof(__pyx_k_protected), 0, 0, 1, 1},
  {0, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {0, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {0, __pyx_k_relstorage__compat, sizeof(__pyx_k_relstorage__compat), 0, 0, 1, 1},
  {0, __pyx_k_relstorage_cache_cache, sizeof(__pyx_k_relstorage_cache_cache), 0, 0, 1, 1},
  {0, __pyx_k_relstorage_cache_interfaces, sizeof(__pyx_k_relstorage_cache_interfaces), 0, 0, 1, 1},
  {0, __pyx_k_reset_stats, sizeof(__pyx_k_reset_stats), 0, 0, 1, 1},
  {0, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {0, __pyx_k_return_count_only, sizeof(__pyx_k_return_count_only), 0, 0, 1, 1},
  {0, __pyx_k_s_r_s_frozen_s, sizeof(__pyx_k_s_r_s_frozen_s), 0, 0, 1, 0},
  {0, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {0, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {0, __pyx_k_set_all_for_tid, sizeof(__pyx_k_set_all_for_tid), 0, 0, 1, 1},
  {0, __pyx_k_sizeof, sizeof(__pyx_k_sizeof), 0, 0, 1, 1},
  {0, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {0, __pyx_k_src_relstorage_cache_cache_pyx, sizeof(__pyx_k_src_relstorage_cache_cache_pyx), 0, 0, 1, 0},
  {0, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {0, __pyx_k_state_oid_iter, sizeof(__pyx_k_state_oid_iter), 0, 0, 1, 1},
  {0, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {0, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {0, __pyx_k_tid, sizeof(__pyx_k_tid), 0, 0, 1, 1},
  {0, __pyx_k_tid_int, sizeof(__pyx_k_tid_int), 0, 0, 1, 1},
  {0, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {0, __pyx_k_value_limit, sizeof(__pyx_k_value_limit), 0, 0, 1, 1},
  {0, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  #else
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_n_s_CacheConsistencyError, __pyx_k_CacheConsistencyError, sizeof(__pyx_k_CacheConsistencyError), 0, 0, 1, 1},
  {&__pyx_n_s_CachedValue, __pyx_k_CachedValue, sizeof(__pyx_k_CachedValue), 0, 0, 1, 1},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Invalid_pointer_cast, __pyx_k_Invalid_pointer_cast, sizeof(__pyx_k_Invalid_pointer_cast), 0, 0, 1, 0},
  {&__pyx_n_s_MultipleValues, __pyx_k_MultipleValues, sizeof(__pyx_k_MultipleValues), 0, 0, 1, 1},
  {&__pyx_n_s_MultipleValues_sizeof, __pyx_k_MultipleValues_sizeof, sizeof(__pyx_k_MultipleValues_sizeof), 0, 0, 1, 1},
  {&__pyx_n_s_NoSuchGeneration, __pyx_k_NoSuchGeneration, sizeof(__pyx_k_NoSuchGeneration), 0, 0, 1, 1},
  {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache, __pyx_k_PyCache, sizeof(__pyx_k_PyCache), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache___iter, __pyx_k_PyCache___iter, sizeof(__pyx_k_PyCache___iter), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_add_MRUs, __pyx_k_PyCache_add_MRUs, sizeof(__pyx_k_PyCache_add_MRUs), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_age_frequencies, __pyx_k_PyCache_age_frequencies, sizeof(__pyx_k_PyCache_age_frequencies), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_contains_oid_with_tid, __pyx_k_PyCache_contains_oid_with_tid, sizeof(__pyx_k_PyCache_contains_oid_with_tid), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_del_oids, __pyx_k_PyCache_del_oids, sizeof(__pyx_k_PyCache_del_oids), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_delitems, __pyx_k_PyCache_delitems, sizeof(__pyx_k_PyCache_delitems), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_freeze, __pyx_k_PyCache_freeze, sizeof(__pyx_k_PyCache_freeze), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_get, __pyx_k_PyCache_get, sizeof(__pyx_k_PyCache_get), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_get_item_with_tid, __pyx_k_PyCache_get_item_with_tid, sizeof(__pyx_k_PyCache_get_item_with_tid), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_iteritems, __pyx_k_PyCache_iteritems, sizeof(__pyx_k_PyCache_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_keys, __pyx_k_PyCache_keys, sizeof(__pyx_k_PyCache_keys), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_peek, __pyx_k_PyCache_peek, sizeof(__pyx_k_PyCache_peek), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_peek_item_with_tid, __pyx_k_PyCache_peek_item_with_tid, sizeof(__pyx_k_PyCache_peek_item_with_tid), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_reset_stats, __pyx_k_PyCache_reset_stats, sizeof(__pyx_k_PyCache_reset_stats), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_set_all_for_tid, __pyx_k_PyCache_set_all_for_tid, sizeof(__pyx_k_PyCache_set_all_for_tid), 0, 0, 1, 1},
  {&__pyx_n_s_PyCache_values, __pyx_k_PyCache_values, sizeof(__pyx_k_PyCache_values), 0, 0, 1, 1},
  {&__pyx_n_s_PyGeneration, __pyx_k_PyGeneration, sizeof(__pyx_k_PyGeneration), 0, 0, 1, 1},
  {&__pyx_n_s_PyGeneration___iter, __pyx_k_PyGeneration___iter, sizeof(__pyx_k_PyGeneration___iter), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_SingleValue_sizeof, __pyx_k_SingleValue_sizeof, sizeof(__pyx_k_SingleValue_sizeof), 0, 0, 1, 1},
  {&__pyx_kp_b__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 0, 0},
  {&__pyx_n_s__38, __pyx_k__38, sizeof(__pyx_k__38), 0, 0, 1, 1},
  {&__pyx_n_s__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 0, 1, 1},
  {&__pyx_n_s_add_MRUs, __pyx_k_add_MRUs, sizeof(__pyx_k_add_MRUs), 0, 0, 1, 1},
  {&__pyx_n_s_added_oids, __pyx_k_added_oids, sizeof(__pyx_k_added_oids), 0, 0, 1, 1},
  {&__pyx_n_s_age_frequencies, __pyx_k_age_frequencies, sizeof(__pyx_k_age_frequencies), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_compress, __pyx_k_compress, sizeof(__pyx_k_compress), 0, 0, 1, 1},
  {&__pyx_n_s_contains_oid_with_tid, __pyx_k_contains_oid_with_tid, sizeof(__pyx_k_contains_oid_with_tid), 0, 0, 1, 1},
  {&__pyx_n_s_del_oids, __pyx_k_del_oids, sizeof(__pyx_k_del_oids), 0, 0, 1, 1},
  {&__pyx_n_s_delitems, __pyx_k_delitems, sizeof(__pyx_k_delitems), 0, 0, 1, 1},
  {&__pyx_n_s_eden, __pyx_k_eden, sizeof(__pyx_k_eden), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_filler, __pyx_k_filler, sizeof(__pyx_k_filler), 0, 0, 1, 1},
  {&__pyx_n_s_freeze, __pyx_k_freeze, sizeof(__pyx_k_freeze), 0, 0, 1, 1},
  {&__pyx_n_s_frequency, __pyx_k_frequency, sizeof(__pyx_k_frequency), 0, 0, 1, 1},
  {&__pyx_n_s_frozen, __pyx_k_frozen, sizeof(__pyx_k_frozen), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_get_item_with_tid, __pyx_k_get_item_with_tid, sizeof(__pyx_k_get_item_with_tid), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
  {&__pyx_n_s_it, __pyx_k_it, sizeof(__pyx_k_it), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_iteroiditems, __pyx_k_iteroiditems, sizeof(__pyx_k_iteroiditems), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_number_nodes, __pyx_k_number_nodes, sizeof(__pyx_k_number_nodes), 0, 0, 1, 1},
  {&__pyx_n_s_oid, __pyx_k_oid, sizeof(__pyx_k_oid), 0, 0, 1, 1},
  {&__pyx_n_s_oids, __pyx_k_oids, sizeof(__pyx_k_oids), 0, 0, 1, 1},
  {&__pyx_n_s_oids_tids, __pyx_k_oids_tids, sizeof(__pyx_k_oids_tids), 0, 0, 1, 1},
  {&__pyx_n_s_ordered_keys, __pyx_k_ordered_keys, sizeof(__pyx_k_ordered_keys), 0, 0, 1, 1},
  {&__pyx_n_s_peek, __pyx_k_peek, sizeof(__pyx_k_peek), 0, 0, 1, 1},
  {&__pyx_n_s_peek_item_with_tid, __pyx_k_peek_item_with_tid, sizeof(__pyx_k_peek_item_with_tid), 0, 0, 1, 1},
  {&__pyx_n_s_probation, __pyx_k_probation, sizeof(__pyx_k_probation), 0, 0, 1, 1},
  {&__pyx_n_s_protected, __pyx_k_protected, sizeof(__pyx_k_protected), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_relstorage__compat, __pyx_k_relstorage__compat, sizeof(__pyx_k_relstorage__compat), 0, 0, 1, 1},
  {&__pyx_n_s_relstorage_cache_cache, __pyx_k_relstorage_cache_cache, sizeof(__pyx_k_relstorage_cache_cache), 0, 0, 1, 1},
  {&__pyx_n_s_relstorage_cache_interfaces, __pyx_k_relstorage_cache_interfaces, sizeof(__pyx_k_relstorage_cache_interfaces), 0, 0, 1, 1},
  {&__pyx_n_s_reset_stats, __pyx_k_reset_stats, sizeof(__pyx_k_reset_stats), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_return_count_only, __pyx_k_return_count_only, sizeof(__pyx_k_return_count_only), 0, 0, 1, 1},
  {&__pyx_kp_s_s_r_s_frozen_s, __pyx_k_s_r_s_frozen_s, sizeof(__pyx_k_s_r_s_frozen_s), 0, 0, 1, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_set_all_for_tid, __pyx_k_set_all_for_tid, sizeof(__pyx_k_set_all_for_tid), 0, 0, 1, 1},
  {&__pyx_n_s_sizeof, __pyx_k_sizeof, sizeof(__pyx_k_sizeof), 0, 0, 1, 1},
  {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
  {&__pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_k_src_relstorage_cache_cache_pyx, sizeof(__pyx_k_src_relstorage_cache_cache_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
  {&__pyx_n_s_state_oid_iter, __pyx_k_state_oid_iter, sizeof(__pyx_k_state_oid_iter), 0, 0, 1, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_tid, __pyx_k_tid, sizeof(__pyx_k_tid), 0, 0, 1, 1},
  {&__pyx_n_s_tid_int, __pyx_k_tid_int, sizeof(__pyx_k_tid_int), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_value_limit, __pyx_k_value_limit, sizeof(__pyx_k_value_limit), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  #endif
  {0, 0, 0, 0, 0, 0, 0}
};
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 193, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 398, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 493, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(1, 61, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "relstorage/cache/cache.pyx":81
 *     if not sve_p and not mve_p:
 *         # Most likely, entry is null
 *         raise AssertionError("Invalid pointer cast")             # <<<<<<<<<<<<<<
 * 
 *     # Keeping in mind the semantics of our constructors
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_Invalid_pointer_cast); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "relstorage/cache/cache.pyx":127
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 88
 *         return sizeof(SVCacheEntry)
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_codeobj__8 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__7, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_sizeof, 127, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__8)) __PYX_ERR(0, 127, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":216
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 72.
 *         return sizeof(MVCacheEntry)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__9, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_sizeof, 216, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(0, 216, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":299
 *         self.sets = self.hits = self.misses = 0
 * 
 *     cpdef reset_stats(self):             # <<<<<<<<<<<<<<
 *         self.hits = self.sets = self.misses = 0
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__11, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_reset_stats, 299, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 299, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":343
 *         return self.cache.size()
 * 
 *     cpdef CachedValue get(self, OID_t key):             # <<<<<<<<<<<<<<
 *         entry = self.cache.get(key)
 *         if not entry:
 */
  __pyx_tuple__13 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_key); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__13, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_get, 343, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 343, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":349
 *         return python_from_entry_p(entry)
 * 
 *     cpdef CachedValue peek(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */
  __pyx_tuple__15 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_key); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_peek, 349, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 349, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":352
 *         return self.get(key)
 * 
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):             # <<<<<<<<<<<<<<
 *         value = self.cache.peek(key, tid)
 *         if value:
 */
  __pyx_tuple__17 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_tid); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_peek_item_with_tid, 352, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 352, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":357
 *             return python_from_entry_p(value)
 * 
 *     cpdef bint contains_oid_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         """
 *         For use during cache validation. Can only be used when
 */
  __pyx_tuple__19 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_tid); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_contains_oid_with_tid, 357, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 357, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":375
 *         return self.get(key)
 * 
 *     cpdef get_item_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)
 */
  __pyx_tuple__21 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_key, __pyx_n_s_tid); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_get_item_with_tid, 375, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 375, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":428
 *             preincr(it)
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the oid/cache_value pairs.
 */
  __pyx_tuple__23 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_it, __pyx_n_s_end); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_iteritems, 428, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(0, 428, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":441
 *             preincr(it)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the OIDs in the cache.
 */
  __pyx_tuple__24 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_it, __pyx_n_s_end); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__4 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_keys, 441, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__4)) __PYX_ERR(0, 441, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":454
 *             preincr(it)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the values in the cache.
 */
  __pyx_tuple__25 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_it, __pyx_n_s_end); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_values, 454, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 454, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":469
 *     # Cache specific operations
 * 
 *     cpdef set_all_for_tid(self, TID_t tid_int, state_oid_iter, compress, Py_ssize_t value_limit):             # <<<<<<<<<<<<<<
 *         # Do all this down here so we don't give up the GIL, assuming the iter is actually a list
 *         # or otherwise not implemented in pure-python
 */
  __pyx_tuple__26 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_tid_int, __pyx_n_s_state_oid_iter, __pyx_n_s_compress, __pyx_n_s_value_limit); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_set_all_for_tid, 469, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 469, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":482
 * 
 * 
 *     def add_MRUs(self, ordered_keys, return_count_only=False):             # <<<<<<<<<<<<<<
 *         cdef OID_t key
 *         cdef TID_t tid
 */
  __pyx_tuple__28 = PyTuple_Pack(14, __pyx_n_s_self, __pyx_n_s_ordered_keys, __pyx_n_s_return_count_only, __pyx_n_s_key, __pyx_n_s_tid, __pyx_n_s_i, __pyx_n_s_number_nodes, __pyx_n_s_filler, __pyx_n_s_state, __pyx_n_s_frozen, __pyx_n_s_frequency, __pyx_n_s_added_oids, __pyx_n_s_result, __pyx_n_s_key); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 14, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_add_MRUs, 482, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 482, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":511
 *         return result
 * 
 *     def age_frequencies(self):             # <<<<<<<<<<<<<<
 *         self.cache.age_frequencies()
 * 
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_age_frequencies, 511, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 511, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":514
 *         self.cache.age_frequencies()
 * 
 *     def delitems(self, oids_tids):             # <<<<<<<<<<<<<<
 *         """
 *         For each OID/TID pair in the items, remove all cached values
 */
  __pyx_tuple__32 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_oids_tids, __pyx_n_s_oid, __pyx_n_s_tid); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_delitems, 514, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 514, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":525
 *             self.cache.delitem(oid, tid)
 * 
 *     def del_oids(self, oids):             # <<<<<<<<<<<<<<
 *         """
 *         For each oid in OIDs, remove it.
 */
  __pyx_tuple__34 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_oids, __pyx_n_s_oid); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_del_oids, 525, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 525, __pyx_L1_error)

  /* "relstorage/cache/cache.pyx":532
 *             self.cache.delitem(oid)
 * 
 *     def freeze(self, oids_tids):             # <<<<<<<<<<<<<<
 *         # The idea is to *move* the data, or make it available,
 *         # *without* copying it.
 */
  __pyx_tuple__36 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_oids_tids, __pyx_n_s_oid, __pyx_n_s_tid); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_relstorage_cache_cache_pyx, __pyx_n_s_freeze, 532, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_InitString(__pyx_string_tab[0], &__pyx_n_s_AssertionError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[1], &__pyx_n_s_CacheConsistencyError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[2], &__pyx_n_s_CachedValue) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[3], &__pyx_n_s_IndexError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[4], &__pyx_kp_s_Invalid_pointer_cast) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[5], &__pyx_n_s_MultipleValues) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[6], &__pyx_n_s_MultipleValues_sizeof) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[7], &__pyx_n_s_NoSuchGeneration) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[8], &__pyx_n_s_NotImplemented) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[9], &__pyx_n_s_PyCache) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[10], &__pyx_n_s_PyCache___iter) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[11], &__pyx_n_s_PyCache_add_MRUs) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[12], &__pyx_n_s_PyCache_age_frequencies) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[13], &__pyx_n_s_PyCache_contains_oid_with_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[14], &__pyx_n_s_PyCache_del_oids) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[15], &__pyx_n_s_PyCache_delitems) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[16], &__pyx_n_s_PyCache_freeze) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[17], &__pyx_n_s_PyCache_get) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[18], &__pyx_n_s_PyCache_get_item_with_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[19], &__pyx_n_s_PyCache_iteritems) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[20], &__pyx_n_s_PyCache_keys) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[21], &__pyx_n_s_PyCache_peek) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[22], &__pyx_n_s_PyCache_peek_item_with_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[23], &__pyx_n_s_PyCache_reset_stats) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[24], &__pyx_n_s_PyCache_set_all_for_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[25], &__pyx_n_s_PyCache_values) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[26], &__pyx_n_s_PyGeneration) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[27], &__pyx_n_s_PyGeneration___iter) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[28], &__pyx_n_s_RuntimeError) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[29], &__pyx_n_s_SingleValue_sizeof) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[30], &__pyx_kp_b__2) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[31], &__pyx_n_s__38) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[32], &__pyx_n_s__6) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[33], &__pyx_n_s_add_MRUs) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[34], &__pyx_n_s_added_oids) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[35], &__pyx_n_s_age_frequencies) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[36], &__pyx_n_s_args) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[37], &__pyx_n_s_class) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[38], &__pyx_n_s_cline_in_traceback) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[39], &__pyx_n_s_close) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[40], &__pyx_n_s_compress) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[41], &__pyx_n_s_contains_oid_with_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[42], &__pyx_n_s_del_oids) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[43], &__pyx_n_s_delitems) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[44], &__pyx_n_s_eden) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[45], &__pyx_n_s_end) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[46], &__pyx_n_s_enumerate) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[47], &__pyx_n_s_filler) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[48], &__pyx_n_s_freeze) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[49], &__pyx_n_s_frequency) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[50], &__pyx_n_s_frozen) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[51], &__pyx_n_s_get) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[52], &__pyx_n_s_get_item_with_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[53], &__pyx_n_s_i) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[54], &__pyx_n_s_import) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[55], &__pyx_n_s_initializing) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[56], &__pyx_n_s_it) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[57], &__pyx_n_s_iter) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[58], &__pyx_n_s_iteritems) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[59], &__pyx_n_s_iteroiditems) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[60], &__pyx_n_s_key) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[61], &__pyx_n_s_keys) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[62], &__pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[63], &__pyx_n_s_name) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[64], &__pyx_n_s_number_nodes) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[65], &__pyx_n_s_oid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[66], &__pyx_n_s_oids) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[67], &__pyx_n_s_oids_tids) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[68], &__pyx_n_s_ordered_keys) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[69], &__pyx_n_s_peek) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[70], &__pyx_n_s_peek_item_with_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[71], &__pyx_n_s_probation) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[72], &__pyx_n_s_protected) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[73], &__pyx_n_s_pyx_vtable) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[74], &__pyx_n_s_range) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[75], &__pyx_n_s_relstorage__compat) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[76], &__pyx_n_s_relstorage_cache_cache) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[77], &__pyx_n_s_relstorage_cache_interfaces) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[78], &__pyx_n_s_reset_stats) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[79], &__pyx_n_s_result) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[80], &__pyx_n_s_return_count_only) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[81], &__pyx_kp_s_s_r_s_frozen_s) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[82], &__pyx_n_s_self) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[83], &__pyx_n_s_send) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[84], &__pyx_n_s_set_all_for_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[85], &__pyx_n_s_sizeof) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[86], &__pyx_n_s_spec) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[87], &__pyx_kp_s_src_relstorage_cache_cache_pyx) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[88], &__pyx_n_s_state) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[89], &__pyx_n_s_state_oid_iter) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[90], &__pyx_n_s_sys) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[91], &__pyx_n_s_test) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[92], &__pyx_n_s_throw) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[93], &__pyx_n_s_tid) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[94], &__pyx_n_s_tid_int) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[95], &__pyx_n_s_value) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[96], &__pyx_n_s_value_limit) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  if (__Pyx_InitString(__pyx_string_tab[97], &__pyx_n_s_values) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  #endif
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache_CachedValue = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache_CachedValue_spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache_CachedValue)) __PYX_ERR(0, 71, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache_CachedValue) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache_CachedValue.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache_CachedValue.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache_CachedValue.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache_CachedValue.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CachedValue, __pyx_type_10relstorage_5cache_5cache_CachedValue) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  #else
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_CachedValue, (PyObject *)&__pyx_type_10relstorage_5cache_5cache_CachedValue) < 0) __PYX_ERR(0, 71, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache_CachedValue = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache_CachedValue;
  #else
  __pyx_ptype_10relstorage_5cache_5cache_CachedValue = &__pyx_type_10relstorage_5cache_5cache_CachedValue;
  #endif
  __pyx_vtabptr_10relstorage_5cache_5cache_SingleValue = &__pyx_vtable_10relstorage_5cache_5cache_SingleValue;
  __pyx_vtable_10relstorage_5cache_5cache_SingleValue.from_entry = (struct __pyx_obj_10relstorage_5cache_5cache_SingleValue *(*)(__pyx_t_10relstorage_5cache_5cache_SVCacheEntry_p))__pyx_f_10relstorage_5cache_5cache_11SingleValue_from_entry;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_10relstorage_5cache_5cache_CachedValue); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_type_10relstorage_5cache_5cache_SingleValue = PyType_FromSpecWithBases(&__pyx_type_10relstorage_5cache_5cache_SingleValue_spec, __pyx_t_1); if (unlikely(!__pyx_type_10relstorage_5cache_5cache_SingleValue)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  #else
  __pyx_type_10relstorage_5cache_5cache_SingleValue.tp_base = __pyx_ptype_10relstorage_5cache_5cache_CachedValue;
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache_SingleValue) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache_SingleValue.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache_SingleValue.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache_SingleValue.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache_SingleValue.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_SingleValue, __pyx_vtabptr_10relstorage_5cache_5cache_SingleValue) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_SingleValue.tp_dict, __pyx_vtabptr_10relstorage_5cache_5cache_SingleValue) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache_SingleValue = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache_SingleValue;
  #else
  __pyx_ptype_10relstorage_5cache_5cache_SingleValue = &__pyx_type_10relstorage_5cache_5cache_SingleValue;
  #endif
  __pyx_vtabptr_10relstorage_5cache_5cache_MultipleValues = &__pyx_vtable_10relstorage_5cache_5cache_MultipleValues;
  __pyx_vtable_10relstorage_5cache_5cache_MultipleValues.from_entry = (struct __pyx_obj_10relstorage_5cache_5cache_MultipleValues *(*)(__pyx_t_10relstorage_5cache_5cache_MVCacheEntry_p))__pyx_f_10relstorage_5cache_5cache_14MultipleValues_from_entry;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_10relstorage_5cache_5cache_CachedValue); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_type_10relstorage_5cache_5cache_MultipleValues = PyType_FromSpecWithBases(&__pyx_type_10relstorage_5cache_5cache_MultipleValues_spec, __pyx_t_1); if (unlikely(!__pyx_type_10relstorage_5cache_5cache_MultipleValues)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  #else
  __pyx_type_10relstorage_5cache_5cache_MultipleValues.tp_base = __pyx_ptype_10relstorage_5cache_5cache_CachedValue;
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache_MultipleValues) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache_MultipleValues.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache_MultipleValues.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache_MultipleValues.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache_MultipleValues.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_MultipleValues, __pyx_vtabptr_10relstorage_5cache_5cache_MultipleValues) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_MultipleValues.tp_dict, __pyx_vtabptr_10relstorage_5cache_5cache_MultipleValues) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MultipleValues, __pyx_type_10relstorage_5cache_5cache_MultipleValues) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
  #else
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MultipleValues, (PyObject *)&__pyx_type_10relstorage_5cache_5cache_MultipleValues) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache_MultipleValues = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache_MultipleValues;
  #else
  __pyx_ptype_10relstorage_5cache_5cache_MultipleValues = &__pyx_type_10relstorage_5cache_5cache_MultipleValues;
  #endif
  __pyx_vtabptr_10relstorage_5cache_5cache_PyGeneration = &__pyx_vtable_10relstorage_5cache_5cache_PyGeneration;
  __pyx_vtable_10relstorage_5cache_5cache_PyGeneration.from_generation = (PyObject *(*)(relstorage::cache::Generation &, PyObject *, PyObject *))__pyx_f_10relstorage_5cache_5cache_12PyGeneration_from_generation;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache_PyGeneration = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache_PyGeneration_spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache_PyGeneration)) __PYX_ERR(0, 245, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache_PyGeneration) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache_PyGeneration.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache_PyGeneration.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache_PyGeneration.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache_PyGeneration.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_10relstorage_5cache_5cache_PyGeneration, "__iter__"); if (unlikely(!wrapper)) __PYX_ERR(0, 245, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_10relstorage_5cache_5cache_12PyGeneration_4__iter__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_10relstorage_5cache_5cache_12PyGeneration_4__iter__.doc = __pyx_doc_10relstorage_5cache_5cache_12PyGeneration_4__iter__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_10relstorage_5cache_5cache_12PyGeneration_4__iter__;
    }
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_PyGeneration, __pyx_vtabptr_10relstorage_5cache_5cache_PyGeneration) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_PyGeneration.tp_dict, __pyx_vtabptr_10relstorage_5cache_5cache_PyGeneration) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PyGeneration, __pyx_type_10relstorage_5cache_5cache_PyGeneration) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  #else
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PyGeneration, (PyObject *)&__pyx_type_10relstorage_5cache_5cache_PyGeneration) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache_PyGeneration = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache_PyGeneration;
  #else
  __pyx_ptype_10relstorage_5cache_5cache_PyGeneration = &__pyx_type_10relstorage_5cache_5cache_PyGeneration;
  #endif
  __pyx_vtabptr_10relstorage_5cache_5cache_PyCache = &__pyx_vtable_10relstorage_5cache_5cache_PyCache;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache.reset_stats = (PyObject *(*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, int __pyx_skip_dispatch))__pyx_f_10relstorage_5cache_5cache_7PyCache_reset_stats;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache.get = (struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *(*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, int __pyx_skip_dispatch))__pyx_f_10relstorage_5cache_5cache_7PyCache_get;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache.peek = (struct __pyx_obj_10relstorage_5cache_5cache_CachedValue *(*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, int __pyx_skip_dispatch))__pyx_f_10relstorage_5cache_5cache_7PyCache_peek;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache.peek_item_with_tid = (PyObject *(*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, relstorage::cache::TID_t, int __pyx_skip_dispatch))__pyx_f_10relstorage_5cache_5cache_7PyCache_peek_item_with_tid;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache.contains_oid_with_tid = (int (*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, int __pyx_skip_dispatch))__pyx_f_10relstorage_5cache_5cache_7PyCache_contains_oid_with_tid;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache.get_item_with_tid = (PyObject *(*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, int __pyx_skip_dispatch))__pyx_f_10relstorage_5cache_5cache_7PyCache_get_item_with_tid;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache._do_set = (PyObject *(*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::OID_t, PyObject *, relstorage::cache::TID_t))__pyx_f_10relstorage_5cache_5cache_7PyCache__do_set;
  __pyx_vtable_10relstorage_5cache_5cache_PyCache.set_all_for_tid = (PyObject *(*)(struct __pyx_obj_10relstorage_5cache_5cache_PyCache *, relstorage::cache::TID_t, PyObject *, PyObject *, Py_ssize_t, int __pyx_skip_dispatch))__pyx_f_10relstorage_5cache_5cache_7PyCache_set_all_for_tid;
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache_PyCache = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache_PyCache_spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache_PyCache)) __PYX_ERR(0, 289, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache_PyCache) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache_PyCache.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache_PyCache.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache_PyCache.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache_PyCache.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_10relstorage_5cache_5cache_PyCache, "__iter__"); if (unlikely(!wrapper)) __PYX_ERR(0, 289, __pyx_L1_error)
    if (__Pyx_IS_TYPE(wrapper, &PyWrapperDescr_Type)) {
      __pyx_wrapperbase_10relstorage_5cache_5cache_7PyCache_26__iter__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_10relstorage_5cache_5cache_7PyCache_26__iter__.doc = __pyx_doc_10relstorage_5cache_5cache_7PyCache_26__iter__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_10relstorage_5cache_5cache_7PyCache_26__iter__;
    }
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_PyCache, __pyx_vtabptr_10relstorage_5cache_5cache_PyCache) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  #else
  if (__Pyx_SetVtable(__pyx_type_10relstorage_5cache_5cache_PyCache.tp_dict, __pyx_vtabptr_10relstorage_5cache_5cache_PyCache) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PyCache, __pyx_type_10relstorage_5cache_5cache_PyCache) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  #else
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PyCache, (PyObject *)&__pyx_type_10relstorage_5cache_5cache_PyCache) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache_PyCache = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache_PyCache;
  #else
  __pyx_ptype_10relstorage_5cache_5cache_PyCache = &__pyx_type_10relstorage_5cache_5cache_PyCache;
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__ = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter___spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__)) __PYX_ERR(0, 279, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 279, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__ = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__;
  #else
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct____iter__ = &__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct____iter__;
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter___spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__)) __PYX_ERR(0, 406, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__) < 0) __PYX_ERR(0, 406, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__;
  #else
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__ = &__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_1___iter__;
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems_spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems)) __PYX_ERR(0, 428, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems) < 0) __PYX_ERR(0, 428, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems;
  #else
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems = &__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_2_iteritems;
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys_spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys)) __PYX_ERR(0, 441, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys) < 0) __PYX_ERR(0, 441, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys;
  #else
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_3_keys = &__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_3_keys;
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values = PyType_FromSpec(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values_spec); if (unlikely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values)) __PYX_ERR(0, 454, __pyx_L1_error)
  #else
  if (PyType_Ready(&__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values) < 0) __PYX_ERR(0, 454, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3
  __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values.tp_dictoffset && __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values = (PyTypeObject *)__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values;
  #else
  __pyx_ptype_10relstorage_5cache_5cache___pyx_scope_struct_4_values = &__pyx_type_10relstorage_5cache_5cache___pyx_scope_struct_4_values;
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cache(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cache},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "cache",
      __pyx_k_Python_wrappers_for_the_generat, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_COMPILING_IN_LIMITED_API
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_COMPILING_IN_LIMITED_API
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcache(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcache(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cache(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cache(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_cache(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'cache' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cache", __pyx_methods, __pyx_k_Python_wrappers_for_the_generat, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_COMPILING_IN_LIMITED_API
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyState_AddModule(__pyx_t_1, &__pyx_moduledef) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cache(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_relstorage__cache__cache) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "relstorage.cache.cache")) {
      if (unlikely(PyDict_SetItemString(modules, "relstorage.cache.cache", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "relstorage/cache/cache.pyx":37
 * from relstorage.cache.c_cache cimport ProposedCacheEntry
 * 
 * import sys             # <<<<<<<<<<<<<<
 * from relstorage.cache.interfaces import NoSuchGeneration
 * from relstorage.cache.interfaces import CacheConsistencyError
 */
  __pyx_t_1 = __Pyx_ImportDottedModule(__pyx_n_s_sys, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":38
 * 
 * import sys
 * from relstorage.cache.interfaces import NoSuchGeneration             # <<<<<<<<<<<<<<
 * from relstorage.cache.interfaces import CacheConsistencyError
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_NoSuchGeneration);
  __Pyx_GIVEREF(__pyx_n_s_NoSuchGeneration);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_NoSuchGeneration);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_relstorage_cache_interfaces, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_NoSuchGeneration); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NoSuchGeneration, __pyx_t_1) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "relstorage/cache/cache.pyx":39
 * import sys
 * from relstorage.cache.interfaces import NoSuchGeneration
 * from relstorage.cache.interfaces import CacheConsistencyError             # <<<<<<<<<<<<<<
 * 
 * from relstorage._compat import iteroiditems
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_CacheConsistencyError);
  __Pyx_GIVEREF(__pyx_n_s_CacheConsistencyError);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_CacheConsistencyError);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_relstorage_cache_interfaces, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CacheConsistencyError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CacheConsistencyError, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "relstorage/cache/cache.pyx":41
 * from relstorage.cache.interfaces import CacheConsistencyError
 * 
 * from relstorage._compat import iteroiditems             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_iteroiditems);
  __Pyx_GIVEREF(__pyx_n_s_iteroiditems);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_iteroiditems);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_relstorage__compat, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_iteroiditems); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_iteroiditems, __pyx_t_1) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "relstorage/cache/cache.pyx":127
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 88
 *         return sizeof(SVCacheEntry)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_11SingleValue_5sizeof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SingleValue_sizeof, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__8)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_SingleValue->tp_dict, __pyx_n_s_sizeof, __pyx_t_2) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_SingleValue);

  /* "relstorage/cache/cache.pyx":216
 *         release_entry(&self.entry)
 * 
 *     def sizeof(self):             # <<<<<<<<<<<<<<
 *         # At this writing, reports 72.
 *         return sizeof(MVCacheEntry)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_14MultipleValues_3sizeof, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MultipleValues_sizeof, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__10)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_MultipleValues->tp_dict, __pyx_n_s_sizeof, __pyx_t_2) < 0) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_MultipleValues);

  /* "relstorage/cache/cache.pyx":299
 *         self.sets = self.hits = self.misses = 0
 * 
 *     cpdef reset_stats(self):             # <<<<<<<<<<<<<<
 *         self.hits = self.sets = self.misses = 0
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_3reset_stats, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_reset_stats, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_reset_stats, __pyx_t_2) < 0) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":343
 *         return self.cache.size()
 * 
 *     cpdef CachedValue get(self, OID_t key):             # <<<<<<<<<<<<<<
 *         entry = self.cache.get(key)
 *         if not entry:
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_11get, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_get, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__14)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_get, __pyx_t_2) < 0) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":349
 *         return python_from_entry_p(entry)
 * 
 *     cpdef CachedValue peek(self, OID_t key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_13peek, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_peek, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__16)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_peek, __pyx_t_2) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":352
 *         return self.get(key)
 * 
 *     cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):             # <<<<<<<<<<<<<<
 *         value = self.cache.peek(key, tid)
 *         if value:
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_15peek_item_with_tid, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_peek_item_with_tid, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_peek_item_with_tid, __pyx_t_2) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":357
 *             return python_from_entry_p(value)
 * 
 *     cpdef bint contains_oid_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         """
 *         For use during cache validation. Can only be used when
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_17contains_oid_with_tid, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_contains_oid_with_tid, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_contains_oid_with_tid, __pyx_t_2) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":375
 *         return self.get(key)
 * 
 *     cpdef get_item_with_tid(self, OID_t key, tid):             # <<<<<<<<<<<<<<
 *         cdef TID_t native_tid = -1 if tid is None else tid
 *         cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_21get_item_with_tid, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_get_item_with_tid, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_get_item_with_tid, __pyx_t_2) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":428
 *             preincr(it)
 * 
 *     def iteritems(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the oid/cache_value pairs.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_30iteritems, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_iteritems, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_iteritems, __pyx_t_2) < 0) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":441
 *             preincr(it)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the OIDs in the cache.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_33keys, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_keys, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_keys, __pyx_t_2) < 0) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":454
 *             preincr(it)
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         """
 *         Iterate across the values in the cache.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_36values, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_values, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__5)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_values, __pyx_t_2) < 0) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":469
 *     # Cache specific operations
 * 
 *     cpdef set_all_for_tid(self, TID_t tid_int, state_oid_iter, compress, Py_ssize_t value_limit):             # <<<<<<<<<<<<<<
 *         # Do all this down here so we don't give up the GIL, assuming the iter is actually a list
 *         # or otherwise not implemented in pure-python
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_39set_all_for_tid, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_set_all_for_tid, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_set_all_for_tid, __pyx_t_2) < 0) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":482
 * 
 * 
 *     def add_MRUs(self, ordered_keys, return_count_only=False):             # <<<<<<<<<<<<<<
 *         cdef OID_t key
 *         cdef TID_t tid
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_41add_MRUs, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_add_MRUs, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_add_MRUs, __pyx_t_2) < 0) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":511
 *         return result
 * 
 *     def age_frequencies(self):             # <<<<<<<<<<<<<<
 *         self.cache.age_frequencies()
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_43age_frequencies, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_age_frequencies, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_age_frequencies, __pyx_t_2) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":514
 *         self.cache.age_frequencies()
 * 
 *     def delitems(self, oids_tids):             # <<<<<<<<<<<<<<
 *         """
 *         For each OID/TID pair in the items, remove all cached values
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_45delitems, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_delitems, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_delitems, __pyx_t_2) < 0) __PYX_ERR(0, 514, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":525
 *             self.cache.delitem(oid, tid)
 * 
 *     def del_oids(self, oids):             # <<<<<<<<<<<<<<
 *         """
 *         For each oid in OIDs, remove it.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_47del_oids, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_del_oids, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_del_oids, __pyx_t_2) < 0) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":532
 *             self.cache.delitem(oid)
 * 
 *     def freeze(self, oids_tids):             # <<<<<<<<<<<<<<
 *         # The idea is to *move* the data, or make it available,
 *         # *without* copying it.
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10relstorage_5cache_5cache_7PyCache_49freeze, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PyCache_freeze, NULL, __pyx_n_s_relstorage_cache_cache, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_10relstorage_5cache_5cache_PyCache->tp_dict, __pyx_n_s_freeze, __pyx_t_2) < 0) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10relstorage_5cache_5cache_PyCache);

  /* "relstorage/cache/cache.pyx":1
 * # -*- coding: utf-8 -*-             # <<<<<<<<<<<<<<
 * # distutils: language = c++
 * # cython: auto_pickle=False,embedsignature=True,always_allow_keywords=False,infer_types=True
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "vector.to_py":60
 * 
 * @cname("__pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t")
 * cdef object __pyx_convert_vector_to_py_relstorage_3a__3a_cache_3a__3a_TID_t(vector[X]& v):             # <<<<<<<<<<<<<<
 *     return [v[i] for i in range(v.size())]
 * 
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init relstorage.cache.cache", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(__pyx_m);
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init relstorage.cache.cache");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY ||  CYTHON_COMPILING_IN_LIMITED_API
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;  // error
            return kwvalues[i];
        }
    }
    return NULL;  // not found (no exception set)
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        if (unlikely(PyTuple_GET_SIZE(kw) == 0))
            return 1;
        if (!kw_allowed) {
            key = PyTuple_GET_ITEM(kw, 0);
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < PyTuple_GET_SIZE(kw); pos++) {
            key = PyTuple_GET_ITEM(kw, pos);
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result;
    Py_ssize_t i;
    argstuple = PyTuple_New(nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        PyTuple_SET_ITEM(argstuple, i, args[i]);
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
    Py_DECREF(argstuple);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
#ifdef __Pyx_CyFunction_USED
        if (__Pyx_IsCyOrPyCFunction(func))
#else
        if (PyCFunction_Check(func))
#endif
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
                return __Pyx_PyObject_CallMethO(func, NULL);
            }
        }
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (PyCFunction_Check(func))
        {
            if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
                return __Pyx_PyObject_CallMethO(func, args[0]);
            }
        }
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    #if CYTHON_VECTORCALL
    vectorcallfunc f = _PyVectorcall_Function(func);
    if (f) {
        return f(func, args, nargs, kwargs);
    }
    #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
    if (__Pyx_CyFunction_CheckExact(func)) {
        __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
        if (f) return f(func, args, nargs, kwargs);
    }
    #endif
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    return __Pyx_PyObject_FastCall_fallback(func, args, nargs, kwargs);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* ReturnWithStopIteration */
static void __Pyx__ReturnWithStopIteration(PyObject* value) {
    PyObject *exc, *args;
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_PYSTON
    __Pyx_PyThreadState_declare
    if ((PY_VERSION_HEX >= 0x03030000 && PY_VERSION_HEX < 0x030500B1)
            || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        args = PyTuple_New(1);
        if (unlikely(!args)) return;
        Py_INCREF(value);
        PyTuple_SET_ITEM(args, 0, value);
        exc = PyType_Type.tp_call(PyExc_StopIteration, args, NULL);
        Py_DECREF(args);
        if (!exc) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_type)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(PyExc_StopIteration);
        __Pyx_ErrRestore(PyExc_StopIteration, exc, NULL);
        return;
    }
    #endif
#else
    args = PyTuple_Pack(1, value);
    if (unlikely(!args)) return;
    exc = PyObject_Call(PyExc_StopIteration, args, NULL);
    Py_DECREF(args);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(PyExc_StopIteration, exc);
    Py_DECREF(exc);
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(CYTHON_UNUSED int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        if (kwds_is_tuple) {
            if (pos >= PyTuple_GET_SIZE(kwds)) break;
            key = PyTuple_GET_ITEM(kwds, pos);
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_SetVtable(PyObject *type, void *vtable) {
#else
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#endif
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (!ob)
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (PyObject_SetAttr(type, __pyx_n_s_pyx_vtable, ob) < 0)
#else
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                #if CYTHON_COMPILING_IN_LIMITED_API
                module = PyImport_ImportModuleLevelObject(
                    name, empty_dict, empty_dict, from_list, 1);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                #endif
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            #if CYTHON_COMPILING_IN_LIMITED_API
            module = PyImport_ImportModuleLevelObject(
                name, empty_dict, empty_dict, from_list, level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
  #if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        PyObject *sep;
        PyObject *slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, CYTHON_UNUSED PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__6;
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    Py_ssize_t i, nparts;
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (likely(module))
        return module;
    return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* FetchCommonType */
  static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    PyObject *abi_module = PyImport_AddModule((char*) __PYX_ABI_MODULE_NAME);
    if (!abi_module) return NULL;
    Py_INCREF(abi_module);
    return abi_module;
}
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, type->tp_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              type->tp_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, type->tp_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *py_basicsize, *cached_type = NULL;
    Py_ssize_t basicsize;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, spec->name);
    if (cached_type) {
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (!py_basicsize) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred()) goto bad;
        if (__Pyx_VerifyCachedType(
              cached_type,
              spec->name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    cached_type = PyType_FromSpecWithBases(spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (PyObject_SetAttrString(abi_module, spec->name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
  #if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, Py_ssize_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (i = 0; i < nargs; i++) newargs[i] = args[i];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, Py_ssize_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
  #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if PY_VERSION_HEX < 0x030500A0
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         cyfunc->func.m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", cyfunc->func.m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = cyfunc->func.m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = cyfunc->func.m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = cyfunc->func.m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = cyfunc->func.m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = cyfunc->func.m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = cyfunc->func.m_self;
        break;
    default:
        return NULL;
    }
    return ((_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(__pyx_CyFunctionObject, func.vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#ifdef _Py_TPFLAGS_HAVE_VECTORCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(void) {
#if CYTHON_COMPILING_IN_LIMITED_API
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(&__pyx_CyFunctionType_spec, NULL);
#else
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
  static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if CYTHON_COMPILING_IN_LIMITED_API
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    if (c_line) {
        (void) __pyx_cfilenm;
        c_line = __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    _PyTraceback_Add(funcname, filename, c_line ? -c_line : py_line);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
    const int64_t neg_one = (int64_t) ((int64_t) 0 - (int64_t) 1), const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_relstorage_3a__3a_cache_3a__3a_generation_num(relstorage::cache::generation_num value) {
    const relstorage::cache::generation_num neg_one = (relstorage::cache::generation_num) ((relstorage::cache::generation_num) 0 - (relstorage::cache::generation_num) 1), const_zero = (relstorage::cache::generation_num) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(relstorage::cache::generation_num) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(relstorage::cache::generation_num) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(relstorage::cache::generation_num) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(relstorage::cache::generation_num) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(relstorage::cache::generation_num) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(relstorage::cache::generation_num),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
    const int64_t neg_one = (int64_t) ((int64_t) 0 - (int64_t) 1), const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 2 * PyLong_SHIFT) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 3 * PyLong_SHIFT) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 4 * PyLong_SHIFT) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int64_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            int64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int64_t) -1;
        }
    } else {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* FormatTypeName */
  #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XSETREF(name, __Pyx_NewRef(__pyx_n_s__38));
    }
    return name;
}
#endif

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if (CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyObjectCall2Args */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallOneArg */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
  static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
  static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* PyObjectCallNoArg */
  static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg = NULL;
    return __Pyx_PyObject_FastCall(func, (&arg)+1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* CoroutineBase */
  #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(self->ob_refcnt > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
  static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
  #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
  static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the coroutine"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM+0 >= 0x06000000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
  static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
  #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}
#endif

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
